@@ -2,7 +2,7 @@
 Timezone utility functions.
 """
 
-# Copyright (C) 2020-2021 The Psycopg Team
+# Copyright (C) 2020 The Psycopg Team
 
 import logging
 from typing import Dict, Optional, Union
@@ -28,9 +28,15 @@ def get_tzinfo(pgconn: Optional[PGconn]) -> tzinfo:
         sname = tzname.decode() if tzname else "UTC"
         try:
             zi: tzinfo = ZoneInfo(sname)
-        except KeyError:
+        except (KeyError, OSError):
+            logger.warning("unknown PostgreSQL timezone: %r; will use UTC", sname)
+            zi = timezone.utc
+        except Exception as ex:
             logger.warning(
-                "unknown PostgreSQL timezone: %r; will use UTC", sname
+                "error handling PostgreSQL timezone: %r; will use UTC (%s - %s)",
+                sname,
+                type(ex).__name__,
+                ex,
             )
             zi = timezone.utc
 
