@@ -6,28 +6,31 @@ Adapters for arrays
 
 import re
 import struct
-from typing import Any, cast, Callable, Iterator, List
-from typing import Optional, Pattern, Set, Tuple, Type
+from typing import Any, cast, Callable, List, Optional, Pattern, Set, Tuple, Type
 
 from .. import pq
 from .. import errors as e
 from .. import postgres
-from ..abc import AdaptContext, Buffer, Dumper, DumperKey, NoneType
+from ..abc import AdaptContext, Buffer, Dumper, DumperKey, NoneType, Loader, Transformer
 from ..adapt import RecursiveDumper, RecursiveLoader, PyFormat
-from .._compat import cache
+from .._compat import cache, prod
 from .._struct import pack_len, unpack_len
+from .._cmodule import _psycopg
 from ..postgres import TEXT_OID, INVALID_OID
 from .._typeinfo import TypeInfo
 
 _struct_head = struct.Struct("!III")  # ndims, hasnull, elem oid
 _pack_head = cast(Callable[[int, int, int], bytes], _struct_head.pack)
-_unpack_head = cast(Callable[[bytes], Tuple[int, int, int]], _struct_head.unpack_from)
+_unpack_head = cast(Callable[[Buffer], Tuple[int, int, int]], _struct_head.unpack_from)
 _struct_dim = struct.Struct("!II")  # dim, lower bound
 _pack_dim = cast(Callable[[int, int], bytes], _struct_dim.pack)
-_unpack_dim = cast(Callable[[bytes, int], Tuple[int, int]], _struct_dim.unpack_from)
+_unpack_dim = cast(Callable[[Buffer, int], Tuple[int, int]], _struct_dim.unpack_from)
 
 TEXT_ARRAY_OID = postgres.types["text"].array_oid
 
+PY_TEXT = PyFormat.TEXT
+PQ_BINARY = pq.Format.BINARY
+
 
 class BaseListDumper(RecursiveDumper):
     element_oid = 0
@@ -121,7 +124,7 @@ class ListDumper(BaseListDumper):
             return self.cls
 
         sd = self._tx.get_dumper(item, format)
-        return (self.cls, sd.get_key(item, format))  # type: ignore
+        return (self.cls, sd.get_key(item, format))
 
     def upgrade(self, obj: List[Any], format: PyFormat) -> "BaseListDumper":
         # If we have an oid we don't need to upgrade
@@ -153,7 +156,7 @@ class ListDumper(BaseListDumper):
     _re_esc = re.compile(rb'(["\\])')
 
     def dump(self, obj: List[Any]) -> bytes:
-        tokens: List[bytes] = []
+        tokens: List[Buffer] = []
         needs_quotes = _get_needs_quotes_regexp(self.delimiter).search
 
         def dump_list(obj: List[Any]) -> None:
@@ -187,7 +190,7 @@ class ListDumper(BaseListDumper):
         if self.sub_dumper:
             return self.sub_dumper.dump(item)
         else:
-            return self._tx.get_dumper(item, PyFormat.TEXT).dump(item)
+            return self._tx.get_dumper(item, PY_TEXT).dump(item)
 
 
 @cache
@@ -223,7 +226,7 @@ class ListBinaryDumper(BaseListDumper):
             return (self.cls,)
 
         sd = self._tx.get_dumper(item, format)
-        return (self.cls, sd.get_key(item, format))  # type: ignore
+        return (self.cls, sd.get_key(item, format))
 
     def upgrade(self, obj: List[Any], format: PyFormat) -> "BaseListDumper":
         # If we have an oid we don't need to upgrade
@@ -249,7 +252,7 @@ class ListBinaryDumper(BaseListDumper):
         if not obj:
             return _pack_head(0, 0, sub_oid)
 
-        data: List[bytes] = [b"", b""]  # placeholders to avoid a resize
+        data: List[Buffer] = [b"", b""]  # placeholders to avoid a resize
         dims: List[int] = []
         hasnull = 0
 
@@ -290,124 +293,32 @@ class ListBinaryDumper(BaseListDumper):
         return b"".join(data)
 
 
-class BaseArrayLoader(RecursiveLoader):
-    base_oid: int
-
-
-class ArrayLoader(BaseArrayLoader):
+class ArrayLoader(RecursiveLoader):
 
     delimiter = b","
+    base_oid: int
 
     def load(self, data: Buffer) -> List[Any]:
-        rv = None
-        stack: List[Any] = []
-        cast = self._tx.get_loader(self.base_oid, self.format).load
-
-        # Remove the dimensions information prefix (``[...]=``)
-        if data and data[0] == b"["[0]:
-            if isinstance(data, memoryview):
-                data = bytes(data)
-            idx = data.find(b"=")
-            if idx == -1:
-                raise e.DataError("malformed array, no '=' after dimension information")
-            data = data[idx + 1 :]
-
-        re_parse = _get_array_parse_regexp(self.delimiter)
-        for m in re_parse.finditer(data):
-            t = m.group(1)
-            if t == b"{":
-                a: List[Any] = []
-                if rv is None:
-                    rv = a
-                if stack:
-                    stack[-1].append(a)
-                stack.append(a)
-
-            elif t == b"}":
-                if not stack:
-                    raise e.DataError("malformed array, unexpected '}'")
-                rv = stack.pop()
-
-            else:
-                if not stack:
-                    wat = (
-                        t[:10].decode("utf8", "replace") + "..." if len(t) > 10 else ""
-                    )
-                    raise e.DataError(f"malformed array, unexpected '{wat}'")
-                if t == b"NULL":
-                    v = None
-                else:
-                    if t.startswith(b'"'):
-                        t = self._re_unescape.sub(rb"\1", t[1:-1])
-                    v = cast(t)
-
-                stack[-1].append(v)
-
-        assert rv is not None
-        return rv
-
-    _re_unescape = re.compile(rb"\\(.)")
+        loader = self._tx.get_loader(self.base_oid, self.format)
+        return _load_text(data, loader, self.delimiter)
 
 
-@cache
-def _get_array_parse_regexp(delimiter: bytes) -> Pattern[bytes]:
-    """
-    Return a regexp to tokenize an array representation into item and brackets
-    """
-    return re.compile(
-        rb"""(?xi)
-        (     [{}]                        # open or closed bracket
-            | " (?: [^"\\] | \\. )* "     # or a quoted string
-            | [^"{}%s\\]+                 # or an unquoted non-empty string
-        ) ,?
-        """
-        % delimiter
-    )
-
-
-class ArrayBinaryLoader(BaseArrayLoader):
+class ArrayBinaryLoader(RecursiveLoader):
 
     format = pq.Format.BINARY
 
     def load(self, data: Buffer) -> List[Any]:
-        ndims, hasnull, oid = _unpack_head(data)
-        if not ndims:
-            return []
-
-        fcast = self._tx.get_loader(oid, self.format).load
-
-        p = 12 + 8 * ndims
-        dims = [_unpack_dim(data, i)[0] for i in list(range(12, p, 8))]
-
-        def consume(p: int) -> Iterator[Any]:
-            while True:
-                size = unpack_len(data, p)[0]
-                p += 4
-                if size != -1:
-                    yield fcast(data[p : p + size])
-                    p += size
-                else:
-                    yield None
-
-        items = consume(p)
-
-        def agg(dims: List[int]) -> List[Any]:
-            if not dims:
-                return next(items)
-            else:
-                dim, dims = dims[0], dims[1:]
-                return [agg(dims) for _ in range(dim)]
-
-        return agg(dims)
+        return _load_binary(data, self._tx)
 
 
 def register_array(info: TypeInfo, context: Optional[AdaptContext] = None) -> None:
     if not info.array_oid:
         raise ValueError(f"the type info {info} doesn't describe an array")
 
+    base: Type[Any]
     adapters = context.adapters if context else postgres.adapters
 
-    base: Type[Any] = ArrayLoader
+    base = getattr(_psycopg, "ArrayLoader", ArrayLoader)
     name = f"{info.name.title()}{base.__name__}"
     attribs = {
         "base_oid": info.oid,
@@ -416,10 +327,7 @@ def register_array(info: TypeInfo, context: Optional[AdaptContext] = None) -> No
     loader = type(name, (base,), attribs)
     adapters.register_loader(info.array_oid, loader)
 
-    base = ArrayBinaryLoader
-    name = f"{info.name.title()}{base.__name__}"
-    attribs = {"base_oid": info.oid}
-    loader = type(name, (base,), attribs)
+    loader = getattr(_psycopg, "ArrayBinaryLoader", ArrayBinaryLoader)
     adapters.register_loader(info.array_oid, loader)
 
     base = ListDumper
@@ -459,3 +367,98 @@ def register_all_arrays(context: AdaptContext) -> None:
     for t in context.adapters.types:
         if t.array_oid:
             t.register(context)
+
+
+def _load_text(
+    data: Buffer,
+    loader: Loader,
+    delimiter: bytes = b",",
+    __re_unescape: Pattern[bytes] = re.compile(rb"\\(.)"),
+) -> List[Any]:
+    rv = None
+    stack: List[Any] = []
+    a: List[Any] = []
+    rv = a
+    load = loader.load
+
+    # Remove the dimensions information prefix (``[...]=``)
+    if data and data[0] == b"["[0]:
+        if isinstance(data, memoryview):
+            data = bytes(data)
+        idx = data.find(b"=")
+        if idx == -1:
+            raise e.DataError("malformed array: no '=' after dimension information")
+        data = data[idx + 1 :]
+
+    re_parse = _get_array_parse_regexp(delimiter)
+    for m in re_parse.finditer(data):
+        t = m.group(1)
+        if t == b"{":
+            if stack:
+                stack[-1].append(a)
+            stack.append(a)
+            a = []
+
+        elif t == b"}":
+            if not stack:
+                raise e.DataError("malformed array: unexpected '}'")
+            rv = stack.pop()
+
+        else:
+            if not stack:
+                wat = t[:10].decode("utf8", "replace") + "..." if len(t) > 10 else ""
+                raise e.DataError(f"malformed array: unexpected '{wat}'")
+            if t == b"NULL":
+                v = None
+            else:
+                if t.startswith(b'"'):
+                    t = __re_unescape.sub(rb"\1", t[1:-1])
+                v = load(t)
+
+            stack[-1].append(v)
+
+    assert rv is not None
+    return rv
+
+
+@cache
+def _get_array_parse_regexp(delimiter: bytes) -> Pattern[bytes]:
+    """
+    Return a regexp to tokenize an array representation into item and brackets
+    """
+    return re.compile(
+        rb"""(?xi)
+        (     [{}]                        # open or closed bracket
+            | " (?: [^"\\] | \\. )* "     # or a quoted string
+            | [^"{}%s\\]+                 # or an unquoted non-empty string
+        ) ,?
+        """
+        % delimiter
+    )
+
+
+def _load_binary(data: Buffer, tx: Transformer) -> List[Any]:
+    ndims, hasnull, oid = _unpack_head(data)
+    load = tx.get_loader(oid, PQ_BINARY).load
+
+    if not ndims:
+        return []
+
+    p = 12 + 8 * ndims
+    dims = [_unpack_dim(data, i)[0] for i in range(12, p, 8)]
+    nelems = prod(dims)
+
+    out: List[Any] = [None] * nelems
+    for i in range(nelems):
+        size = unpack_len(data, p)[0]
+        p += 4
+        if size == -1:
+            continue
+        out[i] = load(data[p : p + size])
+        p += size
+
+    # fon ndims > 1 we have to aggregate the array into sub-arrays
+    for dim in dims[-1:0:-1]:
+        out = [out[i : i + dim] for i in range(0, len(out), dim)]
+
+    return out
