@@ -4,15 +4,15 @@ psycopg connection pool base class and functionalities.
 
 # Copyright (C) 2021 The Psycopg Team
 
+from time import monotonic
 from random import random
-from typing import Any, Callable, Deque, Dict, Generic, Optional
-from typing import TYPE_CHECKING
-from collections import Counter, deque
+from typing import Any, Callable, Dict, Generic, Optional, Tuple
 
+from psycopg import errors as e
 from psycopg.abc import ConnectionType
 
-if TYPE_CHECKING:
-    from typing import Counter as TCounter
+from .errors import PoolClosed
+from ._compat import Counter, Deque
 
 
 class BasePool(Generic[ConnectionType]):
@@ -44,21 +44,18 @@ class BasePool(Generic[ConnectionType]):
         kwargs: Optional[Dict[str, Any]] = None,
         min_size: int = 4,
         max_size: Optional[int] = None,
+        open: bool = True,
         name: Optional[str] = None,
         timeout: float = 30.0,
         max_waiting: int = 0,
         max_lifetime: float = 60 * 60.0,
         max_idle: float = 10 * 60.0,
         reconnect_timeout: float = 5 * 60.0,
-        reconnect_failed: Optional[
-            Callable[["BasePool[ConnectionType]"], None]
-        ] = None,
+        reconnect_failed: Optional[Callable[["BasePool[ConnectionType]"], None]] = None,
         num_workers: int = 3,
     ):
-        if max_size is None:
-            max_size = min_size
-        if max_size < min_size:
-            raise ValueError("max_size must be greater or equal than min_size")
+        min_size, max_size = self._check_size(min_size, max_size)
+
         if not name:
             num = BasePool._num_pool = BasePool._num_pool + 1
             name = f"pool-{num}"
@@ -81,8 +78,8 @@ class BasePool(Generic[ConnectionType]):
         self.num_workers = num_workers
 
         self._nconns = min_size  # currently in the pool, out, being prepared
-        self._pool: Deque[ConnectionType] = deque()
-        self._stats: "TCounter[str]" = Counter()
+        self._pool = Deque[ConnectionType]()
+        self._stats = Counter[str]()
 
         # Min number of connections in the pool in a max_idle unit of time.
         # It is reset periodically by the ShrinkPool scheduled task.
@@ -97,9 +94,8 @@ class BasePool(Generic[ConnectionType]):
         # connections to the pool.
         self._growing = False
 
-        # _close should be the last property to be set in the state
-        # to avoid warning on __del__ in case __init__ fails.
-        self._closed = False
+        self._opened = False
+        self._closed = True
 
     def __repr__(self) -> str:
         return (
@@ -120,6 +116,45 @@ class BasePool(Generic[ConnectionType]):
         """`!True` if the pool is closed."""
         return self._closed
 
+    def _check_size(self, min_size: int, max_size: Optional[int]) -> Tuple[int, int]:
+        if max_size is None:
+            max_size = min_size
+
+        if min_size < 0:
+            raise ValueError("min_size cannot be negative")
+        if max_size < min_size:
+            raise ValueError("max_size must be greater or equal than min_size")
+        if min_size == max_size == 0:
+            raise ValueError("if min_size is 0 max_size must be greater or than 0")
+
+        return min_size, max_size
+
+    def _check_open(self) -> None:
+        if self._closed and self._opened:
+            raise e.OperationalError(
+                "pool has already been opened/closed and cannot be reused"
+            )
+
+    def _check_open_getconn(self) -> None:
+        if self._closed:
+            if self._opened:
+                raise PoolClosed(f"the pool {self.name!r} is already closed")
+            else:
+                raise PoolClosed(f"the pool {self.name!r} is not open yet")
+
+    def _check_pool_putconn(self, conn: ConnectionType) -> None:
+        pool = getattr(conn, "_pool", None)
+        if pool is self:
+            return
+
+        if pool:
+            msg = f"it comes from pool {pool.name!r}"
+        else:
+            msg = "it doesn't come from any pool"
+        raise ValueError(
+            f"can't return connection to pool {self.name!r}, {msg}: {conn}"
+        )
+
     def get_stats(self) -> Dict[str, int]:
         """
         Return current stats about the pool usage.
@@ -157,6 +192,13 @@ class BasePool(Generic[ConnectionType]):
         """
         return value * (1.0 + ((max_pc - min_pc) * random()) + min_pc)
 
+    def _set_connection_expiry_date(self, conn: ConnectionType) -> None:
+        """Set an expiry date on a connection.
+
+        Add some randomness to avoid mass reconnection.
+        """
+        conn._expire_at = monotonic() + self._jitter(self.max_lifetime, -0.05, 0.0)
+
 
 class ConnectionAttempt:
     """Keep the state of a connection attempt."""
