@@ -4,25 +4,24 @@ psycopg asynchronous connection pool
 
 # Copyright (C) 2021 The Psycopg Team
 
-import sys
 import asyncio
 import logging
 from abc import ABC, abstractmethod
 from time import monotonic
 from types import TracebackType
-from typing import Any, AsyncIterator, Awaitable, Callable, Deque
-from typing import Dict, List, Optional, Type
+from typing import Any, AsyncIterator, Awaitable, Callable
+from typing import Dict, List, Optional, Sequence, Type
 from weakref import ref
-from collections import deque
+from contextlib import asynccontextmanager
 
 from psycopg import errors as e
+from psycopg import AsyncConnection
 from psycopg.pq import TransactionStatus
-from psycopg._compat import Task, asynccontextmanager, create_task
-from psycopg.connection_async import AsyncConnection
 
 from .base import ConnectionAttempt, BasePool
 from .sched import AsyncScheduler
 from .errors import PoolClosed, PoolTimeout, TooManyRequests
+from ._compat import Task, create_task, Deque
 
 logger = logging.getLogger("psycopg.pool")
 
@@ -32,59 +31,40 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
         self,
         conninfo: str = "",
         *,
+        open: bool = True,
         connection_class: Type[AsyncConnection[Any]] = AsyncConnection,
-        configure: Optional[
-            Callable[[AsyncConnection[Any]], Awaitable[None]]
-        ] = None,
-        reset: Optional[
-            Callable[[AsyncConnection[Any]], Awaitable[None]]
-        ] = None,
+        configure: Optional[Callable[[AsyncConnection[Any]], Awaitable[None]]] = None,
+        reset: Optional[Callable[[AsyncConnection[Any]], Awaitable[None]]] = None,
         **kwargs: Any,
     ):
-        # https://bugs.python.org/issue42600
-        if sys.version_info < (3, 7):
-            raise e.NotSupportedError(
-                "async pool not supported before Python 3.7"
-            )
-
         self.connection_class = connection_class
         self._configure = configure
         self._reset = reset
 
-        self._lock = asyncio.Lock()
-        self._waiting: Deque["AsyncClient"] = deque()
+        # asyncio objects, created on open to attach them to the right loop.
+        self._lock: asyncio.Lock
+        self._sched: AsyncScheduler
+        self._tasks: "asyncio.Queue[MaintenanceTask]"
+
+        self._waiting = Deque["AsyncClient"]()
 
         # to notify that the pool is full
         self._pool_full_event: Optional[asyncio.Event] = None
 
-        self._sched = AsyncScheduler()
-        self._tasks: "asyncio.Queue[MaintenanceTask]" = asyncio.Queue()
+        self._sched_runner: Optional[Task[None]] = None
         self._workers: List[Task[None]] = []
 
         super().__init__(conninfo, **kwargs)
 
-        self._sched_runner = create_task(
-            self._sched.run(), name=f"{self.name}-scheduler"
-        )
-        for i in range(self.num_workers):
-            t = create_task(
-                self.worker(self._tasks),
-                name=f"{self.name}-worker-{i}",
-            )
-            self._workers.append(t)
-
-        # populate the pool with initial min_size connections in background
-        for i in range(self._nconns):
-            self.run_task(AddConnection(self))
-
-        # Schedule a task to shrink the pool if connections over min_size have
-        # remained unused.
-        self.run_task(Schedule(self, ShrinkPool(self), self.max_idle))
+        if open:
+            self._open()
 
     async def wait(self, timeout: float = 30.0) -> None:
+        self._check_open_getconn()
+
         async with self._lock:
             assert not self._pool_full_event
-            if len(self._pool) >= self._nconns:
+            if len(self._pool) >= self._min_size:
                 return
             self._pool_full_event = asyncio.Event()
 
@@ -92,7 +72,7 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
         try:
             await asyncio.wait_for(self._pool_full_event.wait(), timeout)
         except asyncio.TimeoutError:
-            await self.close()  # stop all the threads
+            await self.close()  # stop all the tasks
             raise PoolTimeout(
                 f"pool initialization incomplete after {timeout} sec"
             ) from None
@@ -117,50 +97,29 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
             self._stats[self._USAGE_MS] += int(1000.0 * (t1 - t0))
             await self.putconn(conn)
 
-    async def getconn(
-        self, timeout: Optional[float] = None
-    ) -> AsyncConnection[Any]:
+    async def getconn(self, timeout: Optional[float] = None) -> AsyncConnection[Any]:
         logger.info("connection requested from %r", self.name)
         self._stats[self._REQUESTS_NUM] += 1
+
+        self._check_open_getconn()
+
         # Critical section: decide here if there's a connection ready
         # or if the client needs to wait.
         async with self._lock:
-            if self._closed:
-                raise PoolClosed(f"the pool {self.name!r} is closed")
-
-            pos: Optional[AsyncClient] = None
-            if self._pool:
-                # Take a connection ready out of the pool
-                conn = self._pool.popleft()
-                if len(self._pool) < self._nconns_min:
-                    self._nconns_min = len(self._pool)
-            else:
-                if self.max_waiting and len(self._waiting) >= self.max_waiting:
-                    self._stats[self._REQUESTS_ERRORS] += 1
-                    raise TooManyRequests(
-                        f"the pool {self.name!r} has aleady"
-                        f" {len(self._waiting)} requests waiting"
-                    )
-
+            conn = await self._get_ready_connection(timeout)
+            if not conn:
                 # No connection available: put the client in the waiting queue
                 t0 = monotonic()
                 pos = AsyncClient()
                 self._waiting.append(pos)
                 self._stats[self._REQUESTS_QUEUED] += 1
 
-                # Allow only one thread at time to grow the pool (or returning
-                # connections might be starved).
-                if self._nconns < self._max_size and not self._growing:
-                    self._nconns += 1
-                    logger.info(
-                        "growing pool %r to %s", self.name, self._nconns
-                    )
-                    self._growing = True
-                    self.run_task(AddConnection(self, growing=True))
+                # If there is space for the pool to grow, let's do it
+                self._maybe_grow_pool()
 
         # If we are in the waiting queue, wait to be assigned a connection
         # (outside the critical section, so only the waiting client is locked)
-        if pos:
+        if not conn:
             if timeout is None:
                 timeout = self.timeout
             try:
@@ -179,33 +138,113 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
         logger.info("connection given by %r", self.name)
         return conn
 
-    async def putconn(self, conn: AsyncConnection[Any]) -> None:
-        # Quick check to discard the wrong connection
-        pool = getattr(conn, "_pool", None)
-        if pool is not self:
-            if pool:
-                msg = f"it comes from pool {pool.name!r}"
-            else:
-                msg = "it doesn't come from any pool"
-            raise ValueError(
-                f"can't return connection to pool {self.name!r}, {msg}: {conn}"
+    async def _get_ready_connection(
+        self, timeout: Optional[float]
+    ) -> Optional[AsyncConnection[Any]]:
+        conn: Optional[AsyncConnection[Any]] = None
+        if self._pool:
+            # Take a connection ready out of the pool
+            conn = self._pool.popleft()
+            if len(self._pool) < self._nconns_min:
+                self._nconns_min = len(self._pool)
+        elif self.max_waiting and len(self._waiting) >= self.max_waiting:
+            self._stats[self._REQUESTS_ERRORS] += 1
+            raise TooManyRequests(
+                f"the pool {self.name!r} has already"
+                f" {len(self._waiting)} requests waiting"
             )
+        return conn
+
+    def _maybe_grow_pool(self) -> None:
+        # Allow only one task at time to grow the pool (or returning
+        # connections might be starved).
+        if self._nconns < self._max_size and not self._growing:
+            self._nconns += 1
+            logger.info("growing pool %r to %s", self.name, self._nconns)
+            self._growing = True
+            self.run_task(AddConnection(self, growing=True))
+
+    async def putconn(self, conn: AsyncConnection[Any]) -> None:
+        self._check_pool_putconn(conn)
 
         logger.info("returning connection to %r", self.name)
-
-        # If the pool is closed just close the connection instead of returning
-        # it to the pool. For extra refcare remove the pool reference from it.
-        if self._closed:
-            conn._pool = None
-            await conn.close()
+        if await self._maybe_close_connection(conn):
             return
 
-        # Use a worker to perform eventual maintenance work in a separate thread
+        # Use a worker to perform eventual maintenance work in a separate task
         if self._reset:
             self.run_task(ReturnConnection(self, conn))
         else:
             await self._return_connection(conn)
 
+    async def _maybe_close_connection(self, conn: AsyncConnection[Any]) -> bool:
+        # If the pool is closed just close the connection instead of returning
+        # it to the pool. For extra refcare remove the pool reference from it.
+        if not self._closed:
+            return False
+
+        conn._pool = None
+        await conn.close()
+        return True
+
+    async def open(self, wait: bool = False, timeout: float = 30.0) -> None:
+        # Make sure the lock is created after there is an event loop
+        try:
+            self._lock
+        except AttributeError:
+            self._lock = asyncio.Lock()
+
+        async with self._lock:
+            self._open()
+
+        if wait:
+            await self.wait(timeout=timeout)
+
+    def _open(self) -> None:
+        if not self._closed:
+            return
+
+        # Throw a RuntimeError if the pool is open outside a running loop.
+        asyncio.get_running_loop()
+
+        self._check_open()
+
+        # Create these objects now to attach them to the right loop.
+        # See #219
+        self._tasks = asyncio.Queue()
+        self._sched = AsyncScheduler()
+        # This has been most likely, but not necessarily, created in `open()`.
+        try:
+            self._lock
+        except AttributeError:
+            self._lock = asyncio.Lock()
+
+        self._closed = False
+        self._opened = True
+
+        self._start_workers()
+        self._start_initial_tasks()
+
+    def _start_workers(self) -> None:
+        self._sched_runner = create_task(
+            self._sched.run(), name=f"{self.name}-scheduler"
+        )
+        for i in range(self.num_workers):
+            t = create_task(
+                self.worker(self._tasks),
+                name=f"{self.name}-worker-{i}",
+            )
+            self._workers.append(t)
+
+    def _start_initial_tasks(self) -> None:
+        # populate the pool with initial min_size connections in background
+        for i in range(self._nconns):
+            self.run_task(AddConnection(self))
+
+        # Schedule a task to shrink the pool if connections over min_size have
+        # remained unused.
+        self.run_task(Schedule(self, ShrinkPool(self), self.max_idle))
+
     async def close(self, timeout: float = 5.0) -> None:
         if self._closed:
             return
@@ -217,33 +256,44 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
             # Take waiting client and pool connections out of the state
             waiting = list(self._waiting)
             self._waiting.clear()
-            pool = list(self._pool)
+            connections = list(self._pool)
             self._pool.clear()
 
         # Now that the flag _closed is set, getconn will fail immediately,
         # putconn will just close the returned connection.
+        await self._stop_workers(waiting, connections, timeout)
 
+    async def _stop_workers(
+        self,
+        waiting_clients: Sequence["AsyncClient"] = (),
+        connections: Sequence[AsyncConnection[Any]] = (),
+        timeout: float = 0.0,
+    ) -> None:
         # Stop the scheduler
         await self._sched.enter(0, None)
 
-        # Stop the worker threads
-        for w in self._workers:
+        # Stop the worker tasks
+        workers, self._workers = self._workers[:], []
+        for w in workers:
             self.run_task(StopWorker(self))
 
         # Signal to eventual clients in the queue that business is closed.
-        for pos in waiting:
+        for pos in waiting_clients:
             await pos.fail(PoolClosed(f"the pool {self.name!r} is closed"))
 
         # Close the connections still in the pool
-        for conn in pool:
+        for conn in connections:
             await conn.close()
 
-        # Wait for the worker threads to terminate
-        wait = asyncio.gather(self._sched_runner, *self._workers)
-        if timeout > 0:
-            wait = asyncio.wait_for(asyncio.shield(wait), timeout=timeout)
+        # Wait for the worker tasks to terminate
+        assert self._sched_runner is not None
+        sched_runner, self._sched_runner = self._sched_runner, None
+        wait = asyncio.gather(sched_runner, *workers)
         try:
-            await wait
+            if timeout > 0:
+                await asyncio.wait_for(asyncio.shield(wait), timeout=timeout)
+            else:
+                await wait
         except asyncio.TimeoutError:
             logger.warning(
                 "couldn't stop pool %r tasks within %s seconds",
@@ -252,6 +302,7 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
             )
 
     async def __aenter__(self) -> "AsyncConnectionPool":
+        await self.open()
         return self
 
     async def __aexit__(
@@ -262,13 +313,8 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
     ) -> None:
         await self.close()
 
-    async def resize(
-        self, min_size: int, max_size: Optional[int] = None
-    ) -> None:
-        if max_size is None:
-            max_size = min_size
-        if max_size < min_size:
-            raise ValueError("max_size must be greater or equal than min_size")
+    async def resize(self, min_size: int, max_size: Optional[int] = None) -> None:
+        min_size, max_size = self._check_size(min_size, max_size)
 
         ngrow = max(0, min_size - self._min_size)
 
@@ -291,10 +337,17 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
             conns = list(self._pool)
             self._pool.clear()
 
+            # Give a chance to the pool to grow if it has no connection.
+            # In case there are enough connection, or the pool is already
+            # growing, this is a no-op.
+            self._maybe_grow_pool()
+
         while conns:
             conn = conns.pop()
             try:
                 await conn.execute("SELECT 1")
+                if conn.pgconn.transaction_status == TransactionStatus.INTRANS:
+                    await conn.rollback()
             except Exception:
                 self._stats[self._CONNECTIONS_LOST] += 1
                 logger.warning("discarding broken connection: %s", conn)
@@ -309,20 +362,18 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
         self._reconnect_failed(self)
 
     def run_task(self, task: "MaintenanceTask") -> None:
-        """Run a maintenance task in a worker thread."""
+        """Run a maintenance task in a worker."""
         self._tasks.put_nowait(task)
 
-    async def schedule_task(
-        self, task: "MaintenanceTask", delay: float
-    ) -> None:
-        """Run a maintenance task in a worker thread in the future."""
+    async def schedule_task(self, task: "MaintenanceTask", delay: float) -> None:
+        """Run a maintenance task in a worker in the future."""
         await self._sched.enter(delay, task.tick)
 
     @classmethod
     async def worker(cls, q: "asyncio.Queue[MaintenanceTask]") -> None:
         """Runner to execute pending maintenance task.
 
-        The function is designed to run as a separate thread.
+        The function is designed to run as a task.
 
         Block on the queue *q*, run a task received. Finish running if a
         StopWorker is received.
@@ -345,15 +396,16 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
                     ex,
                 )
 
-    async def _connect(self) -> AsyncConnection[Any]:
-        """Return a new connection configured for the pool."""
+    async def _connect(self, timeout: Optional[float] = None) -> AsyncConnection[Any]:
         self._stats[self._CONNECTIONS_NUM] += 1
+        kwargs = self.kwargs
+        if timeout:
+            kwargs = kwargs.copy()
+            kwargs["connect_timeout"] = max(round(timeout), 1)
         t0 = monotonic()
         try:
             conn: AsyncConnection[Any]
-            conn = await self.connection_class.connect(
-                self.conninfo, **self.kwargs
-            )
+            conn = await self.connection_class.connect(self.conninfo, **kwargs)
         except Exception:
             self._stats[self._CONNECTIONS_ERRORS] += 1
             raise
@@ -367,16 +419,14 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
             await self._configure(conn)
             status = conn.pgconn.transaction_status
             if status != TransactionStatus.IDLE:
-                nstatus = TransactionStatus(status).name
+                sname = TransactionStatus(status).name
                 raise e.ProgrammingError(
-                    f"connection left in status {nstatus} by configure function"
+                    f"connection left in status {sname} by configure function"
                     f" {self._configure}: discarded"
                 )
 
         # Set an expiry date, with some randomness to avoid mass reconnection
-        conn._expire_at = monotonic() + self._jitter(
-            self.max_lifetime, -0.05, 0.0
-        )
+        self._set_connection_expiry_date(conn)
         return conn
 
     async def _add_connection(
@@ -391,9 +441,7 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
         """
         now = monotonic()
         if not attempt:
-            attempt = ConnectionAttempt(
-                reconnect_timeout=self.reconnect_timeout
-            )
+            attempt = ConnectionAttempt(reconnect_timeout=self.reconnect_timeout)
 
         try:
             conn = await self._connect()
@@ -407,6 +455,9 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
                 )
                 async with self._lock:
                     self._nconns -= 1
+                    # If we have given up with a growing attempt, allow a new one.
+                    if growing and self._growing:
+                        self._growing = False
                 self.reconnect_failed()
             else:
                 attempt.update_delay(now)
@@ -420,11 +471,13 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
         await self._add_to_pool(conn)
         if growing:
             async with self._lock:
-                if self._nconns < self._max_size and self._waiting:
+                # Keep on growing if the pool is not full yet, or if there are
+                # clients waiting and the pool can extend.
+                if self._nconns < self._min_size or (
+                    self._nconns < self._max_size and self._waiting
+                ):
                     self._nconns += 1
-                    logger.info(
-                        "growing pool %r to %s", self.name, self._nconns
-                    )
+                    logger.info("growing pool %r to %s", self.name, self._nconns)
                     self.run_task(AddConnection(self, growing=True))
                 else:
                     self._growing = False
@@ -464,8 +517,6 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
         # Also disable the warning for open connection in conn.__del__
         conn._pool = None
 
-        pos: Optional[AsyncClient] = None
-
         # Critical section: if there is a client waiting give it the connection
         # otherwise put it back into the pool.
         async with self._lock:
@@ -481,7 +532,7 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
 
                 # If we have been asked to wait for pool init, notify the
                 # waiter if the pool is full.
-                if self._pool_full_event and len(self._pool) >= self._nconns:
+                if self._pool_full_event and len(self._pool) >= self._min_size:
                     self._pool_full_event.set()
 
     async def _reset_connection(self, conn: AsyncConnection[Any]) -> None:
@@ -516,9 +567,9 @@ class AsyncConnectionPool(BasePool[AsyncConnection[Any]]):
                 await self._reset(conn)
                 status = conn.pgconn.transaction_status
                 if status != TransactionStatus.IDLE:
-                    nstatus = TransactionStatus(status).name
+                    sname = TransactionStatus(status).name
                     raise e.ProgrammingError(
-                        f"connection left in status {nstatus} by reset function"
+                        f"connection left in status {sname} by reset function"
                         f" {self._reset}: discarded"
                     )
             except Exception as ex:
@@ -635,12 +686,13 @@ class MaintenanceTask(ABC):
     async def run(self) -> None:
         """Run the task.
 
-        This usually happens in a worker thread. Call the concrete _run()
+        This usually happens in a worker. Call the concrete _run()
         implementation, if the pool is still alive.
         """
         pool = self.pool()
         if not pool or pool.closed:
             # Pool is no more working. Quietly discard the operation.
+            logger.debug("task run discarded: %s", self)
             return
 
         await self._run(pool)
@@ -648,12 +700,13 @@ class MaintenanceTask(ABC):
     async def tick(self) -> None:
         """Run the scheduled task
 
-        This function is called by the scheduler thread. Use a worker to
+        This function is called by the scheduler task. Use a worker to
         run the task for real in order to free the scheduler immediately.
         """
         pool = self.pool()
         if not pool or pool.closed:
             # Pool is no more working. Quietly discard the operation.
+            logger.debug("task tick discarded: %s", self)
             return
 
         pool.run_task(self)
@@ -664,7 +717,7 @@ class MaintenanceTask(ABC):
 
 
 class StopWorker(MaintenanceTask):
-    """Signal the maintenance thread to terminate."""
+    """Signal the maintenance worker to terminate."""
 
     async def _run(self, pool: "AsyncConnectionPool") -> None:
         pass
@@ -688,9 +741,7 @@ class AddConnection(MaintenanceTask):
 class ReturnConnection(MaintenanceTask):
     """Clean up and return a connection to the pool."""
 
-    def __init__(
-        self, pool: "AsyncConnectionPool", conn: "AsyncConnection[Any]"
-    ):
+    def __init__(self, pool: "AsyncConnectionPool", conn: "AsyncConnection[Any]"):
         super().__init__(pool)
         self.conn = conn
 
