@@ -905,6 +905,37 @@ async def test_reconnect_after_grow_failed(proxy):
         assert len(p._pool) == p.min_size == 4
 
 
+@pytest.mark.slow
+async def test_refill_on_check(proxy):
+    proxy.start()
+    ev = asyncio.Event()
+
+    def failed(pool):
+        ev.set()
+
+    async with pool.AsyncConnectionPool(
+        proxy.client_dsn, min_size=4, reconnect_timeout=1.0, reconnect_failed=failed
+    ) as p:
+        # The pool is full
+        await p.wait(timeout=2)
+
+        # Break all the connection
+        proxy.stop()
+
+        # Checking the pool will empty it
+        await p.check()
+        await asyncio.wait_for(ev.wait(), 2.0)
+        assert len(p._pool) == 0
+
+        # Allow to connect again
+        proxy.start()
+
+        # Make sure that check has refilled the pool
+        await p.check()
+        await p.wait(timeout=2)
+        assert len(p._pool) == 4
+
+
 @pytest.mark.slow
 async def test_uniform_use(dsn):
     async with pool.AsyncConnectionPool(dsn, min_size=4) as p:
