@@ -1,5 +1,6 @@
 import datetime as dt
 from types import ModuleType
+from typing import Any, List
 
 import pytest
 
@@ -9,6 +10,7 @@ from psycopg import errors as e
 from psycopg.adapt import Transformer, PyFormat, Dumper, Loader
 from psycopg._cmodule import _psycopg
 from psycopg.postgres import types as builtins, TEXT_OID
+from psycopg.types.array import ListDumper, ListBinaryDumper
 
 
 @pytest.mark.parametrize(
@@ -55,22 +57,21 @@ def test_register_dumper_by_class(conn):
 def test_register_dumper_by_class_name(conn):
     dumper = make_dumper("x")
     assert conn.adapters.get_dumper(MyStr, PyFormat.TEXT) is not dumper
-    conn.adapters.register_dumper(
-        f"{MyStr.__module__}.{MyStr.__qualname__}", dumper
-    )
+    conn.adapters.register_dumper(f"{MyStr.__module__}.{MyStr.__qualname__}", dumper)
     assert conn.adapters.get_dumper(MyStr, PyFormat.TEXT) is dumper
 
 
-def test_dump_global_ctx(dsn, global_adapters):
+@pytest.mark.crdb("skip", reason="global adapters don't affect crdb")
+def test_dump_global_ctx(conn_cls, dsn, global_adapters, pgconn):
     psycopg.adapters.register_dumper(MyStr, make_bin_dumper("gb"))
     psycopg.adapters.register_dumper(MyStr, make_dumper("gt"))
-    conn = psycopg.connect(dsn)
-    cur = conn.execute("select %s", [MyStr("hello")])
-    assert cur.fetchone() == ("hellogt",)
-    cur = conn.execute("select %b", [MyStr("hello")])
-    assert cur.fetchone() == ("hellogb",)
-    cur = conn.execute("select %t", [MyStr("hello")])
-    assert cur.fetchone() == ("hellogt",)
+    with conn_cls.connect(dsn) as conn:
+        cur = conn.execute("select %s", [MyStr("hello")])
+        assert cur.fetchone() == ("hellogt",)
+        cur = conn.execute("select %b", [MyStr("hello")])
+        assert cur.fetchone() == ("hellogb",)
+        cur = conn.execute("select %t", [MyStr("hello")])
+        assert cur.fetchone() == ("hellogt",)
 
 
 def test_dump_connection_ctx(conn):
@@ -115,9 +116,7 @@ def test_dump_subclass(conn):
         pass
 
     cur = conn.cursor()
-    cur.execute(
-        "select %s::text, %b::text", [MyString("hello"), MyString("world")]
-    )
+    cur.execute("select %s::text, %b::text", [MyString("hello"), MyString("world")])
     assert cur.fetchone() == ("hello", "world")
 
 
@@ -199,14 +198,15 @@ def test_register_loader_by_type_name(conn):
     assert conn.adapters.get_loader(TEXT_OID, pq.Format.TEXT) is loader
 
 
-def test_load_global_ctx(dsn, global_adapters):
+@pytest.mark.crdb("skip", reason="global adapters don't affect crdb")
+def test_load_global_ctx(conn_cls, dsn, global_adapters):
     psycopg.adapters.register_loader("text", make_loader("gt"))
     psycopg.adapters.register_loader("text", make_bin_loader("gb"))
-    conn = psycopg.connect(dsn)
-    cur = conn.cursor(binary=False).execute("select 'hello'::text")
-    assert cur.fetchone() == ("hellogt",)
-    cur = conn.cursor(binary=True).execute("select 'hello'::text")
-    assert cur.fetchone() == ("hellogb",)
+    with conn_cls.connect(dsn) as conn:
+        cur = conn.cursor(binary=False).execute("select 'hello'::text")
+        assert cur.fetchone() == ("hellogt",)
+        cur = conn.cursor(binary=True).execute("select 'hello'::text")
+        assert cur.fetchone() == ("hellogb",)
 
 
 def test_load_connection_ctx(conn):
@@ -289,25 +289,15 @@ def test_load_cursor_ctx_nested(conn, sql, obj, fmt_out):
 
 
 @pytest.mark.parametrize("fmt_out", pq.Format)
-def test_array_dumper(conn, fmt_out):
+def test_list_dumper(conn, fmt_out):
     t = Transformer(conn)
     fmt_in = PyFormat.from_pq(fmt_out)
     dint = t.get_dumper([0], fmt_in)
-    if fmt_out == pq.Format.BINARY:
-        assert dint.oid == builtins["int2"].array_oid
-        assert dint.sub_dumper.oid == builtins["int2"].oid
-    else:
-        assert dint.oid == builtins["numeric"].array_oid
-        assert dint.sub_dumper is None
+    assert isinstance(dint, (ListDumper, ListBinaryDumper))
+    assert dint.oid == builtins["int2"].array_oid
+    assert dint.sub_dumper and dint.sub_dumper.oid == builtins["int2"].oid
 
     dstr = t.get_dumper([""], fmt_in)
-    if fmt_in == PyFormat.BINARY:
-        assert dstr.oid == builtins["text"].array_oid
-        assert dstr.sub_dumper.oid == builtins["text"].oid
-    else:
-        assert dstr.oid == 0
-        assert dstr.sub_dumper.oid == 0
-
     assert dstr is not dint
 
     assert t.get_dumper([1], fmt_in) is dint
@@ -318,12 +308,29 @@ def test_array_dumper(conn, fmt_out):
     assert dempty.oid == 0
     assert dempty.dump([]) == b"{}"
 
-    L = []
+    L: List[List[Any]] = []
     L.append(L)
     with pytest.raises(psycopg.DataError):
         assert t.get_dumper(L, fmt_in)
 
 
+@pytest.mark.crdb("skip", reason="test in crdb test suite")
+def test_str_list_dumper_text(conn):
+    t = Transformer(conn)
+    dstr = t.get_dumper([""], PyFormat.TEXT)
+    assert isinstance(dstr, ListDumper)
+    assert dstr.oid == 0
+    assert dstr.sub_dumper and dstr.sub_dumper.oid == 0
+
+
+def test_str_list_dumper_binary(conn):
+    t = Transformer(conn)
+    dstr = t.get_dumper([""], PyFormat.BINARY)
+    assert isinstance(dstr, ListBinaryDumper)
+    assert dstr.oid == builtins["text"].array_oid
+    assert dstr.sub_dumper and dstr.sub_dumper.oid == builtins["text"].oid
+
+
 def test_last_dumper_registered_ctx(conn):
     cur = conn.cursor()
 
@@ -344,12 +351,11 @@ def test_last_dumper_registered_ctx(conn):
 def test_none_type_argument(conn, fmt_in):
     cur = conn.cursor()
     cur.execute("create table none_args (id serial primary key, num integer)")
-    cur.execute(
-        "insert into none_args (num) values (%s) returning id", (None,)
-    )
+    cur.execute("insert into none_args (num) values (%s) returning id", (None,))
     assert cur.fetchone()[0]
 
 
+@pytest.mark.crdb("skip", reason="test in crdb test suite")
 @pytest.mark.parametrize("fmt_in", PyFormat)
 def test_return_untyped(conn, fmt_in):
     # Analyze and check for changes using strings in untyped/typed contexts
@@ -357,31 +363,30 @@ def test_return_untyped(conn, fmt_in):
     # Currently string are passed as unknown oid to libpq. This is because
     # unknown is more easily cast by postgres to different types (see jsonb
     # later).
-    cur.execute(f"select %{fmt_in}, %{fmt_in}", ["hello", 10])
+    cur.execute(f"select %{fmt_in.value}, %{fmt_in.value}", ["hello", 10])
     assert cur.fetchone() == ("hello", 10)
 
     cur.execute("create table testjson(data jsonb)")
     if fmt_in != PyFormat.BINARY:
-        cur.execute(f"insert into testjson (data) values (%{fmt_in})", ["{}"])
+        cur.execute(f"insert into testjson (data) values (%{fmt_in.value})", ["{}"])
         assert cur.execute("select data from testjson").fetchone() == ({},)
     else:
         # Binary types cannot be passed as unknown oids.
         with pytest.raises(e.DatatypeMismatch):
-            cur.execute(
-                f"insert into testjson (data) values (%{fmt_in})", ["{}"]
-            )
+            cur.execute(f"insert into testjson (data) values (%{fmt_in.value})", ["{}"])
 
 
-@pytest.mark.parametrize("fmt_in", PyFormat.AUTO)
+@pytest.mark.parametrize("fmt_in", PyFormat)
 def test_no_cast_needed(conn, fmt_in):
     # Verify that there is no need of cast in certain common scenario
-    cur = conn.execute("select '2021-01-01'::date + %s", [3])
+    cur = conn.execute(f"select '2021-01-01'::date + %{fmt_in.value}", [3])
     assert cur.fetchone()[0] == dt.date(2021, 1, 4)
 
-    cur = conn.execute("select '[10, 20, 30]'::jsonb -> %s", [1])
+    cur = conn.execute(f"select '[10, 20, 30]'::jsonb -> %{fmt_in.value}", [1])
     assert cur.fetchone()[0] == 20
 
 
+@pytest.mark.slow
 @pytest.mark.skipif(_psycopg is None, reason="C module test")
 def test_optimised_adapters():
 
@@ -393,27 +398,28 @@ def test_optimised_adapters():
         obj = getattr(_psycopg, n)
         if not isinstance(obj, type):
             continue
-        if not issubclass(obj, (_psycopg.CDumper, _psycopg.CLoader)):
+        if not issubclass(
+            obj,
+            (_psycopg.CDumper, _psycopg.CLoader),  # type: ignore[attr-defined]
+        ):
             continue
         c_adapters[n] = obj
 
     # All the registered adapters
     reg_adapters = set()
-    adapters = (
-        list(postgres.adapters._dumpers.values()) + postgres.adapters._loaders
-    )
+    adapters = list(postgres.adapters._dumpers.values()) + postgres.adapters._loaders
     assert len(adapters) == 5
     for m in adapters:
         reg_adapters |= set(m.values())
 
     # Check that the registered adapters are the optimised one
-    n = 0
+    i = 0
     for cls in reg_adapters:
         if cls.__name__ in c_adapters:
             assert cls is c_adapters[cls.__name__]
-            n += 1
+            i += 1
 
-    assert n >= 10
+    assert i >= 10
 
     # Check that every optimised adapter is the optimised version of a Py one
     for n in dir(psycopg.types):
@@ -431,6 +437,37 @@ def test_optimised_adapters():
     assert not c_adapters
 
 
+def test_dumper_init_error(conn):
+    class BadDumper(Dumper):
+        def __init__(self, cls, context):
+            super().__init__(cls, context)
+            1 / 0
+
+        def dump(self, obj):
+            return obj.encode()
+
+    cur = conn.cursor()
+    cur.adapters.register_dumper(str, BadDumper)
+    with pytest.raises(ZeroDivisionError):
+        cur.execute("select %s::text", ["hi"])
+
+
+def test_loader_init_error(conn):
+    class BadLoader(Loader):
+        def __init__(self, oid, context):
+            super().__init__(oid, context)
+            1 / 0
+
+        def load(self, data):
+            return data.decode()
+
+    cur = conn.cursor()
+    cur.adapters.register_loader("text", BadLoader)
+    with pytest.raises(ZeroDivisionError):
+        cur.execute("select 'hi'::text")
+        assert cur.fetchone() == ("hi",)
+
+
 @pytest.mark.slow
 @pytest.mark.parametrize("fmt", PyFormat)
 @pytest.mark.parametrize("fmt_out", [pq.Format.TEXT, pq.Format.BINARY])
