@@ -1,15 +1,13 @@
-import sys
 import time
-import socket
 import pytest
 import logging
 import weakref
-from threading import Thread
+from typing import Any, List
+from dataclasses import dataclass
 
 import psycopg
-from psycopg import Connection, Notify
+from psycopg import Notify, errors as e
 from psycopg.rows import tuple_row
-from psycopg.errors import UndefinedTable
 from psycopg.conninfo import conninfo_to_dict, make_conninfo
 
 from .utils import gc_collect
@@ -17,44 +15,34 @@ from .test_cursor import my_row_factory
 from .test_adapt import make_bin_dumper, make_dumper
 
 
-def test_connect(dsn):
-    conn = Connection.connect(dsn)
+def test_connect(conn_cls, dsn):
+    conn = conn_cls.connect(dsn)
     assert not conn.closed
     assert conn.pgconn.status == conn.ConnStatus.OK
+    conn.close()
 
 
-def test_connect_str_subclass(dsn):
+def test_connect_str_subclass(conn_cls, dsn):
     class MyString(str):
         pass
 
-    conn = Connection.connect(MyString(dsn))
+    conn = conn_cls.connect(MyString(dsn))
     assert not conn.closed
     assert conn.pgconn.status == conn.ConnStatus.OK
+    conn.close()
 
 
-def test_connect_bad():
+def test_connect_bad(conn_cls):
     with pytest.raises(psycopg.OperationalError):
-        Connection.connect("dbname=nosuchdb")
+        conn_cls.connect("dbname=nosuchdb")
 
 
 @pytest.mark.slow
 @pytest.mark.timing
-@pytest.mark.skipif(sys.platform == "win32", reason="connect() hangs on Win32")
-def test_connect_timeout():
-    s = socket.socket(socket.AF_INET)
-    s.bind(("localhost", 0))
-    port = s.getsockname()[1]
-    s.listen(0)
-
-    def closer():
-        time.sleep(1.5)
-        s.close()
-
-    Thread(target=closer).start()
-
+def test_connect_timeout(conn_cls, deaf_port):
     t0 = time.time()
     with pytest.raises(psycopg.OperationalError, match="timeout expired"):
-        Connection.connect(host="localhost", port=port, connect_timeout=1)
+        conn_cls.connect(host="localhost", port=deaf_port, connect_timeout=1)
     elapsed = time.time() - t0
     assert elapsed == pytest.approx(1.0, abs=0.05)
 
@@ -62,13 +50,14 @@ def test_connect_timeout():
 def test_close(conn):
     assert not conn.closed
     assert not conn.broken
+
+    cur = conn.cursor()
+
     conn.close()
     assert conn.closed
     assert not conn.broken
     assert conn.pgconn.status == conn.ConnStatus.BAD
 
-    cur = conn.cursor()
-
     conn.close()
     assert conn.closed
     assert conn.pgconn.status == conn.ConnStatus.BAD
@@ -77,11 +66,10 @@ def test_close(conn):
         cur.execute("select 1")
 
 
+@pytest.mark.crdb_skip("pg_terminate_backend")
 def test_broken(conn):
     with pytest.raises(psycopg.OperationalError):
-        conn.execute(
-            "select pg_terminate_backend(%s)", [conn.pgconn.backend_pid]
-        )
+        conn.execute("select pg_terminate_backend(%s)", [conn.pgconn.backend_pid])
     assert conn.closed
     assert conn.broken
     conn.close()
@@ -89,22 +77,31 @@ def test_broken(conn):
     assert conn.broken
 
 
-def test_connection_warn_close(dsn, recwarn):
-    conn = Connection.connect(dsn)
+def test_cursor_closed(conn):
+    conn.close()
+    with pytest.raises(psycopg.OperationalError):
+        with conn.cursor("foo"):
+            pass
+    with pytest.raises(psycopg.OperationalError):
+        conn.cursor()
+
+
+def test_connection_warn_close(conn_cls, dsn, recwarn):
+    conn = conn_cls.connect(dsn)
     conn.close()
     del conn
     assert not recwarn, [str(w.message) for w in recwarn.list]
 
-    conn = Connection.connect(dsn)
+    conn = conn_cls.connect(dsn)
     del conn
     assert "IDLE" in str(recwarn.pop(ResourceWarning).message)
 
-    conn = Connection.connect(dsn)
+    conn = conn_cls.connect(dsn)
     conn.execute("select 1")
     del conn
     assert "INTRANS" in str(recwarn.pop(ResourceWarning).message)
 
-    conn = Connection.connect(dsn)
+    conn = conn_cls.connect(dsn)
     try:
         conn.execute("select wat")
     except Exception:
@@ -112,45 +109,47 @@ def test_connection_warn_close(dsn, recwarn):
     del conn
     assert "INERROR" in str(recwarn.pop(ResourceWarning).message)
 
-    with Connection.connect(dsn) as conn:
+    with conn_cls.connect(dsn) as conn:
         pass
     del conn
     assert not recwarn, [str(w.message) for w in recwarn.list]
 
 
-def test_context_commit(conn, dsn):
+@pytest.fixture
+def testctx(svcconn):
+    svcconn.execute("create table if not exists testctx (id int primary key)")
+    svcconn.execute("delete from testctx")
+    return None
+
+
+def test_context_commit(conn_cls, testctx, conn, dsn):
     with conn:
         with conn.cursor() as cur:
-            cur.execute("drop table if exists textctx")
-            cur.execute("create table textctx ()")
+            cur.execute("insert into testctx values (42)")
 
     assert conn.closed
     assert not conn.broken
 
-    with psycopg.connect(dsn) as conn:
+    with conn_cls.connect(dsn) as conn:
         with conn.cursor() as cur:
-            cur.execute("select * from textctx")
-            assert cur.fetchall() == []
+            cur.execute("select * from testctx")
+            assert cur.fetchall() == [(42,)]
 
 
-def test_context_rollback(conn, dsn):
-    with conn.cursor() as cur:
-        cur.execute("drop table if exists textctx")
-    conn.commit()
-
+def test_context_rollback(conn_cls, testctx, conn, dsn):
     with pytest.raises(ZeroDivisionError):
         with conn:
             with conn.cursor() as cur:
-                cur.execute("create table textctx ()")
+                cur.execute("insert into testctx values (42)")
                 1 / 0
 
     assert conn.closed
     assert not conn.broken
 
-    with psycopg.connect(dsn) as conn:
+    with conn_cls.connect(dsn) as conn:
         with conn.cursor() as cur:
-            with pytest.raises(UndefinedTable):
-                cur.execute("select * from textctx")
+            cur.execute("select * from testctx")
+            assert cur.fetchall() == []
 
 
 def test_context_close(conn):
@@ -159,11 +158,12 @@ def test_context_close(conn):
         conn.close()
 
 
-def test_context_rollback_no_clobber(conn, dsn, caplog):
+@pytest.mark.crdb_skip("pg_terminate_backend")
+def test_context_inerror_rollback_no_clobber(conn_cls, conn, dsn, caplog):
     caplog.set_level(logging.WARNING, logger="psycopg")
 
     with pytest.raises(ZeroDivisionError):
-        with psycopg.connect(dsn) as conn2:
+        with conn_cls.connect(dsn) as conn2:
             conn2.execute("select 1")
             conn.execute(
                 "select pg_terminate_backend(%s::int)",
@@ -174,11 +174,30 @@ def test_context_rollback_no_clobber(conn, dsn, caplog):
     assert len(caplog.records) == 1
     rec = caplog.records[0]
     assert rec.levelno == logging.WARNING
-    assert "rolling back" in rec.message
+    assert "in rollback" in rec.message
 
 
-def test_weakref(dsn):
-    conn = psycopg.connect(dsn)
+@pytest.mark.crdb_skip("copy")
+def test_context_active_rollback_no_clobber(conn_cls, dsn, caplog):
+    caplog.set_level(logging.WARNING, logger="psycopg")
+
+    with pytest.raises(ZeroDivisionError):
+        with conn_cls.connect(dsn) as conn:
+            conn.pgconn.exec_(b"copy (select generate_series(1, 10)) to stdout")
+            assert not conn.pgconn.error_message
+            status = conn.info.transaction_status
+            assert status == conn.TransactionStatus.ACTIVE
+            1 / 0
+
+    assert len(caplog.records) == 1
+    rec = caplog.records[0]
+    assert rec.levelno == logging.WARNING
+    assert "in rollback" in rec.message
+
+
+@pytest.mark.slow
+def test_weakref(conn_cls, dsn):
+    conn = conn_cls.connect(dsn)
     w = weakref.ref(conn)
     conn.close()
     del conn
@@ -202,6 +221,26 @@ def test_commit(conn):
         conn.commit()
 
 
+@pytest.mark.crdb_skip("deferrable")
+def test_commit_error(conn):
+    conn.execute(
+        """
+        drop table if exists selfref;
+        create table selfref (
+            x serial primary key,
+            y int references selfref (x) deferrable initially deferred)
+        """
+    )
+    conn.commit()
+
+    conn.execute("insert into selfref (y) values (-1)")
+    with pytest.raises(e.ForeignKeyViolation):
+        conn.commit()
+    assert conn.pgconn.transaction_status == conn.TransactionStatus.IDLE
+    cur = conn.execute("select 1")
+    assert cur.fetchone() == (1,)
+
+
 def test_rollback(conn):
     conn.pgconn.exec_(b"drop table if exists foo")
     conn.pgconn.exec_(b"create table foo (id int primary key)")
@@ -266,14 +305,15 @@ def test_autocommit(conn):
     assert conn.pgconn.transaction_status == conn.TransactionStatus.IDLE
 
     conn.autocommit = ""
-    assert conn.autocommit is False
+    assert conn.autocommit is False  # type: ignore[comparison-overlap]
     conn.autocommit = "yeah"
     assert conn.autocommit is True
 
 
-def test_autocommit_connect(dsn):
-    conn = Connection.connect(dsn, autocommit=True)
+def test_autocommit_connect(conn_cls, dsn):
+    conn = conn_cls.connect(dsn, autocommit=True)
     assert conn.autocommit
+    conn.close()
 
 
 def test_autocommit_intrans(conn):
@@ -298,7 +338,7 @@ def test_autocommit_inerror(conn):
 def test_autocommit_unknown(conn):
     conn.close()
     assert conn.pgconn.transaction_status == conn.TransactionStatus.UNKNOWN
-    with pytest.raises(psycopg.ProgrammingError):
+    with pytest.raises(psycopg.OperationalError):
         conn.autocommit = True
     assert not conn.autocommit
 
@@ -318,8 +358,8 @@ def test_autocommit_unknown(conn):
         (("host=foo",), {"user": None}, "host=foo"),
     ],
 )
-def test_connect_args(monkeypatch, pgconn, args, kwargs, want):
-    the_conninfo = None
+def test_connect_args(conn_cls, monkeypatch, pgconn, args, kwargs, want):
+    the_conninfo: str
 
     def fake_connect(conninfo):
         nonlocal the_conninfo
@@ -328,28 +368,30 @@ def test_connect_args(monkeypatch, pgconn, args, kwargs, want):
         yield
 
     monkeypatch.setattr(psycopg.connection, "connect", fake_connect)
-    psycopg.Connection.connect(*args, **kwargs)
+    conn = conn_cls.connect(*args, **kwargs)
     assert conninfo_to_dict(the_conninfo) == conninfo_to_dict(want)
+    conn.close()
 
 
 @pytest.mark.parametrize(
-    "args, kwargs",
+    "args, kwargs, exctype",
     [
-        (("host=foo", "host=bar"), {}),
-        (("", ""), {}),
-        ((), {"nosuchparam": 42}),
+        (("host=foo", "host=bar"), {}, TypeError),
+        (("", ""), {}, TypeError),
+        ((), {"nosuchparam": 42}, psycopg.ProgrammingError),
     ],
 )
-def test_connect_badargs(monkeypatch, pgconn, args, kwargs):
+def test_connect_badargs(conn_cls, monkeypatch, pgconn, args, kwargs, exctype):
     def fake_connect(conninfo):
         return pgconn
         yield
 
     monkeypatch.setattr(psycopg.connection, "connect", fake_connect)
-    with pytest.raises((TypeError, psycopg.ProgrammingError)):
-        psycopg.Connection.connect(*args, **kwargs)
+    with pytest.raises(exctype):
+        conn_cls.connect(*args, **kwargs)
 
 
+@pytest.mark.crdb_skip("pg_terminate_backend")
 def test_broken_connection(conn):
     cur = conn.cursor()
     with pytest.raises(psycopg.DatabaseError):
@@ -357,6 +399,7 @@ def test_broken_connection(conn):
     assert conn.closed
 
 
+@pytest.mark.crdb_skip("do")
 def test_notice_handlers(conn, caplog):
     caplog.set_level(logging.WARNING, logger="psycopg")
     messages = []
@@ -371,13 +414,11 @@ def test_notice_handlers(conn, caplog):
     conn.add_notice_handler(cb1)
     conn.add_notice_handler(cb2)
     conn.add_notice_handler("the wrong thing")
-    conn.add_notice_handler(lambda diag: severities.append(diag.severity))
+    conn.add_notice_handler(lambda diag: severities.append(diag.severity_nonlocalized))
 
     conn.pgconn.exec_(b"set client_min_messages to notice")
     cur = conn.cursor()
-    cur.execute(
-        "do $$begin raise notice 'hello notice'; end$$ language plpgsql"
-    )
+    cur.execute("do $$begin raise notice 'hello notice'; end$$ language plpgsql")
     assert messages == ["hello notice"]
     assert severities == ["NOTICE"]
 
@@ -391,9 +432,7 @@ def test_notice_handlers(conn, caplog):
 
     conn.remove_notice_handler(cb1)
     conn.remove_notice_handler("the wrong thing")
-    cur.execute(
-        "do $$begin raise warning 'hello warning'; end$$ language plpgsql"
-    )
+    cur.execute("do $$begin raise warning 'hello warning'; end$$ language plpgsql")
     assert len(caplog.records) == 3
     assert messages == ["hello notice"]
     assert severities == ["NOTICE", "WARNING"]
@@ -402,6 +441,7 @@ def test_notice_handlers(conn, caplog):
         conn.remove_notice_handler(cb1)
 
 
+@pytest.mark.crdb_skip("notify")
 def test_notify_handlers(conn):
     nots1 = []
     nots2 = []
@@ -462,27 +502,31 @@ def test_execute_binary(conn):
     assert cur.pgresult.fformat(0) == 1
 
 
-def test_row_factory(dsn):
-    conn = Connection.connect(dsn)
-    assert conn.row_factory is tuple_row
+def test_row_factory(conn_cls, dsn):
+    defaultconn = conn_cls.connect(dsn)
+    assert defaultconn.row_factory is tuple_row
+    defaultconn.close()
 
-    conn = Connection.connect(dsn, row_factory=my_row_factory)
+    conn = conn_cls.connect(dsn, row_factory=my_row_factory)
     assert conn.row_factory is my_row_factory
 
     cur = conn.execute("select 'a' as ve")
     assert cur.fetchone() == ["Ave"]
 
-    with conn.cursor(row_factory=lambda c: set) as cur:
-        cur.execute("select 1, 1, 2")
-        assert cur.fetchall() == [{1, 2}]
+    with conn.cursor(row_factory=lambda c: lambda t: set(t)) as cur1:
+        cur1.execute("select 1, 1, 2")
+        assert cur1.fetchall() == [{1, 2}]
 
-    with conn.cursor(row_factory=tuple_row) as cur:
-        cur.execute("select 1, 1, 2")
-        assert cur.fetchall() == [(1, 1, 2)]
+    with conn.cursor(row_factory=tuple_row) as cur2:
+        cur2.execute("select 1, 1, 2")
+        assert cur2.fetchall() == [(1, 1, 2)]
 
+    # TODO: maybe fix something to get rid of 'type: ignore' below.
     conn.row_factory = tuple_row
-    cur = conn.execute("select 'vale'")
-    assert cur.fetchone() == ("vale",)
+    cur3 = conn.execute("select 'vale'")
+    r = cur3.fetchone()
+    assert r and r == ("vale",)
+    conn.close()
 
 
 def test_str(conn):
@@ -501,7 +545,7 @@ def test_fileno(conn):
 def test_cursor_factory(conn):
     assert conn.cursor_factory is psycopg.Cursor
 
-    class MyCursor(psycopg.Cursor):
+    class MyCursor(psycopg.Cursor[psycopg.rows.Row]):
         pass
 
     conn.cursor_factory = MyCursor
@@ -512,10 +556,20 @@ def test_cursor_factory(conn):
         assert isinstance(cur, MyCursor)
 
 
+def test_cursor_factory_connect(conn_cls, dsn):
+    class MyCursor(psycopg.Cursor[psycopg.rows.Row]):
+        pass
+
+    with conn_cls.connect(dsn, cursor_factory=MyCursor) as conn:
+        assert conn.cursor_factory is MyCursor
+        cur = conn.cursor()
+        assert type(cur) is MyCursor
+
+
 def test_server_cursor_factory(conn):
     assert conn.server_cursor_factory is psycopg.ServerCursor
 
-    class MyServerCursor(psycopg.ServerCursor):
+    class MyServerCursor(psycopg.ServerCursor[psycopg.rows.Row]):
         pass
 
     conn.server_cursor_factory = MyServerCursor
@@ -523,23 +577,28 @@ def test_server_cursor_factory(conn):
         assert isinstance(cur, MyServerCursor)
 
 
-tx_params = {
-    "isolation_level": {
-        "guc": "isolation",
-        "values": list(psycopg.IsolationLevel),
-        "set_method": "set_isolation_level",
-    },
-    "read_only": {
-        "guc": "read_only",
-        "values": [True, False],
-        "set_method": "set_read_only",
-    },
-    "deferrable": {
-        "guc": "deferrable",
-        "values": [True, False],
-        "set_method": "set_deferrable",
-    },
-}
+@dataclass
+class ParamDef:
+    name: str
+    guc: str
+    values: List[Any]
+
+
+param_isolation = ParamDef(
+    name="isolation_level",
+    guc="isolation",
+    values=list(psycopg.IsolationLevel),
+)
+param_read_only = ParamDef(
+    name="read_only",
+    guc="read_only",
+    values=[True, False],
+)
+param_deferrable = ParamDef(
+    name="deferrable",
+    guc="deferrable",
+    values=[True, False],
+)
 
 # Map Python values to Postgres values for the tx_params possible values
 tx_values_map = {
@@ -549,27 +608,45 @@ tx_values_map["on"] = True
 tx_values_map["off"] = False
 
 
-@pytest.mark.parametrize("attr", list(tx_params))
-def test_transaction_param_default(conn, attr):
-    assert getattr(conn, attr) is None
-    guc = tx_params[attr]["guc"]
+tx_params = [
+    param_isolation,
+    param_read_only,
+    pytest.param(param_deferrable, marks=pytest.mark.crdb_skip("deferrable")),
+]
+tx_params_isolation = [
+    pytest.param(
+        param_isolation,
+        id="isolation_level",
+        marks=pytest.mark.crdb("skip", reason="transaction isolation"),
+    ),
+    pytest.param(
+        param_read_only, id="read_only", marks=pytest.mark.crdb_skip("begin_read_only")
+    ),
+    pytest.param(
+        param_deferrable, id="deferrable", marks=pytest.mark.crdb_skip("deferrable")
+    ),
+]
+
+
+@pytest.mark.parametrize("param", tx_params)
+def test_transaction_param_default(conn, param):
+    assert getattr(conn, param.name) is None
     current, default = conn.execute(
         "select current_setting(%s), current_setting(%s)",
-        [f"transaction_{guc}", f"default_transaction_{guc}"],
+        [f"transaction_{param.guc}", f"default_transaction_{param.guc}"],
     ).fetchone()
     assert current == default
 
 
 @pytest.mark.parametrize("autocommit", [True, False])
-@pytest.mark.parametrize("attr", list(tx_params))
-def test_set_transaction_param_implicit(conn, attr, autocommit):
-    guc = tx_params[attr]["guc"]
+@pytest.mark.parametrize("param", tx_params_isolation)
+def test_set_transaction_param_implicit(conn, param, autocommit):
     conn.autocommit = autocommit
-    for value in tx_params[attr]["values"]:
-        setattr(conn, attr, value)
+    for value in param.values:
+        setattr(conn, param.name, value)
         pgval, default = conn.execute(
             "select current_setting(%s), current_setting(%s)",
-            [f"transaction_{guc}", f"default_transaction_{guc}"],
+            [f"transaction_{param.guc}", f"default_transaction_{param.guc}"],
         ).fetchone()
         if autocommit:
             assert pgval == default
@@ -579,50 +656,52 @@ def test_set_transaction_param_implicit(conn, attr, autocommit):
 
 
 @pytest.mark.parametrize("autocommit", [True, False])
-@pytest.mark.parametrize("attr", list(tx_params))
-def test_set_transaction_param_block(conn, attr, autocommit):
-    guc = tx_params[attr]["guc"]
+@pytest.mark.parametrize("param", tx_params_isolation)
+def test_set_transaction_param_block(conn, param, autocommit):
     conn.autocommit = autocommit
-    for value in tx_params[attr]["values"]:
-        setattr(conn, attr, value)
+    for value in param.values:
+        setattr(conn, param.name, value)
         with conn.transaction():
             pgval = conn.execute(
-                "select current_setting(%s)", [f"transaction_{guc}"]
+                "select current_setting(%s)", [f"transaction_{param.guc}"]
             ).fetchone()[0]
         assert tx_values_map[pgval] == value
 
 
-@pytest.mark.parametrize("attr", list(tx_params))
-def test_set_transaction_param_not_intrans_implicit(conn, attr):
+@pytest.mark.parametrize("param", tx_params)
+def test_set_transaction_param_not_intrans_implicit(conn, param):
     conn.execute("select 1")
     with pytest.raises(psycopg.ProgrammingError):
-        setattr(conn, attr, tx_params[attr]["values"][0])
+        setattr(conn, param.name, param.values[0])
 
 
-@pytest.mark.parametrize("attr", list(tx_params))
-def test_set_transaction_param_not_intrans_block(conn, attr):
+@pytest.mark.parametrize("param", tx_params)
+def test_set_transaction_param_not_intrans_block(conn, param):
     with conn.transaction():
         with pytest.raises(psycopg.ProgrammingError):
-            setattr(conn, attr, tx_params[attr]["values"][0])
+            setattr(conn, param.name, param.values[0])
 
 
-@pytest.mark.parametrize("attr", list(tx_params))
-def test_set_transaction_param_not_intrans_external(conn, attr):
+@pytest.mark.parametrize("param", tx_params)
+def test_set_transaction_param_not_intrans_external(conn, param):
     conn.autocommit = True
     conn.execute("begin")
     with pytest.raises(psycopg.ProgrammingError):
-        setattr(conn, attr, tx_params[attr]["values"][0])
+        setattr(conn, param.name, param.values[0])
 
 
+@pytest.mark.crdb("skip", reason="transaction isolation")
 def test_set_transaction_param_all(conn):
-    for attr in tx_params:
-        value = tx_params[attr]["values"][0]
-        setattr(conn, attr, value)
+    params: List[Any] = tx_params[:]
+    params[2] = params[2].values[0]
 
-    for attr in tx_params:
-        guc = tx_params[attr]["guc"]
+    for param in params:
+        value = param.values[0]
+        setattr(conn, param.name, value)
+
+    for param in params:
         pgval = conn.execute(
-            "select current_setting(%s)", [f"transaction_{guc}"]
+            "select current_setting(%s)", [f"transaction_{param.guc}"]
         ).fetchone()[0]
         assert tx_values_map[pgval] == value
 
@@ -645,13 +724,13 @@ def test_set_transaction_param_strange(conn):
 conninfo_params_timeout = [
     (
         "",
-        {"host": "localhost", "connect_timeout": None},
-        ({"host": "localhost"}, None),
+        {"dbname": "mydb", "connect_timeout": None},
+        ({"dbname": "mydb"}, None),
     ),
     (
         "",
-        {"host": "localhost", "connect_timeout": 1},
-        ({"host": "localhost", "connect_timeout": "1"}, 1),
+        {"dbname": "mydb", "connect_timeout": 1},
+        ({"dbname": "mydb", "connect_timeout": "1"}, 1),
     ),
     (
         "dbname=postgres",
@@ -672,33 +751,40 @@ conninfo_params_timeout = [
 
 
 @pytest.mark.parametrize("dsn, kwargs, exp", conninfo_params_timeout)
-def test_get_connection_params(dsn, kwargs, exp):
-    params = Connection._get_connection_params(dsn, **kwargs)
+def test_get_connection_params(conn_cls, dsn, kwargs, exp):
+    params = conn_cls._get_connection_params(dsn, **kwargs)
     conninfo = make_conninfo(**params)
     assert conninfo_to_dict(conninfo) == exp[0]
     assert params.get("connect_timeout") == exp[1]
 
 
-def test_connect_context(dsn):
+def test_connect_context(conn_cls, dsn):
     ctx = psycopg.adapt.AdaptersMap(psycopg.adapters)
     ctx.register_dumper(str, make_bin_dumper("b"))
     ctx.register_dumper(str, make_dumper("t"))
 
-    conn = psycopg.connect(dsn, context=ctx)
+    conn = conn_cls.connect(dsn, context=ctx)
 
     cur = conn.execute("select %s", ["hello"])
     assert cur.fetchone()[0] == "hellot"
     cur = conn.execute("select %b", ["hello"])
     assert cur.fetchone()[0] == "hellob"
+    conn.close()
 
 
-def test_connect_context_copy(dsn, conn):
+def test_connect_context_copy(conn_cls, dsn, conn):
     conn.adapters.register_dumper(str, make_bin_dumper("b"))
     conn.adapters.register_dumper(str, make_dumper("t"))
 
-    conn2 = psycopg.connect(dsn, context=conn)
+    conn2 = conn_cls.connect(dsn, context=conn)
 
     cur = conn2.execute("select %s", ["hello"])
     assert cur.fetchone()[0] == "hellot"
     cur = conn2.execute("select %b", ["hello"])
     assert cur.fetchone()[0] == "hellob"
+    conn2.close()
+
+
+def test_cancel_closed(conn):
+    conn.close()
+    conn.cancel()
