@@ -1,81 +1,68 @@
 import time
-import socket
 import pytest
-import asyncio
 import logging
-import sys
 import weakref
+from typing import List, Any
 
 import psycopg
-from psycopg import AsyncConnection, Notify
+from psycopg import Notify, errors as e
 from psycopg.rows import tuple_row
-from psycopg.errors import InterfaceError, UndefinedTable
 from psycopg.conninfo import conninfo_to_dict, make_conninfo
 
 from .utils import gc_collect
 from .test_cursor import my_row_factory
-from .test_connection import tx_params, tx_values_map, conninfo_params_timeout
+from .test_connection import tx_params, tx_params_isolation, tx_values_map
+from .test_connection import conninfo_params_timeout
+from .test_connection import testctx  # noqa: F401  # fixture
 from .test_adapt import make_bin_dumper, make_dumper
+from .test_conninfo import fake_resolve  # noqa: F401
 
 pytestmark = pytest.mark.asyncio
 
 
-async def test_connect(dsn):
-    conn = await AsyncConnection.connect(dsn)
+async def test_connect(aconn_cls, dsn):
+    conn = await aconn_cls.connect(dsn)
     assert not conn.closed
     assert conn.pgconn.status == conn.ConnStatus.OK
+    await conn.close()
 
 
-async def test_connect_bad():
+async def test_connect_bad(aconn_cls):
     with pytest.raises(psycopg.OperationalError):
-        await AsyncConnection.connect("dbname=nosuchdb")
+        await aconn_cls.connect("dbname=nosuchdb")
 
 
-async def test_connect_str_subclass(dsn):
+async def test_connect_str_subclass(aconn_cls, dsn):
     class MyString(str):
         pass
 
-    conn = await AsyncConnection.connect(MyString(dsn))
+    conn = await aconn_cls.connect(MyString(dsn))
     assert not conn.closed
     assert conn.pgconn.status == conn.ConnStatus.OK
+    await conn.close()
 
 
 @pytest.mark.slow
 @pytest.mark.timing
-async def test_connect_timeout():
-    s = socket.socket(socket.AF_INET)
-    s.bind(("", 0))
-    port = s.getsockname()[1]
-    s.listen(0)
-
-    async def closer():
-        await asyncio.sleep(1.5)
-        s.close()
-
-    async def connect():
-        t0 = time.time()
-        with pytest.raises(psycopg.OperationalError, match="timeout expired"):
-            await AsyncConnection.connect(
-                host="localhost", port=port, connect_timeout=1
-            )
-        nonlocal elapsed
-        elapsed = time.time() - t0
-
-    elapsed = 0
-    await asyncio.gather(closer(), connect())
+async def test_connect_timeout(aconn_cls, deaf_port):
+    t0 = time.time()
+    with pytest.raises(psycopg.OperationalError, match="timeout expired"):
+        await aconn_cls.connect(host="localhost", port=deaf_port, connect_timeout=1)
+    elapsed = time.time() - t0
     assert elapsed == pytest.approx(1.0, abs=0.05)
 
 
 async def test_close(aconn):
     assert not aconn.closed
     assert not aconn.broken
+
+    cur = aconn.cursor()
+
     await aconn.close()
     assert aconn.closed
     assert not aconn.broken
     assert aconn.pgconn.status == aconn.ConnStatus.BAD
 
-    cur = aconn.cursor()
-
     await aconn.close()
     assert aconn.closed
     assert aconn.pgconn.status == aconn.ConnStatus.BAD
@@ -84,6 +71,7 @@ async def test_close(aconn):
         await cur.execute("select 1")
 
 
+@pytest.mark.crdb_skip("pg_terminate_backend")
 async def test_broken(aconn):
     with pytest.raises(psycopg.OperationalError):
         await aconn.execute(
@@ -96,22 +84,32 @@ async def test_broken(aconn):
     assert aconn.broken
 
 
-async def test_connection_warn_close(dsn, recwarn):
-    conn = await AsyncConnection.connect(dsn)
+async def test_cursor_closed(aconn):
+    await aconn.close()
+    with pytest.raises(psycopg.OperationalError):
+        async with aconn.cursor("foo"):
+            pass
+        aconn.cursor("foo")
+    with pytest.raises(psycopg.OperationalError):
+        aconn.cursor()
+
+
+async def test_connection_warn_close(aconn_cls, dsn, recwarn):
+    conn = await aconn_cls.connect(dsn)
     await conn.close()
     del conn
     assert not recwarn, [str(w.message) for w in recwarn.list]
 
-    conn = await AsyncConnection.connect(dsn)
+    conn = await aconn_cls.connect(dsn)
     del conn
     assert "IDLE" in str(recwarn.pop(ResourceWarning).message)
 
-    conn = await AsyncConnection.connect(dsn)
+    conn = await aconn_cls.connect(dsn)
     await conn.execute("select 1")
     del conn
     assert "INTRANS" in str(recwarn.pop(ResourceWarning).message)
 
-    conn = await AsyncConnection.connect(dsn)
+    conn = await aconn_cls.connect(dsn)
     try:
         await conn.execute("select wat")
     except Exception:
@@ -119,45 +117,42 @@ async def test_connection_warn_close(dsn, recwarn):
     del conn
     assert "INERROR" in str(recwarn.pop(ResourceWarning).message)
 
-    async with await AsyncConnection.connect(dsn) as conn:
+    async with await aconn_cls.connect(dsn) as conn:
         pass
     del conn
     assert not recwarn, [str(w.message) for w in recwarn.list]
 
 
-async def test_context_commit(aconn, dsn):
+@pytest.mark.usefixtures("testctx")
+async def test_context_commit(aconn_cls, aconn, dsn):
     async with aconn:
         async with aconn.cursor() as cur:
-            await cur.execute("drop table if exists textctx")
-            await cur.execute("create table textctx ()")
+            await cur.execute("insert into testctx values (42)")
 
     assert aconn.closed
     assert not aconn.broken
 
-    async with await psycopg.AsyncConnection.connect(dsn) as aconn:
+    async with await aconn_cls.connect(dsn) as aconn:
         async with aconn.cursor() as cur:
-            await cur.execute("select * from textctx")
-            assert await cur.fetchall() == []
+            await cur.execute("select * from testctx")
+            assert await cur.fetchall() == [(42,)]
 
 
-async def test_context_rollback(aconn, dsn):
-    async with aconn.cursor() as cur:
-        await cur.execute("drop table if exists textctx")
-    await aconn.commit()
-
+@pytest.mark.usefixtures("testctx")
+async def test_context_rollback(aconn_cls, aconn, dsn):
     with pytest.raises(ZeroDivisionError):
         async with aconn:
             async with aconn.cursor() as cur:
-                await cur.execute("create table textctx ()")
+                await cur.execute("insert into testctx values (42)")
                 1 / 0
 
     assert aconn.closed
     assert not aconn.broken
 
-    async with await psycopg.AsyncConnection.connect(dsn) as aconn:
+    async with await aconn_cls.connect(dsn) as aconn:
         async with aconn.cursor() as cur:
-            with pytest.raises(UndefinedTable):
-                await cur.execute("select * from textctx")
+            await cur.execute("select * from testctx")
+            assert await cur.fetchall() == []
 
 
 async def test_context_close(aconn):
@@ -166,9 +161,10 @@ async def test_context_close(aconn):
         await aconn.close()
 
 
-async def test_context_rollback_no_clobber(conn, dsn, caplog):
+@pytest.mark.crdb_skip("pg_terminate_backend")
+async def test_context_inerror_rollback_no_clobber(aconn_cls, conn, dsn, caplog):
     with pytest.raises(ZeroDivisionError):
-        async with await psycopg.AsyncConnection.connect(dsn) as conn2:
+        async with await aconn_cls.connect(dsn) as conn2:
             await conn2.execute("select 1")
             conn.execute(
                 "select pg_terminate_backend(%s::int)",
@@ -179,11 +175,30 @@ async def test_context_rollback_no_clobber(conn, dsn, caplog):
     assert len(caplog.records) == 1
     rec = caplog.records[0]
     assert rec.levelno == logging.WARNING
-    assert "rolling back" in rec.message
+    assert "in rollback" in rec.message
 
 
-async def test_weakref(dsn):
-    conn = await psycopg.AsyncConnection.connect(dsn)
+@pytest.mark.crdb_skip("copy")
+async def test_context_active_rollback_no_clobber(aconn_cls, dsn, caplog):
+    caplog.set_level(logging.WARNING, logger="psycopg")
+
+    with pytest.raises(ZeroDivisionError):
+        async with await aconn_cls.connect(dsn) as conn:
+            conn.pgconn.exec_(b"copy (select generate_series(1, 10)) to stdout")
+            assert not conn.pgconn.error_message
+            status = conn.info.transaction_status
+            assert status == conn.TransactionStatus.ACTIVE
+            1 / 0
+
+    assert len(caplog.records) == 1
+    rec = caplog.records[0]
+    assert rec.levelno == logging.WARNING
+    assert "in rollback" in rec.message
+
+
+@pytest.mark.slow
+async def test_weakref(aconn_cls, dsn):
+    conn = await aconn_cls.connect(dsn)
     w = weakref.ref(conn)
     await conn.close()
     del conn
@@ -207,6 +222,26 @@ async def test_commit(aconn):
         await aconn.commit()
 
 
+@pytest.mark.crdb_skip("deferrable")
+async def test_commit_error(aconn):
+    await aconn.execute(
+        """
+        drop table if exists selfref;
+        create table selfref (
+            x serial primary key,
+            y int references selfref (x) deferrable initially deferred)
+        """
+    )
+    await aconn.commit()
+
+    await aconn.execute("insert into selfref (y) values (-1)")
+    with pytest.raises(e.ForeignKeyViolation):
+        await aconn.commit()
+    assert aconn.pgconn.transaction_status == aconn.TransactionStatus.IDLE
+    cur = await aconn.execute("select 1")
+    assert await cur.fetchone() == (1,)
+
+
 async def test_rollback(aconn):
     aconn.pgconn.exec_(b"drop table if exists foo")
     aconn.pgconn.exec_(b"create table foo (id int primary key)")
@@ -280,9 +315,10 @@ async def test_autocommit(aconn):
     assert aconn.autocommit is True
 
 
-async def test_autocommit_connect(dsn):
-    aconn = await psycopg.AsyncConnection.connect(dsn, autocommit=True)
+async def test_autocommit_connect(aconn_cls, dsn):
+    aconn = await aconn_cls.connect(dsn, autocommit=True)
     assert aconn.autocommit
+    await aconn.close()
 
 
 async def test_autocommit_intrans(aconn):
@@ -308,7 +344,7 @@ async def test_autocommit_inerror(aconn):
 async def test_autocommit_unknown(aconn):
     await aconn.close()
     assert aconn.pgconn.transaction_status == aconn.TransactionStatus.UNKNOWN
-    with pytest.raises(psycopg.ProgrammingError):
+    with pytest.raises(psycopg.OperationalError):
         await aconn.set_autocommit(True)
     assert not aconn.autocommit
 
@@ -318,18 +354,20 @@ async def test_autocommit_unknown(aconn):
     [
         ((), {}, ""),
         (("",), {}, ""),
-        (("host=foo user=bar",), {}, "host=foo user=bar"),
-        (("host=foo",), {"user": "baz"}, "host=foo user=baz"),
+        (("dbname=foo user=bar",), {}, "dbname=foo user=bar"),
+        (("dbname=foo",), {"user": "baz"}, "dbname=foo user=baz"),
         (
-            ("host=foo port=5432",),
-            {"host": "qux", "user": "joe"},
-            "host=qux user=joe port=5432",
+            ("dbname=foo port=5432",),
+            {"dbname": "qux", "user": "joe"},
+            "dbname=qux user=joe port=5432",
         ),
-        (("host=foo",), {"user": None}, "host=foo"),
+        (("dbname=foo",), {"user": None}, "dbname=foo"),
     ],
 )
-async def test_connect_args(monkeypatch, pgconn, args, kwargs, want):
-    the_conninfo = None
+async def test_connect_args(
+    aconn_cls, monkeypatch, setpgenv, pgconn, args, kwargs, want
+):
+    the_conninfo: str
 
     def fake_connect(conninfo):
         nonlocal the_conninfo
@@ -337,29 +375,32 @@ async def test_connect_args(monkeypatch, pgconn, args, kwargs, want):
         return pgconn
         yield
 
+    setpgenv({})
     monkeypatch.setattr(psycopg.connection, "connect", fake_connect)
-    await psycopg.AsyncConnection.connect(*args, **kwargs)
+    conn = await aconn_cls.connect(*args, **kwargs)
     assert conninfo_to_dict(the_conninfo) == conninfo_to_dict(want)
+    await conn.close()
 
 
 @pytest.mark.parametrize(
-    "args, kwargs",
+    "args, kwargs, exctype",
     [
-        (("host=foo", "host=bar"), {}),
-        (("", ""), {}),
-        ((), {"nosuchparam": 42}),
+        (("host=foo", "host=bar"), {}, TypeError),
+        (("", ""), {}, TypeError),
+        ((), {"nosuchparam": 42}, psycopg.ProgrammingError),
     ],
 )
-async def test_connect_badargs(monkeypatch, pgconn, args, kwargs):
+async def test_connect_badargs(aconn_cls, monkeypatch, pgconn, args, kwargs, exctype):
     def fake_connect(conninfo):
         return pgconn
         yield
 
     monkeypatch.setattr(psycopg.connection, "connect", fake_connect)
-    with pytest.raises((TypeError, psycopg.ProgrammingError)):
-        await psycopg.AsyncConnection.connect(*args, **kwargs)
+    with pytest.raises(exctype):
+        await aconn_cls.connect(*args, **kwargs)
 
 
+@pytest.mark.crdb_skip("pg_terminate_backend")
 async def test_broken_connection(aconn):
     cur = aconn.cursor()
     with pytest.raises(psycopg.DatabaseError):
@@ -367,6 +408,7 @@ async def test_broken_connection(aconn):
     assert aconn.closed
 
 
+@pytest.mark.crdb_skip("do")
 async def test_notice_handlers(aconn, caplog):
     caplog.set_level(logging.WARNING, logger="psycopg")
     messages = []
@@ -381,13 +423,11 @@ async def test_notice_handlers(aconn, caplog):
     aconn.add_notice_handler(cb1)
     aconn.add_notice_handler(cb2)
     aconn.add_notice_handler("the wrong thing")
-    aconn.add_notice_handler(lambda diag: severities.append(diag.severity))
+    aconn.add_notice_handler(lambda diag: severities.append(diag.severity_nonlocalized))
 
     aconn.pgconn.exec_(b"set client_min_messages to notice")
     cur = aconn.cursor()
-    await cur.execute(
-        "do $$begin raise notice 'hello notice'; end$$ language plpgsql"
-    )
+    await cur.execute("do $$begin raise notice 'hello notice'; end$$ language plpgsql")
     assert messages == ["hello notice"]
     assert severities == ["NOTICE"]
 
@@ -412,6 +452,7 @@ async def test_notice_handlers(aconn, caplog):
         aconn.remove_notice_handler(cb1)
 
 
+@pytest.mark.crdb_skip("notify")
 async def test_notify_handlers(aconn):
     nots1 = []
     nots2 = []
@@ -471,27 +512,31 @@ async def test_execute_binary(aconn):
     assert cur.pgresult.fformat(0) == 1
 
 
-async def test_row_factory(dsn):
-    conn = await AsyncConnection.connect(dsn)
-    assert conn.row_factory is tuple_row
+async def test_row_factory(aconn_cls, dsn):
+    defaultconn = await aconn_cls.connect(dsn)
+    assert defaultconn.row_factory is tuple_row
+    await defaultconn.close()
 
-    conn = await AsyncConnection.connect(dsn, row_factory=my_row_factory)
+    conn = await aconn_cls.connect(dsn, row_factory=my_row_factory)
     assert conn.row_factory is my_row_factory
 
     cur = await conn.execute("select 'a' as ve")
     assert await cur.fetchone() == ["Ave"]
 
-    async with conn.cursor(row_factory=lambda c: set) as cur:
-        await cur.execute("select 1, 1, 2")
-        assert await cur.fetchall() == [{1, 2}]
+    async with conn.cursor(row_factory=lambda c: lambda t: set(t)) as cur1:
+        await cur1.execute("select 1, 1, 2")
+        assert await cur1.fetchall() == [{1, 2}]
 
-    async with conn.cursor(row_factory=tuple_row) as cur:
-        await cur.execute("select 1, 1, 2")
-        assert await cur.fetchall() == [(1, 1, 2)]
+    async with conn.cursor(row_factory=tuple_row) as cur2:
+        await cur2.execute("select 1, 1, 2")
+        assert await cur2.fetchall() == [(1, 1, 2)]
 
+    # TODO: maybe fix something to get rid of 'type: ignore' below.
     conn.row_factory = tuple_row
-    cur = await conn.execute("select 'vale'")
-    assert await cur.fetchone() == ("vale",)
+    cur3 = await conn.execute("select 'vale'")
+    r = await cur3.fetchone()
+    assert r and r == ("vale",)
+    await conn.close()
 
 
 async def test_str(aconn):
@@ -510,7 +555,7 @@ async def test_fileno(aconn):
 async def test_cursor_factory(aconn):
     assert aconn.cursor_factory is psycopg.AsyncCursor
 
-    class MyCursor(psycopg.AsyncCursor):
+    class MyCursor(psycopg.AsyncCursor[psycopg.rows.Row]):
         pass
 
     aconn.cursor_factory = MyCursor
@@ -521,10 +566,20 @@ async def test_cursor_factory(aconn):
         assert isinstance(cur, MyCursor)
 
 
+async def test_cursor_factory_connect(aconn_cls, dsn):
+    class MyCursor(psycopg.AsyncCursor[psycopg.rows.Row]):
+        pass
+
+    async with await aconn_cls.connect(dsn, cursor_factory=MyCursor) as conn:
+        assert conn.cursor_factory is MyCursor
+        cur = conn.cursor()
+        assert type(cur) is MyCursor
+
+
 async def test_server_cursor_factory(aconn):
     assert aconn.server_cursor_factory is psycopg.AsyncServerCursor
 
-    class MyServerCursor(psycopg.AsyncServerCursor):
+    class MyServerCursor(psycopg.AsyncServerCursor[psycopg.rows.Row]):
         pass
 
     aconn.server_cursor_factory = MyServerCursor
@@ -532,34 +587,32 @@ async def test_server_cursor_factory(aconn):
         assert isinstance(cur, MyServerCursor)
 
 
-@pytest.mark.parametrize("attr", list(tx_params))
-async def test_transaction_param_default(aconn, attr):
-    assert getattr(aconn, attr) is None
-    guc = tx_params[attr]["guc"]
+@pytest.mark.parametrize("param", tx_params)
+async def test_transaction_param_default(aconn, param):
+    assert getattr(aconn, param.name) is None
     cur = await aconn.execute(
         "select current_setting(%s), current_setting(%s)",
-        [f"transaction_{guc}", f"default_transaction_{guc}"],
+        [f"transaction_{param.guc}", f"default_transaction_{param.guc}"],
     )
     current, default = await cur.fetchone()
     assert current == default
 
 
-@pytest.mark.parametrize("attr", list(tx_params))
-async def test_transaction_param_readonly_property(aconn, attr):
+@pytest.mark.parametrize("param", tx_params)
+async def test_transaction_param_readonly_property(aconn, param):
     with pytest.raises(AttributeError):
-        setattr(aconn, attr, None)
+        setattr(aconn, param.name, None)
 
 
 @pytest.mark.parametrize("autocommit", [True, False])
-@pytest.mark.parametrize("attr", list(tx_params))
-async def test_set_transaction_param_implicit(aconn, attr, autocommit):
-    guc = tx_params[attr]["guc"]
+@pytest.mark.parametrize("param", tx_params_isolation)
+async def test_set_transaction_param_implicit(aconn, param, autocommit):
     await aconn.set_autocommit(autocommit)
-    for value in tx_params[attr]["values"]:
-        await getattr(aconn, tx_params[attr]["set_method"])(value)
+    for value in param.values:
+        await getattr(aconn, f"set_{param.name}")(value)
         cur = await aconn.execute(
             "select current_setting(%s), current_setting(%s)",
-            [f"transaction_{guc}", f"default_transaction_{guc}"],
+            [f"transaction_{param.guc}", f"default_transaction_{param.guc}"],
         )
         pgval, default = await cur.fetchone()
         if autocommit:
@@ -570,54 +623,56 @@ async def test_set_transaction_param_implicit(aconn, attr, autocommit):
 
 
 @pytest.mark.parametrize("autocommit", [True, False])
-@pytest.mark.parametrize("attr", list(tx_params))
-async def test_set_transaction_param_block(aconn, attr, autocommit):
-    guc = tx_params[attr]["guc"]
+@pytest.mark.parametrize("param", tx_params_isolation)
+async def test_set_transaction_param_block(aconn, param, autocommit):
     await aconn.set_autocommit(autocommit)
-    for value in tx_params[attr]["values"]:
-        await getattr(aconn, tx_params[attr]["set_method"])(value)
+    for value in param.values:
+        await getattr(aconn, f"set_{param.name}")(value)
         async with aconn.transaction():
             cur = await aconn.execute(
-                "select current_setting(%s)", [f"transaction_{guc}"]
+                "select current_setting(%s)", [f"transaction_{param.guc}"]
             )
             pgval = (await cur.fetchone())[0]
         assert tx_values_map[pgval] == value
 
 
-@pytest.mark.parametrize("attr", list(tx_params))
-async def test_set_transaction_param_not_intrans_implicit(aconn, attr):
+@pytest.mark.parametrize("param", tx_params)
+async def test_set_transaction_param_not_intrans_implicit(aconn, param):
     await aconn.execute("select 1")
-    value = tx_params[attr]["values"][0]
+    value = param.values[0]
     with pytest.raises(psycopg.ProgrammingError):
-        await getattr(aconn, tx_params[attr]["set_method"])(value)
+        await getattr(aconn, f"set_{param.name}")(value)
 
 
-@pytest.mark.parametrize("attr", list(tx_params))
-async def test_set_transaction_param_not_intrans_block(aconn, attr):
-    value = tx_params[attr]["values"][0]
+@pytest.mark.parametrize("param", tx_params)
+async def test_set_transaction_param_not_intrans_block(aconn, param):
+    value = param.values[0]
     async with aconn.transaction():
         with pytest.raises(psycopg.ProgrammingError):
-            await getattr(aconn, tx_params[attr]["set_method"])(value)
+            await getattr(aconn, f"set_{param.name}")(value)
 
 
-@pytest.mark.parametrize("attr", list(tx_params))
-async def test_set_transaction_param_not_intrans_external(aconn, attr):
-    value = tx_params[attr]["values"][0]
+@pytest.mark.parametrize("param", tx_params)
+async def test_set_transaction_param_not_intrans_external(aconn, param):
+    value = param.values[0]
     await aconn.set_autocommit(True)
     await aconn.execute("begin")
     with pytest.raises(psycopg.ProgrammingError):
-        await getattr(aconn, tx_params[attr]["set_method"])(value)
+        await getattr(aconn, f"set_{param.name}")(value)
 
 
+@pytest.mark.crdb("skip", reason="transaction isolation")
 async def test_set_transaction_param_all(aconn):
-    for attr in tx_params:
-        value = tx_params[attr]["values"][0]
-        await getattr(aconn, tx_params[attr]["set_method"])(value)
+    params: List[Any] = tx_params[:]
+    params[2] = params[2].values[0]
 
-    for attr in tx_params:
-        guc = tx_params[attr]["guc"]
+    for param in params:
+        value = param.values[0]
+        await getattr(aconn, f"set_{param.name}")(value)
+
+    for param in params:
         cur = await aconn.execute(
-            "select current_setting(%s)", [f"transaction_{guc}"]
+            "select current_setting(%s)", [f"transaction_{param.guc}"]
         )
         pgval = (await cur.fetchone())[0]
         assert tx_values_map[pgval] == value
@@ -639,51 +694,58 @@ async def test_set_transaction_param_strange(aconn):
 
 
 @pytest.mark.parametrize("dsn, kwargs, exp", conninfo_params_timeout)
-async def test_get_connection_params(dsn, kwargs, exp):
-    params = await AsyncConnection._get_connection_params(dsn, **kwargs)
+async def test_get_connection_params(aconn_cls, dsn, kwargs, exp, setpgenv):
+    setpgenv({})
+    params = await aconn_cls._get_connection_params(dsn, **kwargs)
     conninfo = make_conninfo(**params)
     assert conninfo_to_dict(conninfo) == exp[0]
     assert params["connect_timeout"] == exp[1]
 
 
-async def test_connect_context_adapters(dsn):
+async def test_connect_context_adapters(aconn_cls, dsn):
     ctx = psycopg.adapt.AdaptersMap(psycopg.adapters)
     ctx.register_dumper(str, make_bin_dumper("b"))
     ctx.register_dumper(str, make_dumper("t"))
 
-    conn = await psycopg.AsyncConnection.connect(dsn, context=ctx)
+    conn = await aconn_cls.connect(dsn, context=ctx)
 
     cur = await conn.execute("select %s", ["hello"])
     assert (await cur.fetchone())[0] == "hellot"
     cur = await conn.execute("select %b", ["hello"])
     assert (await cur.fetchone())[0] == "hellob"
+    await conn.close()
 
 
-async def test_connect_context_copy(dsn, aconn):
+async def test_connect_context_copy(aconn_cls, dsn, aconn):
     aconn.adapters.register_dumper(str, make_bin_dumper("b"))
     aconn.adapters.register_dumper(str, make_dumper("t"))
 
-    aconn2 = await psycopg.AsyncConnection.connect(dsn, context=aconn)
+    aconn2 = await aconn_cls.connect(dsn, context=aconn)
 
     cur = await aconn2.execute("select %s", ["hello"])
     assert (await cur.fetchone())[0] == "hellot"
     cur = await aconn2.execute("select %b", ["hello"])
     assert (await cur.fetchone())[0] == "hellob"
+    await aconn2.close()
 
 
-@pytest.mark.skipif(sys.platform != "win32", reason="windows only test")
-def test_windows_error(dsn):
-    loop = asyncio.ProactorEventLoop()
+async def test_cancel_closed(aconn):
+    await aconn.close()
+    aconn.cancel()
 
-    async def go():
-        with pytest.raises(
-            InterfaceError,
-            match="Psycopg cannot use the 'ProactorEventLoop'",
-        ):
-            await psycopg.AsyncConnection.connect(dsn)
 
-    try:
-        loop.run_until_complete(go())
-    finally:
-        loop.run_until_complete(loop.shutdown_asyncgens())
-        loop.close()
+async def test_resolve_hostaddr_conn(monkeypatch, fake_resolve):  # noqa: F811
+    got = []
+
+    def fake_connect_gen(conninfo, **kwargs):
+        got.append(conninfo)
+        1 / 0
+
+    monkeypatch.setattr(psycopg.AsyncConnection, "_connect_gen", fake_connect_gen)
+
+    with pytest.raises(ZeroDivisionError):
+        await psycopg.AsyncConnection.connect("host=foo.com")
+
+    assert len(got) == 1
+    want = {"host": "foo.com", "hostaddr": "1.1.1.1"}
+    assert conninfo_to_dict(got[0]) == want
