@@ -1,4 +1,3 @@
-import gc
 import string
 import struct
 import hashlib
@@ -19,7 +18,7 @@ from psycopg.types import TypeInfo
 from psycopg.types.hstore import register_hstore
 from psycopg.types.numeric import Int4
 
-from .utils import eur, gc_collect
+from .utils import eur, gc_collect, gc_count
 
 pytestmark = pytest.mark.crdb_skip("copy")
 
@@ -72,19 +71,31 @@ def test_copy_out_read(conn, format):
 
 
 @pytest.mark.parametrize("format", Format)
-def test_copy_out_iter(conn, format):
+@pytest.mark.parametrize("row_factory", ["tuple_row", "dict_row", "namedtuple_row"])
+def test_copy_out_iter(conn, format, row_factory):
     if format == pq.Format.TEXT:
         want = [row + b"\n" for row in sample_text.splitlines()]
     else:
         want = sample_binary_rows
 
-    cur = conn.cursor()
+    rf = getattr(psycopg.rows, row_factory)
+    cur = conn.cursor(row_factory=rf)
     with cur.copy(f"copy ({sample_values}) to stdout (format {format.name})") as copy:
         assert list(copy) == want
 
     assert conn.info.transaction_status == conn.TransactionStatus.INTRANS
 
 
+@pytest.mark.parametrize("format", Format)
+@pytest.mark.parametrize("row_factory", ["tuple_row", "dict_row", "namedtuple_row"])
+def test_copy_out_no_result(conn, format, row_factory):
+    rf = getattr(psycopg.rows, row_factory)
+    cur = conn.cursor(row_factory=rf)
+    with cur.copy(f"copy ({sample_values}) to stdout (format {format.name})"):
+        with pytest.raises(e.ProgrammingError):
+            cur.fetchone()
+
+
 @pytest.mark.parametrize("ph, params", [("%s", (10,)), ("%(n)s", {"n": 10})])
 def test_copy_out_param(conn, ph, params):
     cur = conn.cursor()
@@ -727,7 +738,7 @@ def test_copy_to_leaks(conn_cls, dsn, faker, fmt, set_types, method):
     for i in range(3):
         work()
         gc_collect()
-        n.append(len(gc.get_objects()))
+        n.append(gc_count())
 
     assert n[0] == n[1] == n[2], f"objects leaked: {n[1] - n[0]}, {n[2] - n[1]}"
 
@@ -770,7 +781,7 @@ def test_copy_from_leaks(conn_cls, dsn, faker, fmt, set_types):
     for i in range(3):
         work()
         gc_collect()
-        n.append(len(gc.get_objects()))
+        n.append(gc_count())
 
     assert n[0] == n[1] == n[2], f"objects leaked: {n[1] - n[0]}, {n[2] - n[1]}"
 
