@@ -1,4 +1,3 @@
-import gc
 import pickle
 import weakref
 import datetime as dt
@@ -13,7 +12,7 @@ from psycopg.adapt import PyFormat
 from psycopg.postgres import types as builtins
 from psycopg.rows import RowMaker
 
-from .utils import gc_collect
+from .utils import gc_collect, gc_count
 from .fix_crdb import is_crdb, crdb_encoding, crdb_time_precision
 
 
@@ -428,6 +427,35 @@ def test_rownumber(conn):
     assert cur.rownumber == 42
 
 
+@pytest.mark.parametrize("query", ["", "set timezone to utc"])
+def test_rownumber_none(conn, query):
+    cur = conn.cursor()
+    cur.execute(query)
+    assert cur.rownumber is None
+
+
+def test_rownumber_mixed(conn):
+    cur = conn.cursor()
+    cur.execute(
+        """
+select x from generate_series(1, 3) x;
+set timezone to utc;
+select x from generate_series(4, 6) x;
+"""
+    )
+    assert cur.rownumber == 0
+    assert cur.fetchone() == (1,)
+    assert cur.rownumber == 1
+    assert cur.fetchone() == (2,)
+    assert cur.rownumber == 2
+    cur.nextset()
+    assert cur.rownumber is None
+    cur.nextset()
+    assert cur.rownumber == 0
+    assert cur.fetchone() == (4,)
+    assert cur.rownumber == 1
+
+
 def test_iter(conn):
     cur = conn.cursor()
     cur.execute("select generate_series(1, 3)")
@@ -896,7 +924,7 @@ def test_leak(conn_cls, dsn, faker, fmt, fmt_out, fetch, row_factory):
     for i in range(3):
         work()
         gc_collect()
-        n.append(len(gc.get_objects()))
+        n.append(gc_count())
     assert n[0] == n[1] == n[2], f"objects leaked: {n[1] - n[0]}, {n[2] - n[1]}"
 
 
