@@ -1,16 +1,18 @@
 import pickle
+from typing import List
 from weakref import ref
 
 import pytest
 
+import psycopg
 from psycopg import pq
 from psycopg import errors as e
 
-from .utils import gc_collect
-
-eur = "\u20ac"
+from .utils import eur, gc_collect
+from .fix_crdb import is_crdb
 
 
+@pytest.mark.crdb_skip("severity_nonlocalized")
 def test_error_diag(conn):
     cur = conn.cursor()
     with pytest.raises(e.DatabaseError) as excinfo:
@@ -19,7 +21,7 @@ def test_error_diag(conn):
     exc = excinfo.value
     diag = exc.diag
     assert diag.sqlstate == "42P01"
-    assert diag.severity == "ERROR"
+    assert diag.severity_nonlocalized == "ERROR"
 
 
 def test_diag_all_attrs(pgconn):
@@ -34,7 +36,8 @@ def test_diag_right_attr(pgconn, monkeypatch):
     res = pgconn.make_empty_result(pq.ExecStatus.NONFATAL_ERROR)
     diag = e.Diagnostic(res)
 
-    checked = []
+    to_check: pq.DiagnosticField
+    checked: List[pq.DiagnosticField] = []
 
     def check_val(self, v):
         nonlocal to_check
@@ -51,24 +54,26 @@ def test_diag_right_attr(pgconn, monkeypatch):
 
 
 def test_diag_attr_values(conn):
-    cur = conn.cursor()
-    cur.execute(
+    if is_crdb(conn):
+        conn.execute("set experimental_enable_temp_tables = 'on'")
+    conn.execute(
         """
         create temp table test_exc (
             data int constraint chk_eq1 check (data = 1)
         )"""
     )
     with pytest.raises(e.Error) as exc:
-        cur.execute("insert into test_exc values(2)")
+        conn.execute("insert into test_exc values(2)")
     diag = exc.value.diag
     assert diag.sqlstate == "23514"
-    assert diag.schema_name[:7] == "pg_temp"
-    assert diag.table_name == "test_exc"
     assert diag.constraint_name == "chk_eq1"
-    if conn.pgconn.server_version >= 90600:
+    if not is_crdb(conn):
+        assert diag.table_name == "test_exc"
+        assert diag.schema_name and diag.schema_name[:7] == "pg_temp"
         assert diag.severity_nonlocalized == "ERROR"
 
 
+@pytest.mark.crdb_skip("do")
 @pytest.mark.parametrize("enc", ["utf8", "latin9"])
 def test_diag_encoding(conn, enc):
     msgs = []
@@ -76,12 +81,11 @@ def test_diag_encoding(conn, enc):
     conn.add_notice_handler(lambda diag: msgs.append(diag.message_primary))
     conn.execute(f"set client_encoding to {enc}")
     cur = conn.cursor()
-    cur.execute(
-        "do $$begin raise notice 'hello %', chr(8364); end$$ language plpgsql"
-    )
+    cur.execute("do $$begin raise notice 'hello %', chr(8364); end$$ language plpgsql")
     assert msgs == [f"hello {eur}"]
 
 
+@pytest.mark.crdb_skip("do")
 @pytest.mark.parametrize("enc", ["utf8", "latin9"])
 def test_error_encoding(conn, enc):
     with conn.transaction():
@@ -97,7 +101,7 @@ def test_error_encoding(conn, enc):
         )
 
     diag = excinfo.value.diag
-    assert f'"{eur}"' in diag.message_primary
+    assert diag.message_primary and f'"{eur}"' in diag.message_primary
     assert diag.sqlstate == "42P01"
 
 
@@ -165,6 +169,7 @@ def test_diag_pickle(conn):
     assert diag2.sqlstate == "42P01"
 
 
+@pytest.mark.slow
 def test_diag_survives_cursor(conn):
     cur = conn.cursor()
     with pytest.raises(e.Error) as exc:
@@ -193,6 +198,7 @@ def test_diag_independent(conn):
     assert exc2.value.diag.sqlstate == "42P01"
 
 
+@pytest.mark.crdb_skip("deferrable")
 def test_diag_from_commit(conn):
     cur = conn.cursor()
     cur.execute(
@@ -211,6 +217,7 @@ def test_diag_from_commit(conn):
 
 
 @pytest.mark.asyncio
+@pytest.mark.crdb_skip("deferrable")
 async def test_diag_from_commit_async(aconn):
     cur = aconn.cursor()
     await cur.execute(
@@ -233,7 +240,70 @@ def test_query_context(conn):
         conn.execute("select * from wat")
 
     s = str(exc.value)
-    assert "from wat" in s, s
+    if not is_crdb(conn):
+        assert "from wat" in s, s
+    assert exc.value.diag.message_primary
     assert exc.value.diag.message_primary in s
     assert "ERROR" not in s
     assert not s.endswith("\n")
+
+
+@pytest.mark.crdb_skip("do")
+def test_unknown_sqlstate(conn):
+    code = "PXX99"
+    with pytest.raises(KeyError):
+        e.lookup(code)
+
+    with pytest.raises(e.ProgrammingError) as excinfo:
+        conn.execute(
+            f"""
+            do $$begin
+            raise exception 'made up code' using errcode = '{code}';
+            end$$ language plpgsql
+            """
+        )
+    exc = excinfo.value
+    assert exc.diag.sqlstate == code
+    assert exc.sqlstate == code
+    # Survives pickling too
+    pexc = pickle.loads(pickle.dumps(exc))
+    assert pexc.sqlstate == code
+
+
+def test_pgconn_error(conn_cls):
+    with pytest.raises(psycopg.OperationalError) as excinfo:
+        conn_cls.connect("dbname=nosuchdb")
+
+    exc = excinfo.value
+    assert exc.pgconn
+    assert exc.pgconn.db == b"nosuchdb"
+
+
+def test_pgconn_error_pickle(conn_cls):
+    with pytest.raises(psycopg.OperationalError) as excinfo:
+        conn_cls.connect("dbname=nosuchdb")
+
+    exc = pickle.loads(pickle.dumps(excinfo.value))
+    assert exc.pgconn is None
+
+
+def test_pgresult(conn):
+    with pytest.raises(e.DatabaseError) as excinfo:
+        conn.execute("select 1 from wat")
+
+    exc = excinfo.value
+    assert exc.pgresult
+    assert exc.pgresult.error_field(pq.DiagnosticField.SQLSTATE) == b"42P01"
+
+
+def test_pgresult_pickle(conn):
+    with pytest.raises(e.DatabaseError) as excinfo:
+        conn.execute("select 1 from wat")
+
+    exc = pickle.loads(pickle.dumps(excinfo.value))
+    assert exc.pgresult is None
+    assert exc.diag.sqlstate == "42P01"
+
+
+def test_blank_sqlstate(conn):
+    assert e.get_base_exception("") is e.DatabaseError
