@@ -1,5 +1,7 @@
 import pytest
 
+from psycopg._cmodule import _psycopg
+
 
 @pytest.mark.parametrize(
     "args, kwargs, want_conninfo",
@@ -15,7 +17,7 @@ def test_connect(monkeypatch, dsn, args, kwargs, want_conninfo):
     # Details of the params manipulation are in test_conninfo.
     import psycopg.connection
 
-    orig_connect = psycopg.connection.connect
+    orig_connect = psycopg.connection.connect  # type: ignore
 
     got_conninfo = None
 
@@ -26,5 +28,30 @@ def test_connect(monkeypatch, dsn, args, kwargs, want_conninfo):
 
     monkeypatch.setattr(psycopg.connection, "connect", mock_connect)
 
-    psycopg.connect(*args, **kwargs)
+    conn = psycopg.connect(*args, **kwargs)
     assert got_conninfo == want_conninfo
+    conn.close()
+
+
+def test_version(mypy):
+    cp = mypy.run_on_source(
+        """\
+from psycopg import __version__
+assert __version__
+"""
+    )
+    assert not cp.stdout
+
+
+@pytest.mark.skipif(_psycopg is None, reason="C module test")
+def test_version_c(mypy):
+    # can be psycopg_c, psycopg_binary
+    cpackage = _psycopg.__name__.split(".")[0]
+
+    cp = mypy.run_on_source(
+        f"""\
+from {cpackage} import __version__
+assert __version__
+"""
+    )
+    assert not cp.stdout
