@@ -1,6 +1,6 @@
 # test_sql.py - tests for the psycopg2.sql module
 
-# Copyright (C) 2020-2021 The Psycopg Team
+# Copyright (C) 2020 The Psycopg Team
 
 import re
 import datetime as dt
@@ -10,8 +10,11 @@ import pytest
 from psycopg import pq, sql, ProgrammingError
 from psycopg.adapt import PyFormat
 from psycopg._encodings import py2pgenc
+from psycopg.types import TypeInfo
+from psycopg.types.string import StrDumper
 
-eur = "\u20ac"
+from .utils import eur
+from .fix_crdb import crdb_encoding, crdb_scs_off
 
 
 @pytest.mark.parametrize(
@@ -28,7 +31,7 @@ def test_quote(obj, quoted):
     assert sql.quote(obj) == quoted
 
 
-@pytest.mark.parametrize("scs", ["on", "off"])
+@pytest.mark.parametrize("scs", ["on", crdb_scs_off("off")])
 def test_quote_roundtrip(conn, scs):
     messages = []
     conn.add_notice_handler(lambda msg: messages.append(msg.message_primary))
@@ -44,11 +47,12 @@ def test_quote_roundtrip(conn, scs):
         assert not messages, f"error with {want!r}"
 
 
-def test_quote_stable_despite_deranged_libpq(conn):
+@pytest.mark.parametrize("dummy", [crdb_scs_off("off")])
+def test_quote_stable_despite_deranged_libpq(conn, dummy):
     # Verify the libpq behaviour of PQescapeString using the last setting seen.
     # Check that we are not affected by it.
     good_str = " E'\\\\'"
-    good_bytes = " E'\\\\000'"
+    good_bytes = " E'\\\\000'::bytea"
     conn.execute("set standard_conforming_strings to on")
     assert pq.Escaping().escape_string(b"\\") == b"\\"
     assert sql.quote("\\") == good_str
@@ -109,7 +113,7 @@ class TestSqlFormat:
     def test_compose_literal(self, conn):
         s = sql.SQL("select {0};").format(sql.Literal(dt.date(2016, 12, 31)))
         s1 = s.as_string(conn)
-        assert s1 == "select '2016-12-31';"
+        assert s1 == "select '2016-12-31'::date;"
 
     def test_compose_empty(self, conn):
         s = sql.SQL("select foo;").format()
@@ -160,10 +164,8 @@ class TestSqlFormat:
             s.as_string(conn)
 
     def test_auto_literal(self, conn):
-        s = sql.SQL("select {}, {}, {}").format(
-            "he'lo", 10, dt.date(2020, 1, 1)
-        )
-        assert s.as_string(conn) == "select 'he''lo', 10, '2020-01-01'"
+        s = sql.SQL("select {}, {}, {}").format("he'lo", 10, dt.date(2020, 1, 1))
+        assert s.as_string(conn) == "select 'he''lo', 10, '2020-01-01'::date"
 
     def test_execute(self, conn):
         cur = conn.cursor()
@@ -177,9 +179,7 @@ class TestSqlFormat:
         cur.execute(
             sql.SQL("insert into {0} (id, {1}) values (%s, {2})").format(
                 sql.Identifier("test_compose"),
-                sql.SQL(", ").join(
-                    map(sql.Identifier, ["foo", "bar", "ba'z"])
-                ),
+                sql.SQL(", ").join(map(sql.Identifier, ["foo", "bar", "ba'z"])),
                 (sql.Placeholder() * 3).join(", "),
             ),
             (10, "a", "b", "c"),
@@ -200,9 +200,7 @@ class TestSqlFormat:
         cur.executemany(
             sql.SQL("insert into {0} (id, {1}) values (%s, {2})").format(
                 sql.Identifier("test_compose"),
-                sql.SQL(", ").join(
-                    map(sql.Identifier, ["foo", "bar", "ba'z"])
-                ),
+                sql.SQL(", ").join(map(sql.Identifier, ["foo", "bar", "ba'z"])),
                 (sql.Placeholder() * 3).join(", "),
             ),
             [(10, "a", "b", "c"), (20, "d", "e", "f")],
@@ -211,6 +209,7 @@ class TestSqlFormat:
         cur.execute("select * from test_compose")
         assert cur.fetchall() == [(10, "a", "b", "c"), (20, "d", "e", "f")]
 
+    @pytest.mark.crdb_skip("copy")
     def test_copy(self, conn):
         cur = conn.cursor()
         cur.execute(
@@ -248,9 +247,9 @@ class TestIdentifier:
         with pytest.raises(TypeError):
             sql.Identifier()
         with pytest.raises(TypeError):
-            sql.Identifier(10)
+            sql.Identifier(10)  # type: ignore[arg-type]
         with pytest.raises(TypeError):
-            sql.Identifier(dt.date(2016, 12, 31))
+            sql.Identifier(dt.date(2016, 12, 31))  # type: ignore[arg-type]
 
     def test_repr(self):
         obj = sql.Identifier("fo'o")
@@ -282,11 +281,11 @@ class TestIdentifier:
     @pytest.mark.parametrize(
         "args, want, enc",
         [
-            (("foo",), '"foo"', "ascii"),
-            (("foo", "bar"), '"foo"."bar"', "ascii"),
-            (("fo'o", 'ba"r'), '"fo\'o"."ba""r"', "ascii"),
+            crdb_encoding(("foo",), '"foo"', "ascii"),
+            crdb_encoding(("foo", "bar"), '"foo"."bar"', "ascii"),
+            crdb_encoding(("fo'o", 'ba"r'), '"fo\'o"."ba""r"', "ascii"),
             (("foo", eur), f'"foo"."{eur}"', "utf8"),
-            (("foo", eur), f'"foo"."{eur}"', "latin9"),
+            crdb_encoding(("foo", eur), f'"foo"."{eur}"', "latin9"),
         ],
     )
     def test_as_bytes(self, conn, args, want, enc):
@@ -315,24 +314,20 @@ class TestLiteral:
 
     def test_as_string(self, conn):
         assert sql.Literal(None).as_string(conn) == "NULL"
-        assert noe(sql.Literal("foo").as_string(conn)) == "'foo'"
+        assert no_e(sql.Literal("foo").as_string(conn)) == "'foo'"
         assert sql.Literal(42).as_string(conn) == "42"
-        assert (
-            sql.Literal(dt.date(2017, 1, 1)).as_string(conn) == "'2017-01-01'"
-        )
+        assert sql.Literal(dt.date(2017, 1, 1)).as_string(conn) == "'2017-01-01'::date"
 
     def test_as_bytes(self, conn):
         assert sql.Literal(None).as_bytes(conn) == b"NULL"
-        assert noe(sql.Literal("foo").as_bytes(conn)) == b"'foo'"
+        assert no_e(sql.Literal("foo").as_bytes(conn)) == b"'foo'"
         assert sql.Literal(42).as_bytes(conn) == b"42"
-        assert (
-            sql.Literal(dt.date(2017, 1, 1)).as_bytes(conn) == b"'2017-01-01'"
-        )
+        assert sql.Literal(dt.date(2017, 1, 1)).as_bytes(conn) == b"'2017-01-01'::date"
 
-        conn.execute("set client_encoding to utf8")
-        assert sql.Literal(eur).as_bytes(conn) == f"'{eur}'".encode()
-        conn.execute("set client_encoding to latin9")
-        assert sql.Literal(eur).as_bytes(conn) == f"'{eur}'".encode("latin9")
+    @pytest.mark.parametrize("encoding", ["utf8", crdb_encoding("latin9")])
+    def test_as_bytes_encoding(self, conn, encoding):
+        conn.execute(f"set client_encoding to {encoding}")
+        assert sql.Literal(eur).as_bytes(conn) == f"'{eur}'".encode(encoding)
 
     def test_eq(self):
         assert sql.Literal("foo") == sql.Literal("foo")
@@ -347,6 +342,63 @@ class TestLiteral:
         with pytest.raises(ProgrammingError):
             sql.Literal(Foo()).as_string(conn)
 
+    def test_array(self, conn):
+        assert (
+            sql.Literal([dt.date(2000, 1, 1)]).as_string(conn)
+            == "'{2000-01-01}'::date[]"
+        )
+
+    def test_short_name_builtin(self, conn):
+        assert sql.Literal(dt.time(0, 0)).as_string(conn) == "'00:00:00'::time"
+        assert (
+            sql.Literal(dt.datetime(2000, 1, 1)).as_string(conn)
+            == "'2000-01-01 00:00:00'::timestamp"
+        )
+        assert (
+            sql.Literal([dt.datetime(2000, 1, 1)]).as_string(conn)
+            == "'{\"2000-01-01 00:00:00\"}'::timestamp[]"
+        )
+
+    def test_text_literal(self, conn):
+        conn.adapters.register_dumper(str, StrDumper)
+        assert sql.Literal("foo").as_string(conn) == "'foo'"
+
+    @pytest.mark.crdb_skip("composite")  # create type, actually
+    @pytest.mark.parametrize("name", ["a-b", f"{eur}", "order", "foo bar"])
+    def test_invalid_name(self, conn, name):
+        conn.execute(
+            f"""
+            set client_encoding to utf8;
+            create type "{name}";
+            create function invin(cstring) returns "{name}"
+                language internal immutable strict as 'textin';
+            create function invout("{name}") returns cstring
+                language internal immutable strict as 'textout';
+            create type "{name}" (input=invin, output=invout, like=text);
+            """
+        )
+        info = TypeInfo.fetch(conn, f'"{name}"')
+
+        class InvDumper(StrDumper):
+            oid = info.oid
+
+            def dump(self, obj):
+                rv = super().dump(obj)
+                return b"%s-inv" % rv
+
+        info.register(conn)
+        conn.adapters.register_dumper(str, InvDumper)
+
+        assert sql.Literal("hello").as_string(conn) == f"'hello-inv'::\"{name}\""
+        cur = conn.execute(sql.SQL("select {}").format("hello"))
+        assert cur.fetchone()[0] == "hello-inv"
+
+        assert (
+            sql.Literal(["hello"]).as_string(conn) == f"'{{hello-inv}}'::\"{name}\"[]"
+        )
+        cur = conn.execute(sql.SQL("select {}").format(["hello"]))
+        assert cur.fetchone()[0] == ["hello-inv"]
+
 
 class TestSQL:
     def test_class(self):
@@ -356,9 +408,9 @@ class TestSQL:
         assert isinstance(sql.SQL("foo"), sql.SQL)
         assert isinstance(sql.SQL("foo"), sql.SQL)
         with pytest.raises(TypeError):
-            sql.SQL(10)
+            sql.SQL(10)  # type: ignore[arg-type]
         with pytest.raises(TypeError):
-            sql.SQL(dt.date(2016, 12, 31))
+            sql.SQL(dt.date(2016, 12, 31))  # type: ignore[arg-type]
 
     def test_repr(self, conn):
         assert repr(sql.SQL("foo")) == "SQL('foo')"
@@ -377,7 +429,7 @@ class TestSQL:
         assert obj.as_string(conn) == "foobar"
 
     def test_sum_inplace(self, conn):
-        obj = sql.SQL("foo")
+        obj = sql.SQL("f") + sql.SQL("oo")
         obj += sql.SQL("bar")
         assert isinstance(obj, sql.Composed)
         assert obj.as_string(conn) == "foobar"
@@ -395,9 +447,7 @@ class TestSQL:
         assert obj.as_string(conn) == '"foo", bar, 42'
 
         obj = sql.SQL(", ").join(
-            sql.Composed(
-                [sql.Identifier("foo"), sql.SQL("bar"), sql.Literal(42)]
-            )
+            sql.Composed([sql.Identifier("foo"), sql.SQL("bar"), sql.Literal(42)])
         )
         assert isinstance(obj, sql.Composed)
         assert obj.as_string(conn) == '"foo", bar, 42'
@@ -408,14 +458,12 @@ class TestSQL:
     def test_as_string(self, conn):
         assert sql.SQL("foo").as_string(conn) == "foo"
 
-    def test_as_bytes(self, conn):
-        assert sql.SQL("foo").as_bytes(conn) == b"foo"
+    @pytest.mark.parametrize("encoding", ["utf8", crdb_encoding("latin9")])
+    def test_as_bytes(self, conn, encoding):
+        if encoding:
+            conn.execute(f"set client_encoding to {encoding}")
 
-        conn.execute("set client_encoding to utf8")
-        assert sql.SQL(eur).as_bytes(conn) == eur.encode()
-
-        conn.execute("set client_encoding to latin9")
-        assert sql.SQL(eur).as_bytes(conn) == eur.encode("latin9")
+        assert sql.SQL(eur).as_bytes(conn) == eur.encode(encoding)
 
 
 class TestComposed:
@@ -424,9 +472,7 @@ class TestComposed:
 
     def test_repr(self):
         obj = sql.Composed([sql.Literal("foo"), sql.Identifier("b'ar")])
-        assert (
-            repr(obj) == """Composed([Literal('foo'), Identifier("b'ar")])"""
-        )
+        assert repr(obj) == """Composed([Literal('foo'), Identifier("b'ar")])"""
         assert str(obj) == repr(obj)
 
     def test_eq(self):
@@ -440,30 +486,30 @@ class TestComposed:
         obj = sql.Composed([sql.Literal("foo"), sql.Identifier("b'ar")])
         obj = obj.join(", ")
         assert isinstance(obj, sql.Composed)
-        assert noe(obj.as_string(conn)) == "'foo', \"b'ar\""
+        assert no_e(obj.as_string(conn)) == "'foo', \"b'ar\""
 
     def test_auto_literal(self, conn):
         obj = sql.Composed(["fo'o", dt.date(2020, 1, 1)])
         obj = obj.join(", ")
         assert isinstance(obj, sql.Composed)
-        assert noe(obj.as_string(conn)) == "'fo''o', '2020-01-01'"
+        assert no_e(obj.as_string(conn)) == "'fo''o', '2020-01-01'::date"
 
     def test_sum(self, conn):
         obj = sql.Composed([sql.SQL("foo ")])
         obj = obj + sql.Literal("bar")
         assert isinstance(obj, sql.Composed)
-        assert noe(obj.as_string(conn)) == "foo 'bar'"
+        assert no_e(obj.as_string(conn)) == "foo 'bar'"
 
     def test_sum_inplace(self, conn):
         obj = sql.Composed([sql.SQL("foo ")])
         obj += sql.Literal("bar")
         assert isinstance(obj, sql.Composed)
-        assert noe(obj.as_string(conn)) == "foo 'bar'"
+        assert no_e(obj.as_string(conn)) == "foo 'bar'"
 
         obj = sql.Composed([sql.SQL("foo ")])
         obj += sql.Composed([sql.Literal("bar")])
         assert isinstance(obj, sql.Composed)
-        assert noe(obj.as_string(conn)) == "foo 'bar'"
+        assert no_e(obj.as_string(conn)) == "foo 'bar'"
 
     def test_iter(self):
         obj = sql.Composed([sql.SQL("foo"), sql.SQL("bar")])
@@ -483,13 +529,11 @@ class TestComposed:
         obj = sql.Composed([sql.SQL("foo"), sql.SQL("bar")])
         assert obj.as_bytes(conn) == b"foobar"
 
+    @pytest.mark.parametrize("encoding", ["utf8", crdb_encoding("latin9")])
+    def test_as_bytes_encoding(self, conn, encoding):
         obj = sql.Composed([sql.SQL("foo"), sql.SQL(eur)])
-
-        conn.execute("set client_encoding to utf8")
-        assert obj.as_bytes(conn) == ("foo" + eur).encode()
-
-        conn.execute("set client_encoding to latin9")
-        assert obj.as_bytes(conn) == ("foo" + eur).encode("latin9")
+        conn.execute(f"set client_encoding to {encoding}")
+        assert obj.as_bytes(conn) == ("foo" + eur).encode(encoding)
 
 
 class TestPlaceholder:
@@ -523,18 +567,18 @@ class TestPlaceholder:
     @pytest.mark.parametrize("format", PyFormat)
     def test_as_string(self, conn, format):
         ph = sql.Placeholder(format=format)
-        assert ph.as_string(conn) == f"%{format}"
+        assert ph.as_string(conn) == f"%{format.value}"
 
         ph = sql.Placeholder(name="foo", format=format)
-        assert ph.as_string(conn) == f"%(foo){format}"
+        assert ph.as_string(conn) == f"%(foo){format.value}"
 
     @pytest.mark.parametrize("format", PyFormat)
     def test_as_bytes(self, conn, format):
         ph = sql.Placeholder(format=format)
-        assert ph.as_bytes(conn) == f"%{format}".encode("ascii")
+        assert ph.as_bytes(conn) == f"%{format.value}".encode("ascii")
 
         ph = sql.Placeholder(name="foo", format=format)
-        assert ph.as_bytes(conn) == f"%(foo){format}".encode("ascii")
+        assert ph.as_bytes(conn) == f"%(foo){format.value}".encode("ascii")
 
 
 class TestValues:
@@ -547,7 +591,7 @@ class TestValues:
         assert sql.DEFAULT.as_string(conn) == "DEFAULT"
 
 
-def noe(s):
+def no_e(s):
     """Drop an eventual E from E'' quotes"""
     if isinstance(s, memoryview):
         s = bytes(s)
@@ -555,6 +599,6 @@ def noe(s):
     if isinstance(s, str):
         return re.sub(r"\bE'", "'", s)
     elif isinstance(s, bytes):
-        return re.sub(br"\bE'", "'", s)
+        return re.sub(rb"\bE'", b"'", s)
     else:
         raise TypeError(f"not dealing with {type(s).__name__}: {s}")
