@@ -1,6 +1,22 @@
+import sys
+import logging
+from threading import Thread, Event
+
 import pytest
 
-from psycopg import Connection, ProgrammingError, Rollback
+import psycopg
+from psycopg import Rollback
+from psycopg import errors as e
+
+# TODOCRDB: is this the expected behaviour?
+crdb_skip_external_observer = pytest.mark.crdb(
+    "skip", reason="deadlock on observer connection"
+)
+
+
+@pytest.fixture
+def conn(conn, pipeline):
+    return conn
 
 
 @pytest.fixture(autouse=True)
@@ -15,7 +31,7 @@ def create_test_table(svcconn):
 
 def insert_row(conn, value):
     sql = "INSERT INTO test_table VALUES (%s)"
-    if isinstance(conn, Connection):
+    if isinstance(conn, psycopg.Connection):
         conn.cursor().execute(sql, (value,))
     else:
 
@@ -29,7 +45,7 @@ def insert_row(conn, value):
 def inserted(conn):
     """Return the values inserted in the test table."""
     sql = "SELECT * FROM test_table"
-    if isinstance(conn, Connection):
+    if isinstance(conn, psycopg.Connection):
         rows = conn.cursor().execute(sql).fetchall()
         return set(v for (v,) in rows)
     else:
@@ -52,14 +68,26 @@ def in_transaction(conn):
         assert False, conn.pgconn.transaction_status
 
 
+def get_exc_info(exc):
+    """Return the exc info for an exception or a success if exc is None"""
+    if not exc:
+        return (None,) * 3
+    try:
+        raise exc
+    except exc:
+        return sys.exc_info()
+
+
 class ExpectedException(Exception):
     pass
 
 
-def test_basic(conn):
+def test_basic(conn, pipeline):
     """Basic use of transaction() to BEGIN and COMMIT a transaction."""
     assert not in_transaction(conn)
     with conn.transaction():
+        if pipeline:
+            pipeline.sync()
         assert in_transaction(conn)
     assert not in_transaction(conn)
 
@@ -89,11 +117,13 @@ def test_cant_reenter(conn):
             pass
 
 
-def test_begins_on_enter(conn):
+def test_begins_on_enter(conn, pipeline):
     """Transaction does not begin until __enter__() is called."""
     tx = conn.transaction()
     assert not in_transaction(conn)
     with tx:
+        if pipeline:
+            pipeline.sync()
         assert in_transaction(conn)
     assert not in_transaction(conn)
 
@@ -118,6 +148,51 @@ def test_rollback_on_exception_exit(conn):
     assert not inserted(conn)
 
 
+@pytest.mark.crdb_skip("pg_terminate_backend")
+def test_context_inerror_rollback_no_clobber(conn_cls, conn, pipeline, dsn, caplog):
+    if pipeline:
+        # Only 'conn' is possibly in pipeline mode, but the transaction and
+        # checks are on 'conn2'.
+        pytest.skip("not applicable")
+    caplog.set_level(logging.WARNING, logger="psycopg")
+
+    with pytest.raises(ZeroDivisionError):
+        with conn_cls.connect(dsn) as conn2:
+            with conn2.transaction():
+                conn2.execute("select 1")
+                conn.execute(
+                    "select pg_terminate_backend(%s::int)",
+                    [conn2.pgconn.backend_pid],
+                )
+                1 / 0
+
+    assert len(caplog.records) == 1
+    rec = caplog.records[0]
+    assert rec.levelno == logging.WARNING
+    assert "in rollback" in rec.message
+
+
+@pytest.mark.crdb_skip("copy")
+def test_context_active_rollback_no_clobber(conn_cls, dsn, caplog):
+    caplog.set_level(logging.WARNING, logger="psycopg")
+
+    conn = conn_cls.connect(dsn)
+    try:
+        with pytest.raises(ZeroDivisionError):
+            with conn.transaction():
+                conn.pgconn.exec_(b"copy (select generate_series(1, 10)) to stdout")
+                status = conn.info.transaction_status
+                assert status == conn.TransactionStatus.ACTIVE
+                1 / 0
+
+        assert len(caplog.records) == 1
+        rec = caplog.records[0]
+        assert rec.levelno == logging.WARNING
+        assert "in rollback" in rec.message
+    finally:
+        conn.close()
+
+
 def test_interaction_dbapi_transaction(conn):
     insert_row(conn, "foo")
 
@@ -144,11 +219,11 @@ def test_prohibits_use_of_commit_rollback_autocommit(conn):
     conn.rollback()
 
     with conn.transaction():
-        with pytest.raises(ProgrammingError):
+        with pytest.raises(e.ProgrammingError):
             conn.autocommit = False
-        with pytest.raises(ProgrammingError):
+        with pytest.raises(e.ProgrammingError):
             conn.commit()
-        with pytest.raises(ProgrammingError):
+        with pytest.raises(e.ProgrammingError):
             conn.rollback()
 
     conn.autocommit = False
@@ -211,7 +286,8 @@ def test_autocommit_off_but_no_tx_started_exception_exit(conn, svcconn):
     assert not inserted(svcconn)
 
 
-def test_autocommit_off_and_tx_in_progress_successful_exit(conn, svcconn):
+@crdb_skip_external_observer
+def test_autocommit_off_and_tx_in_progress_successful_exit(conn, pipeline, svcconn):
     """
     Scenario:
     * Connection has autocommit off but and a transaction is already in
@@ -225,6 +301,8 @@ def test_autocommit_off_and_tx_in_progress_successful_exit(conn, svcconn):
     """
     conn.autocommit = False
     insert_row(conn, "prior")
+    if pipeline:
+        pipeline.sync()
     assert in_transaction(conn)
     with conn.transaction():
         insert_row(conn, "new")
@@ -234,7 +312,8 @@ def test_autocommit_off_and_tx_in_progress_successful_exit(conn, svcconn):
     assert not inserted(svcconn)
 
 
-def test_autocommit_off_and_tx_in_progress_exception_exit(conn, svcconn):
+@crdb_skip_external_observer
+def test_autocommit_off_and_tx_in_progress_exception_exit(conn, pipeline, svcconn):
     """
     Scenario:
     * Connection has autocommit off but and a transaction is already in
@@ -249,6 +328,8 @@ def test_autocommit_off_and_tx_in_progress_exception_exit(conn, svcconn):
     """
     conn.autocommit = False
     insert_row(conn, "prior")
+    if pipeline:
+        pipeline.sync()
     assert in_transaction(conn)
     with pytest.raises(ExpectedException):
         with conn.transaction():
@@ -354,7 +435,7 @@ def test_named_savepoints_successful_exit(conn, commands):
     ...and exiting the context successfully will "commit" the same.
     """
     # Case 1
-    # Using Transaction explicitly becase conn.transaction() enters the contetx
+    # Using Transaction explicitly because conn.transaction() enters the contetx
     assert not commands
     with conn.transaction() as tx:
         assert commands.popall() == ["BEGIN"]
@@ -373,7 +454,7 @@ def test_named_savepoints_successful_exit(conn, commands):
 
     # Case 2
     with conn.transaction(savepoint_name="foo") as tx:
-        assert commands.popall() == ['BEGIN; SAVEPOINT "foo"']
+        assert commands.popall() == ["BEGIN", 'SAVEPOINT "foo"']
         assert tx.savepoint_name == "foo"
     assert commands.popall() == ["COMMIT"]
 
@@ -413,7 +494,7 @@ def test_named_savepoints_exception_exit(conn, commands):
     # Case 2
     with pytest.raises(ExpectedException):
         with conn.transaction(savepoint_name="foo") as tx:
-            assert commands.popall() == ['BEGIN; SAVEPOINT "foo"']
+            assert commands.popall() == ["BEGIN", 'SAVEPOINT "foo"']
             assert tx.savepoint_name == "foo"
             raise ExpectedException
     assert commands.popall() == ["ROLLBACK"]
@@ -426,7 +507,7 @@ def test_named_savepoints_exception_exit(conn, commands):
                 assert commands.popall() == ['SAVEPOINT "bar"']
                 assert tx.savepoint_name == "bar"
                 raise ExpectedException
-        assert commands.popall() == ['ROLLBACK TO "bar"; RELEASE "bar"']
+        assert commands.popall() == ['ROLLBACK TO "bar"', 'RELEASE "bar"']
     assert commands.popall() == ["COMMIT"]
 
     # Case 3 (with savepoint name auto-generated)
@@ -437,7 +518,10 @@ def test_named_savepoints_exception_exit(conn, commands):
                 assert commands.popall() == ['SAVEPOINT "_pg3_2"']
                 assert tx.savepoint_name == "_pg3_2"
                 raise ExpectedException
-        assert commands.popall() == ['ROLLBACK TO "_pg3_2"; RELEASE "_pg3_2"']
+        assert commands.popall() == [
+            'ROLLBACK TO "_pg3_2"',
+            'RELEASE "_pg3_2"',
+        ]
     assert commands.popall() == ["COMMIT"]
 
 
@@ -480,15 +564,6 @@ def test_named_savepoints_with_repeated_names_works(conn):
             assert inserted(conn) == {"tx1"}
         assert inserted(conn) == {"tx1"}
 
-    # Will not (always) catch out-of-order exits
-    with conn.transaction(force_rollback=True):
-        tx1 = conn.transaction("s1")
-        tx2 = conn.transaction("s1")
-        tx1.__enter__()
-        tx2.__enter__()
-        tx1.__exit__(None, None, None)
-        tx2.__exit__(None, None, None)
-
 
 def test_force_rollback_successful_exit(conn, svcconn):
     """
@@ -514,6 +589,7 @@ def test_force_rollback_exception_exit(conn, svcconn):
     assert not inserted(svcconn)
 
 
+@crdb_skip_external_observer
 def test_explicit_rollback_discards_changes(conn, svcconn):
     """
     Raising a Rollback exception in the middle of a block exits the block and
@@ -546,6 +622,7 @@ def test_explicit_rollback_discards_changes(conn, svcconn):
     assert_no_rows()
 
 
+@crdb_skip_external_observer
 def test_explicit_rollback_outer_tx_unaffected(conn, svcconn):
     """
     Raising a Rollback exception in the middle of a block does not impact an
@@ -577,6 +654,7 @@ def test_explicit_rollback_of_outer_transaction(conn):
     assert not inserted(conn)
 
 
+@crdb_skip_external_observer
 def test_explicit_rollback_of_enclosing_tx_outer_tx_unaffected(conn, svcconn):
     """
     Rolling-back an enclosing transaction does not impact an outer transaction.
@@ -596,14 +674,123 @@ def test_explicit_rollback_of_enclosing_tx_outer_tx_unaffected(conn, svcconn):
     assert inserted(svcconn) == {"outer-before", "outer-after"}
 
 
-def test_str(conn):
+def test_str(conn, pipeline):
     with conn.transaction() as tx:
-        assert "[INTRANS]" in str(tx)
+        if pipeline:
+            assert "[INTRANS, pipeline=ON]" in str(tx)
+        else:
+            assert "[INTRANS]" in str(tx)
         assert "(active)" in str(tx)
         assert "'" not in str(tx)
         with conn.transaction("wat") as tx2:
-            assert "[INTRANS]" in str(tx2)
+            if pipeline:
+                assert "[INTRANS, pipeline=ON]" in str(tx2)
+            else:
+                assert "[INTRANS]" in str(tx2)
             assert "'wat'" in str(tx2)
 
-    assert "[IDLE]" in str(tx)
+    if pipeline:
+        assert "[IDLE, pipeline=ON]" in str(tx)
+    else:
+        assert "[IDLE]" in str(tx)
     assert "(terminated)" in str(tx)
+
+    with pytest.raises(ZeroDivisionError):
+        with conn.transaction() as tx:
+            1 / 0
+
+    assert "(terminated)" in str(tx)
+
+
+@pytest.mark.parametrize("exit_error", [None, ZeroDivisionError, Rollback])
+def test_out_of_order_exit(conn, exit_error):
+    conn.autocommit = True
+
+    t1 = conn.transaction()
+    t1.__enter__()
+
+    t2 = conn.transaction()
+    t2.__enter__()
+
+    with pytest.raises(e.ProgrammingError):
+        t1.__exit__(*get_exc_info(exit_error))
+
+    with pytest.raises(e.ProgrammingError):
+        t2.__exit__(*get_exc_info(exit_error))
+
+
+@pytest.mark.parametrize("exit_error", [None, ZeroDivisionError, Rollback])
+def test_out_of_order_implicit_begin(conn, exit_error):
+    conn.execute("select 1")
+
+    t1 = conn.transaction()
+    t1.__enter__()
+
+    t2 = conn.transaction()
+    t2.__enter__()
+
+    with pytest.raises(e.ProgrammingError):
+        t1.__exit__(*get_exc_info(exit_error))
+
+    with pytest.raises(e.ProgrammingError):
+        t2.__exit__(*get_exc_info(exit_error))
+
+
+@pytest.mark.parametrize("exit_error", [None, ZeroDivisionError, Rollback])
+def test_out_of_order_exit_same_name(conn, exit_error):
+    conn.autocommit = True
+
+    t1 = conn.transaction("save")
+    t1.__enter__()
+    t2 = conn.transaction("save")
+    t2.__enter__()
+
+    with pytest.raises(e.ProgrammingError):
+        t1.__exit__(*get_exc_info(exit_error))
+
+    with pytest.raises(e.ProgrammingError):
+        t2.__exit__(*get_exc_info(exit_error))
+
+
+@pytest.mark.parametrize("what", ["commit", "rollback", "error"])
+def test_concurrency(conn, what):
+    conn.autocommit = True
+
+    evs = [Event() for i in range(3)]
+
+    def worker(unlock, wait_on):
+        with pytest.raises(e.ProgrammingError) as ex:
+            with conn.transaction():
+                unlock.set()
+                wait_on.wait()
+                conn.execute("select 1")
+
+                if what == "error":
+                    1 / 0
+                elif what == "rollback":
+                    raise Rollback()
+                else:
+                    assert what == "commit"
+
+        if what == "error":
+            assert "transaction rollback" in str(ex.value)
+            assert isinstance(ex.value.__context__, ZeroDivisionError)
+        elif what == "rollback":
+            assert "transaction rollback" in str(ex.value)
+            assert isinstance(ex.value.__context__, Rollback)
+        else:
+            assert "transaction commit" in str(ex.value)
+
+    # Start a first transaction in a thread
+    t1 = Thread(target=worker, kwargs={"unlock": evs[0], "wait_on": evs[1]})
+    t1.start()
+    evs[0].wait()
+
+    # Start a nested transaction in a thread
+    t2 = Thread(target=worker, kwargs={"unlock": evs[1], "wait_on": evs[2]})
+    t2.start()
+
+    # Terminate the first transaction before the second does
+    t1.join()
+    evs[2].set()
+    t2.join()
