@@ -1,18 +1,30 @@
+import asyncio
+import logging
+
 import pytest
 
-from psycopg import ProgrammingError, Rollback
+from psycopg import Rollback
+from psycopg import errors as e
+from psycopg._compat import create_task
 
-from .test_transaction import in_transaction, insert_row, inserted
-from .test_transaction import ExpectedException
+from .test_transaction import in_transaction, insert_row, inserted, get_exc_info
+from .test_transaction import ExpectedException, crdb_skip_external_observer
 from .test_transaction import create_test_table  # noqa  # autouse fixture
 
 pytestmark = pytest.mark.asyncio
 
 
-async def test_basic(aconn):
+@pytest.fixture
+async def aconn(aconn, apipeline):
+    return aconn
+
+
+async def test_basic(aconn, apipeline):
     """Basic use of transaction() to BEGIN and COMMIT a transaction."""
     assert not in_transaction(aconn)
     async with aconn.transaction():
+        if apipeline:
+            await apipeline.sync()
         assert in_transaction(aconn)
     assert not in_transaction(aconn)
 
@@ -42,11 +54,13 @@ async def test_cant_reenter(aconn):
             pass
 
 
-async def test_begins_on_enter(aconn):
+async def test_begins_on_enter(aconn, apipeline):
     """Transaction does not begin until __enter__() is called."""
     tx = aconn.transaction()
     assert not in_transaction(aconn)
     async with tx:
+        if apipeline:
+            await apipeline.sync()
         assert in_transaction(aconn)
     assert not in_transaction(aconn)
 
@@ -71,6 +85,53 @@ async def test_rollback_on_exception_exit(aconn):
     assert not await inserted(aconn)
 
 
+@pytest.mark.crdb_skip("pg_terminate_backend")
+async def test_context_inerror_rollback_no_clobber(
+    aconn_cls, aconn, apipeline, dsn, caplog
+):
+    if apipeline:
+        # Only 'aconn' is possibly in pipeline mode, but the transaction and
+        # checks are on 'conn2'.
+        pytest.skip("not applicable")
+    caplog.set_level(logging.WARNING, logger="psycopg")
+
+    with pytest.raises(ZeroDivisionError):
+        async with await aconn_cls.connect(dsn) as conn2:
+            async with conn2.transaction():
+                await conn2.execute("select 1")
+                await aconn.execute(
+                    "select pg_terminate_backend(%s::int)",
+                    [conn2.pgconn.backend_pid],
+                )
+                1 / 0
+
+    assert len(caplog.records) == 1
+    rec = caplog.records[0]
+    assert rec.levelno == logging.WARNING
+    assert "in rollback" in rec.message
+
+
+@pytest.mark.crdb_skip("copy")
+async def test_context_active_rollback_no_clobber(aconn_cls, dsn, caplog):
+    caplog.set_level(logging.WARNING, logger="psycopg")
+
+    conn = await aconn_cls.connect(dsn)
+    try:
+        with pytest.raises(ZeroDivisionError):
+            async with conn.transaction():
+                conn.pgconn.exec_(b"copy (select generate_series(1, 10)) to stdout")
+                status = conn.info.transaction_status
+                assert status == conn.TransactionStatus.ACTIVE
+                1 / 0
+
+        assert len(caplog.records) == 1
+        rec = caplog.records[0]
+        assert rec.levelno == logging.WARNING
+        assert "in rollback" in rec.message
+    finally:
+        await conn.close()
+
+
 async def test_interaction_dbapi_transaction(aconn):
     await insert_row(aconn, "foo")
 
@@ -97,11 +158,11 @@ async def test_prohibits_use_of_commit_rollback_autocommit(aconn):
     await aconn.rollback()
 
     async with aconn.transaction():
-        with pytest.raises(ProgrammingError):
+        with pytest.raises(e.ProgrammingError):
             await aconn.set_autocommit(False)
-        with pytest.raises(ProgrammingError):
+        with pytest.raises(e.ProgrammingError):
             await aconn.commit()
-        with pytest.raises(ProgrammingError):
+        with pytest.raises(e.ProgrammingError):
             await aconn.rollback()
 
     await aconn.set_autocommit(False)
@@ -120,9 +181,7 @@ async def test_preserves_autocommit(aconn, autocommit):
     assert aconn.autocommit is autocommit
 
 
-async def test_autocommit_off_but_no_tx_started_successful_exit(
-    aconn, svcconn
-):
+async def test_autocommit_off_but_no_tx_started_successful_exit(aconn, svcconn):
     """
     Scenario:
     * Connection has autocommit off but no transaction has been initiated
@@ -166,8 +225,9 @@ async def test_autocommit_off_but_no_tx_started_exception_exit(aconn, svcconn):
     assert not inserted(svcconn)
 
 
+@crdb_skip_external_observer
 async def test_autocommit_off_and_tx_in_progress_successful_exit(
-    aconn, svcconn
+    aconn, apipeline, svcconn
 ):
     """
     Scenario:
@@ -182,6 +242,8 @@ async def test_autocommit_off_and_tx_in_progress_successful_exit(
     """
     await aconn.set_autocommit(False)
     await insert_row(aconn, "prior")
+    if apipeline:
+        await apipeline.sync()
     assert in_transaction(aconn)
     async with aconn.transaction():
         await insert_row(aconn, "new")
@@ -191,8 +253,9 @@ async def test_autocommit_off_and_tx_in_progress_successful_exit(
     assert not inserted(svcconn)
 
 
+@crdb_skip_external_observer
 async def test_autocommit_off_and_tx_in_progress_exception_exit(
-    aconn, svcconn
+    aconn, apipeline, svcconn
 ):
     """
     Scenario:
@@ -208,6 +271,8 @@ async def test_autocommit_off_and_tx_in_progress_exception_exit(
     """
     await aconn.set_autocommit(False)
     await insert_row(aconn, "prior")
+    if apipeline:
+        await apipeline.sync()
     assert in_transaction(aconn)
     with pytest.raises(ExpectedException):
         async with aconn.transaction():
@@ -263,9 +328,7 @@ async def test_nested_all_changes_discarded_on_inner_exception(aconn, svcconn):
     assert not inserted(svcconn)
 
 
-async def test_nested_inner_scope_exception_handled_in_outer_scope(
-    aconn, svcconn
-):
+async def test_nested_inner_scope_exception_handled_in_outer_scope(aconn, svcconn):
     """
     An exception escaping the inner transaction context causes changes made
     within that inner context to be discarded, but the error can then be
@@ -317,7 +380,7 @@ async def test_named_savepoints_successful_exit(aconn, acommands):
     commands = acommands
 
     # Case 1
-    # Using Transaction explicitly becase conn.transaction() enters the contetx
+    # Using Transaction explicitly because conn.transaction() enters the contetx
     async with aconn.transaction() as tx:
         assert commands.popall() == ["BEGIN"]
         assert not tx.savepoint_name
@@ -336,7 +399,7 @@ async def test_named_savepoints_successful_exit(aconn, acommands):
 
     # Case 2
     async with aconn.transaction(savepoint_name="foo") as tx:
-        assert commands.popall() == ['BEGIN; SAVEPOINT "foo"']
+        assert commands.popall() == ["BEGIN", 'SAVEPOINT "foo"']
         assert tx.savepoint_name == "foo"
     assert commands.popall() == ["COMMIT"]
 
@@ -378,7 +441,7 @@ async def test_named_savepoints_exception_exit(aconn, acommands):
     # Case 2
     with pytest.raises(ExpectedException):
         async with aconn.transaction(savepoint_name="foo") as tx:
-            assert commands.popall() == ['BEGIN; SAVEPOINT "foo"']
+            assert commands.popall() == ["BEGIN", 'SAVEPOINT "foo"']
             assert tx.savepoint_name == "foo"
             raise ExpectedException
     assert commands.popall() == ["ROLLBACK"]
@@ -391,7 +454,7 @@ async def test_named_savepoints_exception_exit(aconn, acommands):
                 assert commands.popall() == ['SAVEPOINT "bar"']
                 assert tx.savepoint_name == "bar"
                 raise ExpectedException
-        assert commands.popall() == ['ROLLBACK TO "bar"; RELEASE "bar"']
+        assert commands.popall() == ['ROLLBACK TO "bar"', 'RELEASE "bar"']
     assert commands.popall() == ["COMMIT"]
 
     # Case 3 (with savepoint name auto-generated)
@@ -402,7 +465,10 @@ async def test_named_savepoints_exception_exit(aconn, acommands):
                 assert commands.popall() == ['SAVEPOINT "_pg3_2"']
                 assert tx.savepoint_name == "_pg3_2"
                 raise ExpectedException
-        assert commands.popall() == ['ROLLBACK TO "_pg3_2"; RELEASE "_pg3_2"']
+        assert commands.popall() == [
+            'ROLLBACK TO "_pg3_2"',
+            'RELEASE "_pg3_2"',
+        ]
     assert commands.popall() == ["COMMIT"]
 
 
@@ -445,15 +511,6 @@ async def test_named_savepoints_with_repeated_names_works(aconn):
             assert await inserted(aconn) == {"tx1"}
         assert await inserted(aconn) == {"tx1"}
 
-    # Will not (always) catch out-of-order exits
-    async with aconn.transaction(force_rollback=True):
-        tx1 = aconn.transaction("s1")
-        tx2 = aconn.transaction("s1")
-        await tx1.__aenter__()
-        await tx2.__aenter__()
-        await tx1.__aexit__(None, None, None)
-        await tx2.__aexit__(None, None, None)
-
 
 async def test_force_rollback_successful_exit(aconn, svcconn):
     """
@@ -479,6 +536,7 @@ async def test_force_rollback_exception_exit(aconn, svcconn):
     assert not inserted(svcconn)
 
 
+@crdb_skip_external_observer
 async def test_explicit_rollback_discards_changes(aconn, svcconn):
     """
     Raising a Rollback exception in the middle of a block exits the block and
@@ -511,6 +569,7 @@ async def test_explicit_rollback_discards_changes(aconn, svcconn):
     await assert_no_rows()
 
 
+@crdb_skip_external_observer
 async def test_explicit_rollback_outer_tx_unaffected(aconn, svcconn):
     """
     Raising a Rollback exception in the middle of a block does not impact an
@@ -542,9 +601,8 @@ async def test_explicit_rollback_of_outer_transaction(aconn):
     assert not await inserted(aconn)
 
 
-async def test_explicit_rollback_of_enclosing_tx_outer_tx_unaffected(
-    aconn, svcconn
-):
+@crdb_skip_external_observer
+async def test_explicit_rollback_of_enclosing_tx_outer_tx_unaffected(aconn, svcconn):
     """
     Rolling-back an enclosing transaction does not impact an outer transaction.
     """
@@ -563,14 +621,123 @@ async def test_explicit_rollback_of_enclosing_tx_outer_tx_unaffected(
     assert inserted(svcconn) == {"outer-before", "outer-after"}
 
 
-async def test_str(aconn):
+async def test_str(aconn, apipeline):
     async with aconn.transaction() as tx:
-        assert "[INTRANS]" in str(tx)
+        if apipeline:
+            assert "[INTRANS]" not in str(tx)
+            await apipeline.sync()
+            assert "[INTRANS, pipeline=ON]" in str(tx)
+        else:
+            assert "[INTRANS]" in str(tx)
         assert "(active)" in str(tx)
         assert "'" not in str(tx)
         async with aconn.transaction("wat") as tx2:
-            assert "[INTRANS]" in str(tx2)
+            if apipeline:
+                assert "[INTRANS, pipeline=ON]" in str(tx2)
+            else:
+                assert "[INTRANS]" in str(tx2)
             assert "'wat'" in str(tx2)
 
-    assert "[IDLE]" in str(tx)
+    if apipeline:
+        assert "[IDLE, pipeline=ON]" in str(tx)
+    else:
+        assert "[IDLE]" in str(tx)
     assert "(terminated)" in str(tx)
+
+    with pytest.raises(ZeroDivisionError):
+        async with aconn.transaction() as tx:
+            1 / 0
+
+    assert "(terminated)" in str(tx)
+
+
+@pytest.mark.parametrize("exit_error", [None, ZeroDivisionError, Rollback])
+async def test_out_of_order_exit(aconn, exit_error):
+    await aconn.set_autocommit(True)
+
+    t1 = aconn.transaction()
+    await t1.__aenter__()
+
+    t2 = aconn.transaction()
+    await t2.__aenter__()
+
+    with pytest.raises(e.ProgrammingError):
+        await t1.__aexit__(*get_exc_info(exit_error))
+
+    with pytest.raises(e.ProgrammingError):
+        await t2.__aexit__(*get_exc_info(exit_error))
+
+
+@pytest.mark.parametrize("exit_error", [None, ZeroDivisionError, Rollback])
+async def test_out_of_order_implicit_begin(aconn, exit_error):
+    await aconn.execute("select 1")
+
+    t1 = aconn.transaction()
+    await t1.__aenter__()
+
+    t2 = aconn.transaction()
+    await t2.__aenter__()
+
+    with pytest.raises(e.ProgrammingError):
+        await t1.__aexit__(*get_exc_info(exit_error))
+
+    with pytest.raises(e.ProgrammingError):
+        await t2.__aexit__(*get_exc_info(exit_error))
+
+
+@pytest.mark.parametrize("exit_error", [None, ZeroDivisionError, Rollback])
+async def test_out_of_order_exit_same_name(aconn, exit_error):
+    await aconn.set_autocommit(True)
+
+    t1 = aconn.transaction("save")
+    await t1.__aenter__()
+    t2 = aconn.transaction("save")
+    await t2.__aenter__()
+
+    with pytest.raises(e.ProgrammingError):
+        await t1.__aexit__(*get_exc_info(exit_error))
+
+    with pytest.raises(e.ProgrammingError):
+        await t2.__aexit__(*get_exc_info(exit_error))
+
+
+@pytest.mark.parametrize("what", ["commit", "rollback", "error"])
+async def test_concurrency(aconn, what):
+    await aconn.set_autocommit(True)
+
+    evs = [asyncio.Event() for i in range(3)]
+
+    async def worker(unlock, wait_on):
+        with pytest.raises(e.ProgrammingError) as ex:
+            async with aconn.transaction():
+                unlock.set()
+                await wait_on.wait()
+                await aconn.execute("select 1")
+
+                if what == "error":
+                    1 / 0
+                elif what == "rollback":
+                    raise Rollback()
+                else:
+                    assert what == "commit"
+
+        if what == "error":
+            assert "transaction rollback" in str(ex.value)
+            assert isinstance(ex.value.__context__, ZeroDivisionError)
+        elif what == "rollback":
+            assert "transaction rollback" in str(ex.value)
+            assert isinstance(ex.value.__context__, Rollback)
+        else:
+            assert "transaction commit" in str(ex.value)
+
+    # Start a first transaction in a task
+    t1 = create_task(worker(unlock=evs[0], wait_on=evs[1]))
+    await evs[0].wait()
+
+    # Start a nested transaction in a task
+    t2 = create_task(worker(unlock=evs[1], wait_on=evs[2]))
+
+    # Terminate the first transaction before the second does
+    await asyncio.gather(t1)
+    evs[2].set()
+    await asyncio.gather(t2)
