@@ -1,4 +1,4 @@
-import select
+import select  # noqa: used in pytest.mark.skipif
 import socket
 import sys
 
@@ -9,23 +9,25 @@ from psycopg import waiting
 from psycopg import generators
 from psycopg.pq import ConnStatus, ExecStatus
 
-
-hasepoll = hasattr(select, "epoll")
-skip_no_epoll = pytest.mark.skipif(not hasepoll, reason="epoll not available")
-
-timeouts = [
-    {},
-    {"timeout": None},
-    {"timeout": 0},
-    {"timeout": 0.2},
-    {"timeout": 10},
-]
-
-
 skip_if_not_linux = pytest.mark.skipif(
     not sys.platform.startswith("linux"), reason="non-Linux platform"
 )
 
+waitfns = [
+    "wait",
+    "wait_selector",
+    pytest.param(
+        "wait_select", marks=pytest.mark.skipif("not hasattr(select, 'select')")
+    ),
+    pytest.param(
+        "wait_epoll", marks=pytest.mark.skipif("not hasattr(select, 'epoll')")
+    ),
+    pytest.param("wait_c", marks=pytest.mark.skipif("not psycopg._cmodule._psycopg")),
+]
+
+timeouts = [pytest.param({}, id="blank")]
+timeouts += [pytest.param({"timeout": x}, id=str(x)) for x in [None, 0, 0.2, 10]]
+
 
 @pytest.mark.parametrize("timeout", timeouts)
 def test_wait_conn(dsn, timeout):
@@ -40,28 +42,12 @@ def test_wait_conn_bad(dsn):
         waiting.wait_conn(gen)
 
 
-@pytest.mark.parametrize("timeout", timeouts)
-def test_wait(pgconn, timeout):
-    pgconn.send_query(b"select 1")
-    gen = generators.execute(pgconn)
-    (res,) = waiting.wait(gen, pgconn.socket, **timeout)
-    assert res.status == ExecStatus.TUPLES_OK
-
-
-waits_and_ids = [
-    (waiting.wait, "wait"),
-    (waiting.wait_selector, "wait_selector"),
-]
-if hasepoll:
-    waits_and_ids.append((waiting.wait_epoll, "wait_epoll"))
-
-waits, wids = list(zip(*waits_and_ids))
-
-
-@pytest.mark.parametrize("waitfn", waits, ids=wids)
+@pytest.mark.parametrize("waitfn", waitfns)
 @pytest.mark.parametrize("wait, ready", zip(waiting.Wait, waiting.Ready))
 @skip_if_not_linux
 def test_wait_ready(waitfn, wait, ready):
+    waitfn = getattr(waiting, waitfn)
+
     def gen():
         r = yield wait
         return r
@@ -71,37 +57,60 @@ def test_wait_ready(waitfn, wait, ready):
     assert r & ready
 
 
+@pytest.mark.parametrize("waitfn", waitfns)
 @pytest.mark.parametrize("timeout", timeouts)
-def test_wait_selector(pgconn, timeout):
+def test_wait(pgconn, waitfn, timeout):
+    waitfn = getattr(waiting, waitfn)
+
     pgconn.send_query(b"select 1")
     gen = generators.execute(pgconn)
-    (res,) = waiting.wait_selector(gen, pgconn.socket, **timeout)
+    (res,) = waitfn(gen, pgconn.socket, **timeout)
     assert res.status == ExecStatus.TUPLES_OK
 
 
-def test_wait_selector_bad(pgconn):
+@pytest.mark.parametrize("waitfn", waitfns)
+def test_wait_bad(pgconn, waitfn):
+    waitfn = getattr(waiting, waitfn)
+
     pgconn.send_query(b"select 1")
     gen = generators.execute(pgconn)
     pgconn.finish()
     with pytest.raises(psycopg.OperationalError):
-        waiting.wait_selector(gen, pgconn.socket)
+        waitfn(gen, pgconn.socket)
 
 
-@skip_no_epoll
-@pytest.mark.parametrize("timeout", timeouts)
-def test_wait_epoll(pgconn, timeout):
-    pgconn.send_query(b"select 1")
-    gen = generators.execute(pgconn)
-    (res,) = waiting.wait_epoll(gen, pgconn.socket, **timeout)
-    assert res.status == ExecStatus.TUPLES_OK
+@pytest.mark.slow
+@pytest.mark.skipif(
+    "sys.platform == 'win32'", reason="win32 works ok, but FDs are mysterious"
+)
+@pytest.mark.parametrize("waitfn", waitfns)
+def test_wait_large_fd(dsn, waitfn):
+    waitfn = getattr(waiting, waitfn)
 
+    files = []
+    try:
+        try:
+            for i in range(1100):
+                files.append(open(__file__))
+        except OSError:
+            pytest.skip("can't open the number of files needed for the test")
 
-@skip_no_epoll
-def test_wait_epoll_bad(pgconn):
-    pgconn.send_query(b"select 1")
-    gen = generators.execute(pgconn)
-    (res,) = waiting.wait_epoll(gen, pgconn.socket)
-    assert res.status == ExecStatus.TUPLES_OK
+        pgconn = psycopg.pq.PGconn.connect(dsn.encode())
+        try:
+            assert pgconn.socket > 1024
+            pgconn.send_query(b"select 1")
+            gen = generators.execute(pgconn)
+            if waitfn is waiting.wait_select:
+                with pytest.raises(ValueError):
+                    waitfn(gen, pgconn.socket)
+            else:
+                (res,) = waitfn(gen, pgconn.socket)
+                assert res.status == ExecStatus.TUPLES_OK
+        finally:
+            pgconn.finish()
+    finally:
+        for f in files:
+            f.close()
 
 
 @pytest.mark.parametrize("timeout", timeouts)
@@ -119,14 +128,6 @@ async def test_wait_conn_async_bad(dsn):
         await waiting.wait_conn_async(gen)
 
 
-@pytest.mark.asyncio
-async def test_wait_async(pgconn):
-    pgconn.send_query(b"select 1")
-    gen = generators.execute(pgconn)
-    (res,) = await waiting.wait_async(gen, pgconn.socket)
-    assert res.status == ExecStatus.TUPLES_OK
-
-
 @pytest.mark.asyncio
 @pytest.mark.parametrize("wait, ready", zip(waiting.Wait, waiting.Ready))
 @skip_if_not_linux
@@ -140,6 +141,14 @@ async def test_wait_ready_async(wait, ready):
     assert r & ready
 
 
+@pytest.mark.asyncio
+async def test_wait_async(pgconn):
+    pgconn.send_query(b"select 1")
+    gen = generators.execute(pgconn)
+    (res,) = await waiting.wait_async(gen, pgconn.socket)
+    assert res.status == ExecStatus.TUPLES_OK
+
+
 @pytest.mark.asyncio
 async def test_wait_async_bad(pgconn):
     pgconn.send_query(b"select 1")
