@@ -1,4 +1,3 @@
-import sys
 import datetime as dt
 
 import pytest
@@ -6,6 +5,26 @@ import pytest
 from psycopg import DataError, pq, sql
 from psycopg.adapt import PyFormat
 
+crdb_skip_datestyle = pytest.mark.crdb("skip", reason="set datestyle/intervalstyle")
+crdb_skip_negative_interval = pytest.mark.crdb("skip", reason="negative interval")
+crdb_skip_invalid_tz = pytest.mark.crdb(
+    "skip", reason="crdb doesn't allow invalid timezones"
+)
+
+datestyles_in = [
+    pytest.param(datestyle, marks=crdb_skip_datestyle)
+    for datestyle in ["DMY", "MDY", "YMD"]
+]
+datestyles_out = [
+    pytest.param(datestyle, marks=crdb_skip_datestyle)
+    for datestyle in ["ISO", "Postgres", "SQL", "German"]
+]
+
+intervalstyles = [
+    pytest.param(datestyle, marks=crdb_skip_datestyle)
+    for datestyle in ["sql_standard", "postgres", "postgres_verbose", "iso_8601"]
+]
+
 
 class TestDate:
     @pytest.mark.parametrize(
@@ -23,7 +42,7 @@ class TestDate:
     def test_dump_date(self, conn, val, expr, fmt_in):
         val = as_date(val)
         cur = conn.cursor()
-        cur.execute(f"select '{expr}'::date = %{fmt_in}", (val,))
+        cur.execute(f"select '{expr}'::date = %{fmt_in.value}", (val,))
         assert cur.fetchone()[0] is True
 
         cur.execute(
@@ -34,10 +53,10 @@ class TestDate:
         )
         assert cur.fetchone()[0] is True
 
-    @pytest.mark.parametrize("datestyle_in", ["DMY", "MDY", "YMD"])
+    @pytest.mark.parametrize("datestyle_in", datestyles_in)
     def test_dump_date_datestyle(self, conn, datestyle_in):
         cur = conn.cursor(binary=False)
-        cur.execute(f"set datestyle = ISO, {datestyle_in}")
+        cur.execute(f"set datestyle = ISO,{datestyle_in}")
         cur.execute("select 'epoch'::date + 1 = %t", (dt.date(1970, 1, 2),))
         assert cur.fetchone()[0] is True
 
@@ -58,9 +77,7 @@ class TestDate:
         cur.execute(f"select '{expr}'::date")
         assert cur.fetchone()[0] == as_date(val)
 
-    @pytest.mark.parametrize(
-        "datestyle_out", ["ISO", "Postgres", "SQL", "German"]
-    )
+    @pytest.mark.parametrize("datestyle_out", datestyles_out)
     def test_load_date_datestyle(self, conn, datestyle_out):
         cur = conn.cursor(binary=False)
         cur.execute(f"set datestyle = {datestyle_out}, YMD")
@@ -68,27 +85,77 @@ class TestDate:
         assert cur.fetchone()[0] == dt.date(2000, 1, 2)
 
     @pytest.mark.parametrize("val", ["min", "max"])
-    @pytest.mark.parametrize(
-        "datestyle_out", ["ISO", "Postgres", "SQL", "German"]
-    )
+    @pytest.mark.parametrize("datestyle_out", datestyles_out)
     def test_load_date_overflow(self, conn, val, datestyle_out):
         cur = conn.cursor(binary=False)
         cur.execute(f"set datestyle = {datestyle_out}, YMD")
-        cur.execute(
-            "select %t + %s::int", (as_date(val), -1 if val == "min" else 1)
-        )
+        cur.execute("select %t + %s::int", (as_date(val), -1 if val == "min" else 1))
         with pytest.raises(DataError):
             cur.fetchone()[0]
 
     @pytest.mark.parametrize("val", ["min", "max"])
     def test_load_date_overflow_binary(self, conn, val):
         cur = conn.cursor(binary=True)
-        cur.execute(
-            "select %s + %s::int", (as_date(val), -1 if val == "min" else 1)
-        )
+        cur.execute("select %s + %s::int", (as_date(val), -1 if val == "min" else 1))
         with pytest.raises(DataError):
             cur.fetchone()[0]
 
+    overflow_samples = [
+        ("-infinity", "date too small"),
+        ("1000-01-01 BC", "date too small"),
+        ("10000-01-01", "date too large"),
+        ("infinity", "date too large"),
+    ]
+
+    @pytest.mark.parametrize("datestyle_out", datestyles_out)
+    @pytest.mark.parametrize("val, msg", overflow_samples)
+    def test_load_overflow_message(self, conn, datestyle_out, val, msg):
+        cur = conn.cursor()
+        cur.execute(f"set datestyle = {datestyle_out}, YMD")
+        cur.execute("select %s::date", (val,))
+        with pytest.raises(DataError) as excinfo:
+            cur.fetchone()[0]
+        assert msg in str(excinfo.value)
+
+    @pytest.mark.parametrize("val, msg", overflow_samples)
+    def test_load_overflow_message_binary(self, conn, val, msg):
+        cur = conn.cursor(binary=True)
+        cur.execute("select %s::date", (val,))
+        with pytest.raises(DataError) as excinfo:
+            cur.fetchone()[0]
+        assert msg in str(excinfo.value)
+
+    def test_infinity_date_example(self, conn):
+        # NOTE: this is an example in the docs. Make sure it doesn't regress when
+        # adding binary datetime adapters
+        from datetime import date
+        from psycopg.types.datetime import DateLoader, DateDumper
+
+        class InfDateDumper(DateDumper):
+            def dump(self, obj):
+                if obj == date.max:
+                    return b"infinity"
+                else:
+                    return super().dump(obj)
+
+        class InfDateLoader(DateLoader):
+            def load(self, data):
+                if data == b"infinity":
+                    return date.max
+                else:
+                    return super().load(data)
+
+        cur = conn.cursor()
+        cur.adapters.register_dumper(date, InfDateDumper)
+        cur.adapters.register_loader("date", InfDateLoader)
+
+        rec = cur.execute(
+            "SELECT %s::text, %s::text", [date(2020, 12, 31), date.max]
+        ).fetchone()
+        assert rec == ("2020-12-31", "infinity")
+        rec = cur.execute("select '2020-12-31'::date, 'infinity'::date").fetchone()
+        assert rec == (date(2020, 12, 31), date(9999, 12, 31))
+
 
 class TestDatetime:
     @pytest.mark.parametrize(
@@ -113,19 +180,19 @@ class TestDatetime:
     def test_dump_datetime(self, conn, val, expr, fmt_in):
         cur = conn.cursor()
         cur.execute("set timezone to '+02:00'")
-        cur.execute(f"select %{fmt_in}", (as_dt(val),))
-        cur.execute(f"select '{expr}'::timestamp = %{fmt_in}", (as_dt(val),))
+        cur.execute(f"select %{fmt_in.value}", (as_dt(val),))
+        cur.execute(f"select '{expr}'::timestamp = %{fmt_in.value}", (as_dt(val),))
         cur.execute(
             f"""
-            select '{expr}'::timestamp = %(val){fmt_in},
-            '{expr}', %(val){fmt_in}::text
+            select '{expr}'::timestamp = %(val){fmt_in.value},
+            '{expr}', %(val){fmt_in.value}::text
             """,
             {"val": as_dt(val)},
         )
         ok, want, got = cur.fetchone()
         assert ok, (want, got)
 
-    @pytest.mark.parametrize("datestyle_in", ["DMY", "MDY", "YMD"])
+    @pytest.mark.parametrize("datestyle_in", datestyles_in)
     def test_dump_datetime_datestyle(self, conn, datestyle_in):
         cur = conn.cursor(binary=False)
         cur.execute(f"set datestyle = ISO, {datestyle_in}")
@@ -148,10 +215,8 @@ class TestDatetime:
     ]
 
     @pytest.mark.parametrize("val, expr", load_datetime_samples)
-    @pytest.mark.parametrize(
-        "datestyle_out", ["ISO", "Postgres", "SQL", "German"]
-    )
-    @pytest.mark.parametrize("datestyle_in", ["DMY", "MDY", "YMD"])
+    @pytest.mark.parametrize("datestyle_out", datestyles_out)
+    @pytest.mark.parametrize("datestyle_in", datestyles_in)
     def test_load_datetime(self, conn, val, expr, datestyle_in, datestyle_out):
         cur = conn.cursor(binary=False)
         cur.execute(f"set datestyle = {datestyle_out}, {datestyle_in}")
@@ -167,9 +232,7 @@ class TestDatetime:
         assert cur.fetchone()[0] == as_dt(val)
 
     @pytest.mark.parametrize("val", ["min", "max"])
-    @pytest.mark.parametrize(
-        "datestyle_out", ["ISO", "Postgres", "SQL", "German"]
-    )
+    @pytest.mark.parametrize("datestyle_out", datestyles_out)
     def test_load_datetime_overflow(self, conn, val, datestyle_out):
         cur = conn.cursor(binary=False)
         cur.execute(f"set datestyle = {datestyle_out}, YMD")
@@ -190,6 +253,32 @@ class TestDatetime:
         with pytest.raises(DataError):
             cur.fetchone()[0]
 
+    overflow_samples = [
+        ("-infinity", "timestamp too small"),
+        ("1000-01-01 12:00 BC", "timestamp too small"),
+        ("10000-01-01 12:00", "timestamp too large"),
+        ("infinity", "timestamp too large"),
+    ]
+
+    @pytest.mark.parametrize("datestyle_out", datestyles_out)
+    @pytest.mark.parametrize("val, msg", overflow_samples)
+    def test_overflow_message(self, conn, datestyle_out, val, msg):
+        cur = conn.cursor()
+        cur.execute(f"set datestyle = {datestyle_out}, YMD")
+        cur.execute("select %s::timestamp", (val,))
+        with pytest.raises(DataError) as excinfo:
+            cur.fetchone()[0]
+        assert msg in str(excinfo.value)
+
+    @pytest.mark.parametrize("val, msg", overflow_samples)
+    def test_overflow_message_binary(self, conn, val, msg):
+        cur = conn.cursor(binary=True)
+        cur.execute("select %s::timestamp", (val,))
+        with pytest.raises(DataError) as excinfo:
+            cur.fetchone()[0]
+        assert msg in str(excinfo.value)
+
+    @crdb_skip_datestyle
     def test_load_all_month_names(self, conn):
         cur = conn.cursor(binary=False)
         cur.execute("set datestyle = 'Postgres'")
@@ -232,15 +321,15 @@ class TestDateTimeTz:
         cur.execute("set timezone to '-02:00'")
         cur.execute(
             f"""
-            select '{expr}'::timestamptz = %(val){fmt_in},
-            '{expr}', %(val){fmt_in}::text
+            select '{expr}'::timestamptz = %(val){fmt_in.value},
+            '{expr}', %(val){fmt_in.value}::text
             """,
             {"val": as_dt(val)},
         )
         ok, want, got = cur.fetchone()
         assert ok, (want, got)
 
-    @pytest.mark.parametrize("datestyle_in", ["DMY", "MDY", "YMD"])
+    @pytest.mark.parametrize("datestyle_in", datestyles_in)
     def test_dump_datetimetz_datestyle(self, conn, datestyle_in):
         tzinfo = dt.timezone(dt.timedelta(hours=2))
         cur = conn.cursor(binary=False)
@@ -263,6 +352,7 @@ class TestDateTimeTz:
         ("1900,1,1~05:21:10", "1900-01-01", "Asia/Calcutta"),
     ]
 
+    @crdb_skip_datestyle
     @pytest.mark.parametrize("val, expr, timezone", load_datetimetz_samples)
     @pytest.mark.parametrize("datestyle_out", ["ISO"])
     def test_load_datetimetz(self, conn, val, expr, timezone, datestyle_out):
@@ -280,12 +370,11 @@ class TestDateTimeTz:
         assert got == as_dt(val)
 
     @pytest.mark.xfail  # parse timezone names
+    @crdb_skip_datestyle
     @pytest.mark.parametrize("val, expr", [("2000,1,1~2", "2000-01-01")])
     @pytest.mark.parametrize("datestyle_out", ["SQL", "Postgres", "German"])
-    @pytest.mark.parametrize("datestyle_in", ["DMY", "MDY", "YMD"])
-    def test_load_datetimetz_tzname(
-        self, conn, val, expr, datestyle_in, datestyle_out
-    ):
+    @pytest.mark.parametrize("datestyle_in", datestyles_in)
+    def test_load_datetimetz_tzname(self, conn, val, expr, datestyle_in, datestyle_out):
         cur = conn.cursor(binary=False)
         cur.execute(f"set datestyle = {datestyle_out}, {datestyle_in}")
         cur.execute("set timezone to '-02:00'")
@@ -300,13 +389,12 @@ class TestDateTimeTz:
             ("Europe/Rome", "2000-1-1", 3600),
             ("Europe/Rome", "2000-7-1", 7200),
             ("Europe/Rome", "1000-1-1", 2996),
-            ("NOSUCH0", "2000-1-1", 0),
+            pytest.param("NOSUCH0", "2000-1-1", 0, marks=crdb_skip_invalid_tz),
+            pytest.param("0", "2000-1-1", 0, marks=crdb_skip_invalid_tz),
         ],
     )
     @pytest.mark.parametrize("fmt_out", pq.Format)
     def test_load_datetimetz_tz(self, conn, fmt_out, tzname, expr, tzoff):
-        if "/" in tzname and sys.platform == "win32":
-            pytest.skip("no IANA db on Windows")
         conn.execute("select set_config('TimeZone', %s, true)", [tzname])
         cur = conn.cursor(binary=fmt_out)
         ts = cur.execute("select %s::timestamptz", [expr]).fetchone()[0]
@@ -325,13 +413,16 @@ class TestDateTimeTz:
         val = as_dt(val)
         cur = conn.cursor()
         cur.execute(
-            f"select pg_typeof(%{fmt_in}) = %s::regtype, %{fmt_in}",
+            f"""
+            select pg_typeof(%{fmt_in.value})::regtype = %s::regtype, %{fmt_in.value}
+            """,
             [val, type, val],
         )
         rec = cur.fetchone()
         assert rec[0] is True, type
         assert rec[1] == val
 
+    @pytest.mark.crdb_skip("copy")
     def test_load_copy(self, conn):
         cur = conn.cursor(binary=False)
         with cur.copy(
@@ -351,16 +442,40 @@ class TestDateTimeTz:
         assert rec[0] == want
         assert rec[1] == 11111111
 
-    tz_sec = pytest.mark.skipif(
-        sys.version_info < (3, 7), reason="no seconds in tz offset"
-    )
+    overflow_samples = [
+        ("-infinity", "timestamp too small"),
+        ("1000-01-01 12:00+00 BC", "timestamp too small"),
+        ("10000-01-01 12:00+00", "timestamp too large"),
+        ("infinity", "timestamp too large"),
+    ]
+
+    @pytest.mark.parametrize("datestyle_out", datestyles_out)
+    @pytest.mark.parametrize("val, msg", overflow_samples)
+    def test_overflow_message(self, conn, datestyle_out, val, msg):
+        cur = conn.cursor()
+        cur.execute(f"set datestyle = {datestyle_out}, YMD")
+        cur.execute("select %s::timestamptz", (val,))
+        if datestyle_out == "ISO":
+            with pytest.raises(DataError) as excinfo:
+                cur.fetchone()[0]
+            assert msg in str(excinfo.value)
+        else:
+            with pytest.raises(NotImplementedError):
+                cur.fetchone()[0]
+
+    @pytest.mark.parametrize("val, msg", overflow_samples)
+    def test_overflow_message_binary(self, conn, val, msg):
+        cur = conn.cursor(binary=True)
+        cur.execute("select %s::timestamptz", (val,))
+        with pytest.raises(DataError) as excinfo:
+            cur.fetchone()[0]
+        assert msg in str(excinfo.value)
 
-    @pytest.mark.xfail(sys.platform == "win32", reason="no IANA db on Windows")
     @pytest.mark.parametrize(
         "valname, tzval, tzname",
         [
             ("max", "-06", "America/Chicago"),
-            pytest.param("min", "+09:18:59", "Asia/Tokyo", marks=[tz_sec]),
+            ("min", "+09:18:59", "Asia/Tokyo"),
         ],
     )
     @pytest.mark.parametrize("fmt_out", pq.Format)
@@ -401,8 +516,8 @@ class TestTime:
         cur = conn.cursor()
         cur.execute(
             f"""
-            select '{expr}'::time = %(val){fmt_in},
-                '{expr}'::time::text, %(val){fmt_in}::text
+            select '{expr}'::time = %(val){fmt_in.value},
+                '{expr}'::time::text, %(val){fmt_in.value}::text
             """,
             {"val": as_time(val)},
         )
@@ -451,7 +566,7 @@ class TestTimeTz:
     def test_dump_timetz(self, conn, val, expr, fmt_in):
         cur = conn.cursor()
         cur.execute("set timezone to '-02:00'")
-        cur.execute(f"select '{expr}'::timetz = %{fmt_in}", (as_time(val),))
+        cur.execute(f"select '{expr}'::timetz = %{fmt_in.value}", (as_time(val),))
         assert cur.fetchone()[0] is True
 
     @pytest.mark.parametrize(
@@ -492,13 +607,16 @@ class TestTimeTz:
         val = as_time(val)
         cur = conn.cursor()
         cur.execute(
-            f"select pg_typeof(%{fmt_in}) = %s::regtype, %{fmt_in}",
+            f"""
+            select pg_typeof(%{fmt_in.value})::regtype = %s::regtype, %{fmt_in.value}
+            """,
             [val, type, val],
         )
         rec = cur.fetchone()
         assert rec[0] is True, type
         assert rec[1] == val
 
+    @pytest.mark.crdb_skip("copy")
     def test_load_copy(self, conn):
         cur = conn.cursor(binary=False)
         with cur.copy(
@@ -523,29 +641,26 @@ class TestInterval:
     dump_timedelta_samples = [
         ("min", "-999999999 days"),
         ("1d", "1 day"),
-        ("-1d", "-1 day"),
+        pytest.param("-1d", "-1 day", marks=crdb_skip_negative_interval),
         ("1s", "1 s"),
-        ("-1s", "-1 s"),
-        ("-1m", "-0.000001 s"),
+        pytest.param("-1s", "-1 s", marks=crdb_skip_negative_interval),
+        pytest.param("-1m", "-0.000001 s", marks=crdb_skip_negative_interval),
         ("1m", "0.000001 s"),
         ("max", "999999999 days 23:59:59.999999"),
     ]
 
     @pytest.mark.parametrize("val, expr", dump_timedelta_samples)
-    @pytest.mark.parametrize(
-        "intervalstyle",
-        ["sql_standard", "postgres", "postgres_verbose", "iso_8601"],
-    )
+    @pytest.mark.parametrize("intervalstyle", intervalstyles)
     def test_dump_interval(self, conn, val, expr, intervalstyle):
-        cur = conn.cursor(binary=False)
+        cur = conn.cursor()
         cur.execute(f"set IntervalStyle to '{intervalstyle}'")
         cur.execute(f"select '{expr}'::interval = %t", (as_td(val),))
         assert cur.fetchone()[0] is True
 
     @pytest.mark.parametrize("val, expr", dump_timedelta_samples)
     def test_dump_interval_binary(self, conn, val, expr):
-        cur = conn.cursor(binary=True)
-        cur.execute(f"select '{expr}'::interval = %t", (as_td(val),))
+        cur = conn.cursor()
+        cur.execute(f"select '{expr}'::interval = %b", (as_td(val),))
         assert cur.fetchone()[0] is True
 
     @pytest.mark.parametrize(
@@ -581,6 +696,7 @@ class TestInterval:
         cur.execute(f"select '{expr}'::interval")
         assert cur.fetchone()[0] == as_td(val)
 
+    @crdb_skip_datestyle
     @pytest.mark.xfail  # weird interval outputs
     @pytest.mark.parametrize("val, expr", [("1d,1s", "1 day 1 sec")])
     @pytest.mark.parametrize(
@@ -593,9 +709,10 @@ class TestInterval:
         cur.execute(f"select '{expr}'::interval")
         assert cur.fetchone()[0] == as_td(val)
 
+    @pytest.mark.parametrize("fmt_out", pq.Format)
     @pytest.mark.parametrize("val", ["min", "max"])
-    def test_load_interval_overflow(self, conn, val):
-        cur = conn.cursor()
+    def test_load_interval_overflow(self, conn, val, fmt_out):
+        cur = conn.cursor(binary=fmt_out)
         cur.execute(
             "select %s + %s * '1s'::interval",
             (as_td(val), -1 if val == "min" else 1),
@@ -603,39 +720,7 @@ class TestInterval:
         with pytest.raises(DataError):
             cur.fetchone()[0]
 
-    def test_infinity_date_example(self, conn):
-        # NOTE: this is an example in the docs. Make sure it doesn't regress when
-        # adding binary datetime adapters
-        from datetime import date
-        from psycopg.types.datetime import DateLoader, DateDumper
-
-        class InfDateDumper(DateDumper):
-            def dump(self, obj):
-                if obj == date.max:
-                    return b"infinity"
-                else:
-                    return super().dump(obj)
-
-        class InfDateLoader(DateLoader):
-            def load(self, data):
-                if data == b"infinity":
-                    return date.max
-                else:
-                    return super().load(data)
-
-        cur = conn.cursor()
-        cur.adapters.register_dumper(date, InfDateDumper)
-        cur.adapters.register_loader("date", InfDateLoader)
-
-        rec = cur.execute(
-            "SELECT %s::text, %s::text", [date(2020, 12, 31), date.max]
-        ).fetchone()
-        assert rec == ("2020-12-31", "infinity")
-        rec = cur.execute(
-            "select '2020-12-31'::date, 'infinity'::date"
-        ).fetchone()
-        assert rec == (date(2020, 12, 31), date(9999, 12, 31))
-
+    @pytest.mark.crdb_skip("copy")
     def test_load_copy(self, conn):
         cur = conn.cursor(binary=False)
         with cur.copy(
@@ -661,9 +746,7 @@ class TestInterval:
 
 
 def as_date(s):
-    return (
-        dt.date(*map(int, s.split(","))) if "," in s else getattr(dt.date, s)
-    )
+    return dt.date(*map(int, s.split(","))) if "," in s else getattr(dt.date, s)
 
 
 def as_time(s):
@@ -672,7 +755,10 @@ def as_time(s):
     else:
         off = None
 
-    rv = dt.time(*map(int, s.split(","))) if "," in s else getattr(dt.time, s)
+    if "," in s:
+        rv = dt.time(*map(int, s.split(",")))  # type: ignore[arg-type]
+    else:
+        rv = getattr(dt.time, s)
     if off:
         rv = rv.replace(tzinfo=as_tzinfo(off))
 
@@ -692,7 +778,7 @@ def as_dt(s):
 
 def as_naive_dt(s):
     if "," in s:
-        rv = dt.datetime(*map(int, s.split(",")))
+        rv = dt.datetime(*map(int, s.split(",")))  # type: ignore[arg-type]
     else:
         rv = getattr(dt.datetime, s)
 
