@@ -4,13 +4,16 @@ from decimal import Decimal
 
 import pytest
 
-from psycopg import pq
+from psycopg import pq, sql
 from psycopg import errors as e
-from psycopg.sql import Identifier
 from psycopg.adapt import PyFormat
 from psycopg.types import range as range_module
 from psycopg.types.range import Range, RangeInfo, register_range
 
+from ..utils import eur
+from ..fix_crdb import is_crdb, crdb_skip_message
+
+pytestmark = pytest.mark.crdb_skip("range")
 
 type2sub = {
     "int4range": "int4",
@@ -26,10 +29,10 @@ tzinfo = dt.timezone(dt.timedelta(hours=2))
 samples = [
     ("int4range", None, None, "()"),
     ("int4range", 10, 20, "[]"),
-    ("int4range", -(2 ** 31), (2 ** 31) - 1, "[)"),
+    ("int4range", -(2**31), (2**31) - 1, "[)"),
     ("int8range", None, None, "()"),
     ("int8range", 10, 20, "[)"),
-    ("int8range", -(2 ** 63), (2 ** 63) - 1, "[)"),
+    ("int8range", -(2**63), (2**63) - 1, "[)"),
     ("numrange", Decimal(-100), Decimal("100.123"), "(]"),
     ("numrange", Decimal(100), None, "()"),
     ("numrange", None, Decimal(100), "()"),
@@ -60,8 +63,8 @@ range_classes = """
 @pytest.mark.parametrize("pgtype", range_names)
 @pytest.mark.parametrize("fmt_in", PyFormat)
 def test_dump_builtin_empty(conn, pgtype, fmt_in):
-    r = Range(empty=True)
-    cur = conn.execute(f"select 'empty'::{pgtype} = %{fmt_in}", (r,))
+    r = Range(empty=True)  # type: ignore[var-annotated]
+    cur = conn.execute(f"select 'empty'::{pgtype} = %{fmt_in.value}", (r,))
     assert cur.fetchone()[0] is True
 
 
@@ -70,7 +73,7 @@ def test_dump_builtin_empty(conn, pgtype, fmt_in):
 def test_dump_builtin_empty_wrapper(conn, wrapper, fmt_in):
     wrapper = getattr(range_module, wrapper)
     r = wrapper(empty=True)
-    cur = conn.execute(f"select 'empty' = %{fmt_in}", (r,))
+    cur = conn.execute(f"select 'empty' = %{fmt_in.value}", (r,))
     assert cur.fetchone()[0] is True
 
 
@@ -91,10 +94,10 @@ def test_dump_builtin_empty_wrapper(conn, wrapper, fmt_in):
     ],
 )
 def test_dump_builtin_array(conn, pgtype, fmt_in):
-    r1 = Range(empty=True)
-    r2 = Range(bounds="()")
+    r1 = Range(empty=True)  # type: ignore[var-annotated]
+    r2 = Range(bounds="()")  # type: ignore[var-annotated]
     cur = conn.execute(
-        f"select array['empty'::{pgtype}, '(,)'::{pgtype}] = %{fmt_in}",
+        f"select array['empty'::{pgtype}, '(,)'::{pgtype}] = %{fmt_in.value}",
         ([r1, r2],),
     )
     assert cur.fetchone()[0] is True
@@ -103,10 +106,11 @@ def test_dump_builtin_array(conn, pgtype, fmt_in):
 @pytest.mark.parametrize("pgtype", range_names)
 @pytest.mark.parametrize("fmt_in", PyFormat)
 def test_dump_builtin_array_with_cast(conn, pgtype, fmt_in):
-    r1 = Range(empty=True)
-    r2 = Range(bounds="()")
+    r1 = Range(empty=True)  # type: ignore[var-annotated]
+    r2 = Range(bounds="()")  # type: ignore[var-annotated]
     cur = conn.execute(
-        f"select array['empty'::{pgtype}, '(,)'::{pgtype}] = %{fmt_in}::{pgtype}[]",
+        f"select array['empty'::{pgtype}, '(,)'::{pgtype}] "
+        f"= %{fmt_in.value}::{pgtype}[]",
         ([r1, r2],),
     )
     assert cur.fetchone()[0] is True
@@ -118,19 +122,17 @@ def test_dump_builtin_array_wrapper(conn, wrapper, fmt_in):
     wrapper = getattr(range_module, wrapper)
     r1 = wrapper(empty=True)
     r2 = wrapper(bounds="()")
-    cur = conn.execute(
-        f"""select '{{empty,"(,)"}}' = %{fmt_in}""", ([r1, r2],)
-    )
+    cur = conn.execute(f"""select '{{empty,"(,)"}}' = %{fmt_in.value}""", ([r1, r2],))
     assert cur.fetchone()[0] is True
 
 
 @pytest.mark.parametrize("pgtype, min, max, bounds", samples)
 @pytest.mark.parametrize("fmt_in", PyFormat)
 def test_dump_builtin_range(conn, pgtype, min, max, bounds, fmt_in):
-    r = Range(min, max, bounds)
+    r = Range(min, max, bounds)  # type: ignore[var-annotated]
     sub = type2sub[pgtype]
     cur = conn.execute(
-        f"select {pgtype}(%s::{sub}, %s::{sub}, %s) = %{fmt_in}",
+        f"select {pgtype}(%s::{sub}, %s::{sub}, %s) = %{fmt_in.value}",
         (min, max, bounds, r),
     )
     assert cur.fetchone()[0] is True
@@ -139,7 +141,7 @@ def test_dump_builtin_range(conn, pgtype, min, max, bounds, fmt_in):
 @pytest.mark.parametrize("pgtype", range_names)
 @pytest.mark.parametrize("fmt_out", pq.Format)
 def test_load_builtin_empty(conn, pgtype, fmt_out):
-    r = Range(empty=True)
+    r = Range(empty=True)  # type: ignore[var-annotated]
     cur = conn.cursor(binary=fmt_out)
     (got,) = cur.execute(f"select 'empty'::{pgtype}").fetchone()
     assert type(got) is Range
@@ -151,7 +153,7 @@ def test_load_builtin_empty(conn, pgtype, fmt_out):
 @pytest.mark.parametrize("pgtype", range_names)
 @pytest.mark.parametrize("fmt_out", pq.Format)
 def test_load_builtin_inf(conn, pgtype, fmt_out):
-    r = Range(bounds="()")
+    r = Range(bounds="()")  # type: ignore[var-annotated]
     cur = conn.cursor(binary=fmt_out)
     (got,) = cur.execute(f"select '(,)'::{pgtype}").fetchone()
     assert type(got) is Range
@@ -165,24 +167,20 @@ def test_load_builtin_inf(conn, pgtype, fmt_out):
 @pytest.mark.parametrize("pgtype", range_names)
 @pytest.mark.parametrize("fmt_out", pq.Format)
 def test_load_builtin_array(conn, pgtype, fmt_out):
-    r1 = Range(empty=True)
-    r2 = Range(bounds="()")
+    r1 = Range(empty=True)  # type: ignore[var-annotated]
+    r2 = Range(bounds="()")  # type: ignore[var-annotated]
     cur = conn.cursor(binary=fmt_out)
-    (got,) = cur.execute(
-        f"select array['empty'::{pgtype}, '(,)'::{pgtype}]"
-    ).fetchone()
+    (got,) = cur.execute(f"select array['empty'::{pgtype}, '(,)'::{pgtype}]").fetchone()
     assert got == [r1, r2]
 
 
 @pytest.mark.parametrize("pgtype, min, max, bounds", samples)
 @pytest.mark.parametrize("fmt_out", pq.Format)
 def test_load_builtin_range(conn, pgtype, min, max, bounds, fmt_out):
-    r = Range(min, max, bounds)
+    r = Range(min, max, bounds)  # type: ignore[var-annotated]
     sub = type2sub[pgtype]
     cur = conn.cursor(binary=fmt_out)
-    cur.execute(
-        f"select {pgtype}(%s::{sub}, %s::{sub}, %s)", (min, max, bounds)
-    )
+    cur.execute(f"select {pgtype}(%s::{sub}, %s::{sub}, %s)", (min, max, bounds))
     # normalise discrete ranges
     if r.upper_inc and isinstance(r.upper, int):
         bounds = "[)" if r.lower_inc else "()"
@@ -208,20 +206,16 @@ def test_copy_in(conn, min, max, bounds, format):
     if bounds != "empty":
         min = dt.date(*map(int, min.split(","))) if min else None
         max = dt.date(*map(int, max.split(","))) if max else None
-        r = Range(min, max, bounds)
+        r = Range[dt.date](min, max, bounds)
     else:
         r = Range(empty=True)
 
     try:
-        with cur.copy(
-            f"copy copyrange (r) from stdin (format {format.name})"
-        ) as copy:
+        with cur.copy(f"copy copyrange (r) from stdin (format {format.name})") as copy:
             copy.write_row([r])
     except e.ProtocolViolation:
         if not min and not max and format == pq.Format.BINARY:
-            pytest.xfail(
-                "TODO: add annotation to dump ranges with no type info"
-            )
+            pytest.xfail("TODO: add annotation to dump ranges with no type info")
         else:
             raise
 
@@ -239,9 +233,7 @@ def test_copy_in_empty_wrappers(conn, bounds, wrapper, format):
     cls = getattr(range_module, wrapper)
     r = cls(empty=True) if bounds == "empty" else cls(None, None, bounds)
 
-    with cur.copy(
-        f"copy copyrange (r) from stdin (format {format.name})"
-    ) as copy:
+    with cur.copy(f"copy copyrange (r) from stdin (format {format.name})") as copy:
         copy.write_row([r])
 
     rec = cur.execute("select r from copyrange order by id").fetchone()
@@ -255,11 +247,12 @@ def test_copy_in_empty_set_type(conn, bounds, pgtype, format):
     cur = conn.cursor()
     cur.execute(f"create table copyrange (id serial primary key, r {pgtype})")
 
-    r = Range(empty=True) if bounds == "empty" else Range(None, None, bounds)
+    if bounds == "empty":
+        r = Range(empty=True)  # type: ignore[var-annotated]
+    else:
+        r = Range(None, None, bounds)
 
-    with cur.copy(
-        f"copy copyrange (r) from stdin (format {format.name})"
-    ) as copy:
+    with cur.copy(f"copy copyrange (r) from stdin (format {format.name})") as copy:
         copy.set_types([pgtype])
         copy.write_row([r])
 
@@ -273,6 +266,9 @@ def testrange(svcconn):
 
 
 def create_test_range(conn):
+    if is_crdb(conn):
+        pytest.skip(crdb_skip_message("range"))
+
     conn.execute(
         """
         create schema if not exists testschema;
@@ -289,8 +285,8 @@ def create_test_range(conn):
 fetch_cases = [
     ("testrange", "text"),
     ("testschema.testrange", "float8"),
-    (Identifier("testrange"), "text"),
-    (Identifier("testschema", "testrange"), "float8"),
+    (sql.Identifier("testrange"), "text"),
+    (sql.Identifier("testschema", "testrange"), "float8"),
 ]
 
 
@@ -326,7 +322,7 @@ def test_dump_custom_empty(conn, testrange):
     info = RangeInfo.fetch(conn, "testrange")
     register_range(info, conn)
 
-    r = Range(empty=True)
+    r = Range[str](empty=True)
     cur = conn.execute("select 'empty'::testrange = %s", (r,))
     assert cur.fetchone()[0] is True
 
@@ -367,10 +363,10 @@ def test_load_quoting(conn, testrange, fmt_out):
             "select testrange(chr(%(low)s::int), chr(%(up)s::int))",
             {"low": i, "up": i + 1},
         )
-        got = cur.fetchone()[0]
+        got: Range[str] = cur.fetchone()[0]
         assert isinstance(got, Range)
-        assert ord(got.lower) == i
-        assert ord(got.upper) == i + 1
+        assert got.lower and ord(got.lower) == i
+        assert got.upper and ord(got.upper) == i + 1
 
 
 @pytest.mark.parametrize("fmt_out", pq.Format)
@@ -389,7 +385,7 @@ def test_mixed_array_types(conn, fmt_out):
 
 class TestRangeObject:
     def test_noparam(self):
-        r = Range()
+        r = Range()  # type: ignore[var-annotated]
 
         assert not r.isempty
         assert r.lower is None
@@ -400,7 +396,7 @@ class TestRangeObject:
         assert not r.upper_inc
 
     def test_empty(self):
-        r = Range(empty=True)
+        r = Range(empty=True)  # type: ignore[var-annotated]
 
         assert r.isempty
         assert r.lower is None
@@ -463,8 +459,9 @@ class TestRangeObject:
             Range(bounds="[}")
 
     def test_in(self):
-        r = Range(empty=True)
+        r = Range[int](empty=True)
         assert 10 not in r
+        assert "x" not in r  # type: ignore[operator]
 
         r = Range()
         assert 10 in r
@@ -665,4 +662,16 @@ class TestRangeObject:
 
 def test_no_info_error(conn):
     with pytest.raises(TypeError, match="range"):
-        register_range(None, conn)
+        register_range(None, conn)  # type: ignore[arg-type]
+
+
+@pytest.mark.parametrize("name", ["a-b", f"{eur}", "order"])
+def test_literal_invalid_name(conn, name):
+    conn.execute("set client_encoding to utf8")
+    conn.execute(f'create type "{name}" as range (subtype = text)')
+    info = RangeInfo.fetch(conn, f'"{name}"')
+    register_range(info, conn)
+    obj = Range("a", "z", "[]")
+    assert sql.Literal(obj).as_string(conn) == f"'[a,z]'::\"{name}\""
+    cur = conn.execute(sql.SQL("select {}").format(obj))
+    assert cur.fetchone()[0] == obj
